#include <iprs.h>
#include "sure.h"
#include "sureextern.h"

extern int verbosity;

/* This procedure is for calculating the mean of the points in each cluster.
 * The idea is for using these means for the fuzzy clustering stuff.
 */

extern int nunary_features, nbinary_features;

/*
 * This function calculates the distance between the given point and all
 * the clusters in the cluster list in a fuzzy c-means sense.
 */

float *bren_calc_distances_fcm(float *pt, LIST clusters, int nfeatures,
						   float fuzzyq)
{
	float *distances, sum=0;
	int cluster_index=0;
	int i;

	distances = (float*)iprs_calloc(list_length(clusters),sizeof(float));
	foreach(CLUSTER, c, clusters)
	{
		distances[cluster_index]=multid_feuclid_dist(pt,c->mean,nfeatures);
		distances[cluster_index]= pow(((float)1.0/distances[cluster_index]), 
							1.0/(fuzzyq-1));
		sum += distances[cluster_index++];
	}
	endforeach;
	cluster_index=0;
	foreach (CLUSTER, c, clusters)
		distances[cluster_index++] /= sum;
	endforeach;
	return(distances);
}

/*
 * This function calculates the distance between the given point and all
 * the clusters in the cluster list in a possibilistic sense.
 */

float *bren_calc_distances_possibilistic(float *pt, LIST clusters, 
										 int nfeatures, float fuzzyq)
{
	float *distances, sum=0;
	int cluster_index=0;
	int i;

	distances = (float*)iprs_calloc(list_length(clusters),sizeof(float));
	foreach(CLUSTER, c, clusters)
	{
		distances[cluster_index]=multid_feuclid_dist(pt,c->mean,nfeatures);
		distances[cluster_index]= ((float)1.0/(1.0+
											   pow(distances[cluster_index], 
												   2.0/(fuzzyq-1)))); 
		cluster_index++;
	}
	endforeach;
	return(distances);
}


/*
 * Choose the best ncluster clusters from ctree and add them to cqueue.
 * Add the parts to pqueue and the fuzzy classification to fqueue.
 */

LIST bren_sure_add_best_clusters(CLUSTER *parent, SNAKE *snake,
								 LIST fuzzy_cfn, char type,
								 int nclusters, float fuzzyq)
{
	float *distances, *best_cluster_values, sum, *point;
	int cluster_index, i, nfeatures;
	CLUSTER **best_clusters;
	LIST clusters;
	LIST new_snake_list;

	clusters = parent->children;
	new_snake_list = NIL;

	if (type=='U')
	{
		nfeatures=nunary_features;
		point = snake->path->unary[snake->path->length-1]->value;
	}
	else 
	{
		nfeatures=nbinary_features;
		point = snake->path->binary[snake->path->length-1]->value;
	}

	if (verbosity>1)
	{
		int i;
		printf("classifying path: ");
		for (i=0; i<snake->path->length; i++)
		{
			printf("%d ", snake->path->part[i]->name);
		}
		printf(": (");
		for (i=0; i<nfeatures; i++)
			printf("%f ", point[i]);
		printf(")\n");
		printf("expanding cluster %s\n", parent->name);
	}

	distances=bren_calc_distances_possibilistic(point, clusters,
												nfeatures, fuzzyq);

	best_cluster_values=(float*)iprs_calloc(nclusters,sizeof(float));
	best_clusters=(CLUSTER**)iprs_calloc(nclusters,sizeof(CLUSTER*));
	cluster_index=0;
	/* Pick out the best ncluster clusters for further expansion */
	foreach(CLUSTER,c1,clusters)
	{
		if (verbosity>1)
		{
			int i;
			for (i=0; i<nfeatures;i++)
				printf("%f ", c1->mean[i]);
			printf("distances[%d]=%f\n",cluster_index,distances[cluster_index]);
		}
		for (i=0;i<nclusters;i++)
		{
			if (distances[cluster_index] > best_cluster_values[i])
			{
				int j;
				/* shuffle along */
				for (j=i;j<nclusters-1;j++)
				{
					best_cluster_values[j+1] = best_cluster_values[j];
					best_clusters[j+1] = best_clusters[j];
				}
				best_cluster_values[i]=distances[cluster_index];
				best_clusters[i]=c1;
				break;
			}
		}
		cluster_index++;
	}
	endforeach;
	if (verbosity>1) printf("best clusters are:\n");
	for (i=0; i<nclusters; i++)
	if ((best_clusters[i])&&(best_cluster_values[i]>minimum_fuzzy_membership))
	{
		CLUSTER *c1;
		CLASSIFICATION *cfn;
		SNAKE *new_snake;
		
		c1 = best_clusters[i];
		if (verbosity>1)
		{
			printf("RULE2: %f\n", best_cluster_values[i]); 
/* 			print_cluster(c1); */
		}
		new_snake = create_snake(snake->path);
		cfn = create_classification(new_snake->path, c1, fuzzy_cfn, 
									best_cluster_values[i]);
		list_append(cfn, new_snake->classifications);
		list_append(new_snake, new_snake_list);
	}
	if (verbosity>1) printf("\n");
	bren_free(distances, iprs_free);
	bren_free(best_cluster_values, iprs_free);
	bren_free(best_clusters, iprs_free);
	return(new_snake_list);
}


