/* ====================================================================
 * Program:     CRG (Conditional Rule Generation)
 * Author:      Walter F. Bischof
 * ==================================================================== */

/* Include files */

#include <iprs.h>
#include "crg.h"

/* Global Declarations */

LIST      binary_list;
float    *binary_lower;
float    *binary_difference_threshold;
float    *binary_mean;
float    *binary_range;
float    *binary_sdev;
float    *binary_upper;
float     expected_rule_cost;
int       full_feature_range;
int       max_path_length;
float     min_cluster_size;            /* experimental */
int       nbinary_features;
int       nclasses;
int       nnodes_analyzed;
int       nparts;
int       npaths;
int       npaths_analyzed;
int       nunary_features;
int       nviews;
LIST      part_list;
PATHNODE *path_tree;
LIST      unary_list;
float    *unary_lower;
float    *unary_difference_threshold;
float    *unary_mean;
float    *unary_range;
float    *unary_sdev;
float    *unary_upper;

/* Parameters */

int       breadth_first;
int       crg_debug_clusters;
int       crg_debug_parts;
int       crg_debug_paths;
int       crg_debug_varia;
float     entropy_threshold;
int       entropy_wrt_all_classes;
int       feature_ordering;
int       fixed_path_length;
float     leader_threshold;
int       max_level;
int       path_splitting;
int       symmetric_relations;
int		  use_leader_cluster;
int       zscore;
/*
 * The leaveout flags are used for removing a particular feature from
 * the input feature spaces. This is useful for testing the performance
 * of the classifier without having to re-extract the features from scratch.
 */
int       nunary_leaveout=0, unary_leaveout[40];
int       nbinary_leaveout=0, binary_leaveout[40];
int ninitial_clusters=10; /* number of initial clusters for fuzzy kmeans */
float fuzzyq=1.5;  /* determines the fuzziness of clusters (must be > 1.0) */
float delta_centres_threshold=0.01; /* dist centres must not move before */
                                  /* stopping k-means clustering */

/* --------------------------------------------------------------------
 * Procedure: Main
 * Usage:     crg [switches] filename
 *            for switches see below
 * -------------------------------------------------------------------- */

int main(int argc, char **argv)
{
	char *infilename,*outfilename;
	CLUSTER *root;
	int ndebug_options=0;
	char *debug_option[20];
	float feature_difference_threshold;
	extern char *optarg;
	extern int optind;
	int opt;

	char *help = 
	"crg [switches] filename\n"
	"-a        full feature range (see crgcluster)\n"
	"-b        breadth-first expansion of cluster tree (default=depth-first)\n"
	"-c <f>    min cluster size for splitting (default=0)\n"
	"-d option print debugging information\n"
	"-e <f>    min entropy for splitting (default H>0)\n"
	"-E        compute entropy wrt to all classes\n"
	"-f <f>    feature difference threshold for splitting (default=0)\n"
	"-l n      expand cluster tree to level n (default=5)\n"
	"-L        use fixed length paths only\n"
	"-P        use path splitting\n"
	"-r file   save results in file (default=training.result)\n"
	"-s        binary relations are symmetric\n"
	"-t <f>    threshold for leader cluster (default=no cluster)\n"
	"-z        z-transform feature values\n"
	"filename  (default=training.data)\n"
	"-x <d>    leave out unary feature d"
	"-y <d>    leave out binary feature d"
	"\n"
	"debugging options\n"
	"clusters  print clusters during cluster tree generation\n"
	"parts     print part information\n"
	"paths     print path information\n"
	"varia     print various informations\n";

	timer();

	unary_list  = NIL;
	binary_list = NIL;
	part_list   = NIL;

	breadth_first=0;
	entropy_threshold=0.0;
	entropy_wrt_all_classes=0;
	feature_difference_threshold=0.02;
	feature_ordering=0;
	fixed_path_length=0;
	full_feature_range=0;
	leader_threshold=1.0e+9;
	max_level=5;
	path_splitting=0;
	symmetric_relations=0;
	use_leader_cluster=0;
	zscore=0;
	debug_option[0]=(char *)NULL;

	infilename="training.data";
	outfilename="training.result";

	while ((opt=getopt(argc,argv,"d:e:l:r:szx:y:"))!=EOF) 
	{
		switch (opt) 
		{
			case 'd':
				debug_option[ndebug_options++] = optarg;
				break;
			case 'e':
				entropy_threshold = atof(optarg);
				break;
			case 'l':
				max_level = atoi(optarg);
				break;
			case 'r':
				outfilename = optarg;
				break;
			case 's':
				symmetric_relations=1;
				break;
			case 'z': 
				zscore=1;
				break;
			case 'x':
				unary_leaveout[nunary_leaveout++]=atoi(optarg);
				break;
			case 'y':
				binary_leaveout[nbinary_leaveout++] = 
					atoi(optarg);
				break;
		}
	}
	if (optind < argc)
		infilename = argv[optind];
	printf("reading file %s\n", infilename);

	if (use_leader_cluster)
		printf("leader cluster threshold is:%f\n", leader_threshold);

	process_debug_options(ndebug_options,debug_option);
	read_database(infilename);

/*
 * Modified by B. McCane 1/8/94 so that zscore transformations are saved in
 * the training results file. Then, if an unsighted object is used for
 * testing, the same zscore transformations will first be performed. Of
 * course if zscore is not used then the default values are mean=0 and sdev=1,
 * which results in no transformation.
 */
	unary_mean=(float *)iprs_calloc(nunary_features,sizeof(float));
	unary_sdev=(float *)iprs_calloc(nunary_features,sizeof(float));
	binary_mean=(float *)iprs_calloc(nbinary_features,sizeof(float));
	binary_sdev=(float *)iprs_calloc(nbinary_features,sizeof(float));
	compute_mean_sdev();
	print_parts();
	if (zscore) compute_zscore();

	compute_range(feature_difference_threshold);
	if(crg_debug_parts)
	{ 
		int i;

		printf("unary_mean = [ ");
		for (i=0; i<nunary_features; i++)
			printf("%f ", unary_mean[i]);
		printf("]\nunary sdev = [ ");
		for (i=0; i<nunary_features; i++)
			printf("%f ", unary_sdev[i]);

		printf("]\nbinary_mean = [ ");
		for (i=0; i<nbinary_features; i++)
			printf("%f ", binary_mean[i]);
		printf("\nbinary sdev = [ ");
		for (i=0; i<nbinary_features; i++)
			printf("%f ", binary_sdev[i]);
		printf("]\n");

		print_parts();
		printf("infilename is: %s\n", infilename);
		output_xgobi(infilename);
	}
	path_tree=create_path_tree();
	root=create_root_cluster();
	create_cluster_tree(root,ninitial_clusters,fuzzyq,delta_centres_threshold);
/* 	bren_calculate_means(root); */
/* 	sort_cluster_tree(root); */
/* 	analyze_cluster_tree(root); */
    {
		LIST leaf_list=NIL;
		int leaf_number;
		print_cluster_tree(root, leaf_list, &leaf_number);
	}
	save_training(root,infilename,outfilename);
/* 	if (crg_debug_clusters) */
/* 		output_clustertree_preorder(root); */

	printf("Time:                      %-3.1f secs\n",timer());

	return(0);
}

/* --------------------------------------------------------------------
 * Procedure: read_database
 * Comments:  Create unary, binary and part descriptors from database
 *            We assume fs0=unary, fs1=binary and fs2=label.
 * -------------------------------------------------------------------- */

void read_database(char *filename)
{
	IPRS_IMAGE *db,*cptr,*vptr;
	IPRS_FEATURE_SPACE *fs0ptr,*fs1ptr,*fs2ptr;
	int c,v,i,n,p1,p2, j, tmp;
	PART **part;
	BINARY *binary;
	UNARY *unary;
	int *unary_usefeatures, *binary_usefeatures;

	if (crg_debug_varia)
		printf("Reading database %s ...",filename);

	db=iprs_load_fsdesc(filename);

	if (crg_debug_varia)
		printf(" creating descriptions ...");

	/* nclasses, nunary_features and nbinary_features */

	nclasses = db->numimages;
	cptr=(IPRS_IMAGE *)(db->image[0]);
	vptr=(IPRS_IMAGE *)(cptr->image[0]);
	fs0ptr=(IPRS_FEATURE_SPACE *)(vptr->image[0]);
	nunary_features=fs0ptr->numfeatures;
	fs1ptr=(IPRS_FEATURE_SPACE *)(vptr->image[1]);
	nbinary_features=fs1ptr->numfeatures;
	nviews=0;
	nparts=0;

	unary_usefeatures = (int*)iprs_calloc(nunary_features, sizeof(int));
	for (i=0; i<nunary_leaveout; i++)
	{
		if (unary_leaveout[i]<nunary_features) 
			unary_usefeatures[unary_leaveout[i]]=1;
		else fprintf(stderr,
			     "Not that many unary features to leave out\n");
	}
	tmp = nunary_features;
	for (i=0; i<tmp; i++)
		if (unary_usefeatures[i]) nunary_features--;
	binary_usefeatures = (int*)iprs_calloc(nbinary_features, sizeof(int));
	for (i=0; i<nbinary_leaveout; i++)
	{
		if (binary_leaveout[i]<nbinary_features) 
			binary_usefeatures[binary_leaveout[i]]=1;
		else fprintf(stderr,
			     "Not that many binary features to leave out\n");
	}
	tmp = nbinary_features;
	for (i=0; i<tmp; i++)
		if (binary_usefeatures[i]) nbinary_features--;

	/* create part, unary and binary descriptors */

	for(c=0;c<db->numimages;c++) {
		cptr=(IPRS_IMAGE *)(db->image[c]);
		for(v=0;v<cptr->numimages;v++) {
			vptr=(IPRS_IMAGE *)(cptr->image[v]);
			fs0ptr=bren_removefeatures_featurespace(
					   (IPRS_FEATURE_SPACE *)(vptr->image[0]), 
					   unary_usefeatures, nunary_features);
			fs1ptr=bren_removefeatures_featurespace(
                       (IPRS_FEATURE_SPACE *)(vptr->image[1]),
					   binary_usefeatures, nbinary_features);
			fs2ptr=(IPRS_FEATURE_SPACE *)(vptr->image[2]);

			/* part */

			n=fs0ptr->numpoints;
			part=(PART **)iprs_calloc(n,sizeof(PART *));
			for (i=0;i<n;i++) {
				part[i]=(PART *)iprs_calloc(1,sizeof(PART));
				part[i]->name=i+nparts;
				part[i]->view=v+nviews;
				part[i]->class=c;
				part[i]->binary=NIL;
				list_append(part[i],part_list);
			}

			/* unary */

			for(i=0;i<n;i++) {
				unary=(UNARY *)iprs_calloc(1,sizeof(UNARY));
				unary->part=part[i];
				unary->view=v+nviews;
				unary->class=c;
				unary->value=fs0ptr->point[i];
				list_append(unary,unary_list);
				part[i]->unary=unary;
			}

			/* binary */

			check_labels(fs2ptr->point,fs2ptr->numpoints,fs1ptr->numpoints,
						 n,c,v);

			for (i=0;i<fs1ptr->numpoints;i++) {
				p1=(int)(fs2ptr->point[i][0]);
				p2=(int)(fs2ptr->point[i][1]);
				binary=(BINARY *)iprs_calloc(1,sizeof(BINARY));
				binary->part1=part[p1];
				binary->part2=part[p2];
				binary->view=v+nviews;
				binary->class=c;
				binary->value=fs1ptr->point[i];
				list_append(binary,binary_list);
				list_append(binary,part[p1]->binary);

				if (symmetric_relations) {
					binary=(BINARY *)iprs_calloc(1,sizeof(BINARY));
					binary->part1=part[p2];
					binary->part2=part[p1];
					binary->view=v+nviews;
					binary->class=c;
					binary->value=fs1ptr->point[i];
					list_append(binary,binary_list);
					list_append(binary,part[p2]->binary);
				}
			}
			nparts+=n;
			iprs_free(part);
		}
		nviews+=cptr->numimages;
	}

	if (crg_debug_varia) printf(" done\n");
}

/* --------------------------------------------------------------------
 * Procedure: create_root_cluster
 * Comment:   create root of cluster tree.
 * -------------------------------------------------------------------- */

CLUSTER *create_root_cluster()
{
	CLUSTER *c;
	int i;

	c=(CLUSTER *)iprs_calloc(1,sizeof(CLUSTER));
	c->type='0';
	c->level=0;
	c->name="";
	c->parent=(CLUSTER *)NULL;
	c->children=NIL;
	c->elements=NIL;
	c->pathnodes=NIL;
	c->nclass=nclasses;
	c->class=(int *)iprs_calloc(nclasses,sizeof(int));
	for (i=0;i<nclasses;i++) c->class[i]=0;
	foreach(PART,part,part_list)
		(c->class[part->class])++;
	endforeach
	c->entropy=compute_entropy(c->class);
/* 	c->weighted_entropy=c->entropy; */
	c->cfs=(FEATURE_LIST *)NULL;
	c->cfs_possible=1;
	return(c);
}

/* --------------------------------------------------------------------
 * Procedure: create_cluster_tree
 * Comment:   create cluster tree. Generation is limited by max_level,
 *            by cluster entropy and by number of parts.
 *
 *            Termination of cluster tree generation ("termination" below)
 *            for complete resolution:
 *                 if (c->nclass > 1)
 *            using entropy (pretty useless if you think about it)
 *                 if (c->entropy > entropy_threshold)
 *            using weighted entropy
 *                 see below
 * -------------------------------------------------------------------- */

void create_cluster_tree(CLUSTER *root, int nclusters, float fuzzyq,
						 float delta_centres_threshold)
{
	LIST queue, clist;
	CLUSTER *c;
	FEATURE_LIST *f;

	nnodes_analyzed=0;
	queue = list_list(root);
	while ( !list_empty(queue) ) {
		nnodes_analyzed++;
		c = list_remove_first(queue);

		if (c->entropy > entropy_threshold) {    /* termination */
			if (crg_debug_clusters)
				printf("entropy=%f\n", c->entropy);
			if (c->level >= max_level) {
				if (crg_debug_clusters)
				{
					printf("CCT: level > maxlevel: splitting\n");
				}
				clist=fuzzy_split2(c,queue, fuzzyq, delta_centres_threshold);
				add_clist_to_queue(clist,queue);
			}
			else {
				if (c->cfs_possible==0) {
					if (crg_debug_clusters)
						printf("CCT: not expansible, splitting\n");
					clist=fuzzy_split2(c,queue,fuzzyq,delta_centres_threshold);
					add_clist_to_queue(clist,queue);
				}
				else {
					f = create_feature_list(c);
					if (f != (FEATURE_LIST *)NULL) {
						if (crg_debug_clusters)
							printf("CCT: expanding tree\n");
						clist = cluster_feature_list(f, nclusters, fuzzyq,
													 delta_centres_threshold);
						add_clist_to_queue(clist,queue);
					}
					else {
						if (crg_debug_clusters)
							printf("CCT: no more paths, must split\n");
/* 						if (path_splitting) { */
/* 							clist=split_paths(c,queue); */
/* 							add_clist_to_queue(clist,queue); */
/* 						} */
						{
							clist=fuzzy_split2(c,queue,fuzzyq,
											   delta_centres_threshold);
							add_clist_to_queue(clist,queue);
						}
					}
				}
			}
		}
	}
}

/* --------------------------------------------------------------------
 * Procedure: create_feature_list
 * Comments:  determine all features for the feature space conditional
 *            on a given cluster.
 * -------------------------------------------------------------------- */

FEATURE_LIST *create_feature_list(CLUSTER *c)
{
	FEATURE_LIST *f;

	f=(FEATURE_LIST *)iprs_calloc(1,sizeof(FEATURE_LIST));

	/* --------------------------------------------------------
	   root cluster: f->type=U f->elements={u(p)}
	   -------------------------------------------------------- */

	if (c->level == 0) {
		f->type='U';
		f->level=1;
		f->parent=c;
		f->pathnodes=list_copy(path_tree->children);
		f->elements=NIL;
		foreach(PATHNODE,p,f->pathnodes)
			list_union(p->unary,f->elements);
		endforeach
/* 		f->npathnodes=list_length(f->pathnodes); */
		c->cfs=f;
		c->cfs_possible=1;
		return(f);
	}

	/* --------------------------------------------------------
	   c->type=U c->elements={u(p)}
	   f->type=B f->elements={b(p,q)}
	   -------------------------------------------------------- */

	if (c->type == 'U') {
		f->type = 'B';
		f->level=c->level+1;
		f->parent=c;
		f->pathnodes=NIL;
		f->elements=NIL;
		foreach(PATHNODE,p1,c->pathnodes) {

			if (list_empty(p1->children)) {

				/* path cannot be extended. fail */

				c->cfs_possible= -1;
				list_free(f->pathnodes);
				list_free(f->elements);
				iprs_free(f);
				return((FEATURE_LIST *)NULL);
			}

			foreach(PATHNODE,p2,p1->children) {
				list_append(p2,f->pathnodes);
				list_union(p2->binary,f->elements);
			}
			endforeach
		}
		endforeach
/* 		f->npathnodes=list_length(f->pathnodes); */
		c->cfs=f;
		c->cfs_possible=1;
		return(f);
	}

	/* --------------------------------------------------------
	   c->type=B c->elements={b(p,q)}
	   f->type=U f->elements={u(q)}
	   -------------------------------------------------------- */

	else /* c->type == 'B' */ {
		f->type = 'U';
		f->level=c->level+1;
		f->parent=c;
		f->pathnodes=list_copy(c->pathnodes);
		f->elements=NIL;
		foreach(PATHNODE,p,f->pathnodes)
			list_union(p->unary,f->elements);
		endforeach
/* 		f->npathnodes=list_length(f->pathnodes); */
		c->cfs=f;
		c->cfs_possible=1;
		return(f);
	}
}

/* --------------------------------------------------------------------
 * Procedure: create_feature_list_for_cluster
 * Comments:  determine all features in a given cluster as opposed to
 *            determining all features that are children of the given
 *            cluster.
 * -------------------------------------------------------------------- */

FEATURE_LIST *create_feature_list_for_cluster(CLUSTER *c)
{
	FEATURE_LIST *f;

	f=(FEATURE_LIST *)iprs_calloc(1,sizeof(FEATURE_LIST));


	f->type = c->type;
	f->level=c->level;
	f->parent=c->parent;
	f->elements=list_copy(c->elements);
	f->pathnodes=list_copy(c->pathnodes);
/* 	f->npathnodes=list_length(f->pathnodes); */
	return(f);
}

void free_feature_list(FEATURE_LIST *f)
{
	if (f)
	{
		if (f->elements) list_free(f->elements);
		if (f->pathnodes) list_free(f->pathnodes);
		iprs_free(f);
	}
}

/* --------------------------------------------------------------------
 * Procedure: cluster_feature_list
 * Comments: cluster a feature list. List of clusters is returned.
 * -------------------------------------------------------------------- */

LIST cluster_feature_list(FEATURE_LIST *f, int max_nclusters, float fuzzyq,
						  float delta_centres_threshold)
{
	float **tmp_cluster_centres, **cluster_centres=NULL;
	CLUSTER *c,*parent;
	LIST clist;
	int nrules,i,r,nfeatures, nclusters;
	float pratio;
	char *s;
	float valid, min_valid=FLT_MAX;
	int best_nclusters;
	FILE *listfile;

	nclusters = list_length(f->elements)/max_nclusters + 1; 
/* 	if (list_length(f->elements)/4+1<2) */
	if (nclusters<2)
	{
		cluster_centres = fuzzy_kmeans_cluster(f, 1, fuzzyq, 
											   delta_centres_threshold);
		min_valid=0.0;
		best_nclusters=1;
	}

	/* updated this so the thing would learn quicker */
	for (nclusters=2; (nclusters<=max_nclusters)&&
		              (nclusters<=list_length(f->elements)/4+1); nclusters++)
	{
		tmp_cluster_centres = fuzzy_kmeans_cluster(f, nclusters, fuzzyq,
											   delta_centres_threshold);

		valid = fuzzy_validity(f, tmp_cluster_centres, nclusters, fuzzyq);
		if (valid<min_valid)
		{
			if (cluster_centres) 
			{
				for (i=0; i<best_nclusters; i++)
					iprs_free(cluster_centres[i]);
				iprs_free(cluster_centres);
			}
			cluster_centres = tmp_cluster_centres;
			min_valid = valid;
			best_nclusters = nclusters;
		}
		else
		{
			for (i=0; i<nclusters; i++)
				iprs_free(tmp_cluster_centres[i]);
			iprs_free(tmp_cluster_centres);
		}
	}

	clist=NIL;
    nrules=best_nclusters;
    for (r=0;r<nrules;r++) {
        c=(CLUSTER *)iprs_calloc(1,sizeof(CLUSTER));
        list_append(c,clist);
		c->type=f->type;
		c->level=f->level;
		if (c->type=='U') nfeatures=nunary_features;
		else nfeatures=nbinary_features;
		s=(char *)iprs_calloc(strlen(f->parent->name)+10,sizeof(char));
		sprintf(s,"%s%c%-d",f->parent->name,f->type,r);
		c->name=s;
		c->parent=f->parent;
		c->children=NIL;
/* 		c->rule=(RULE *)iprs_calloc(1,sizeof(RULE)); */
/*         c->rule->nfeatures=nfeatures; */
/* 		c->rule->important=(int *)iprs_calloc(nfeatures,sizeof(int)); */
/*         c->rule->lower=(float*)iprs_calloc(nfeatures, sizeof(float)); */
/*         c->rule->upper=(float*)iprs_calloc(nfeatures, sizeof(float)); */
		c->mean = (float*)iprs_calloc(nfeatures, sizeof(float));
		for (i=0; i<nfeatures; i++)
			c->mean[i] = cluster_centres[r][i];
        c->elements=NIL;
		c->pathnodes=NIL;
		c->nclass=0;
		c->class=(int *)iprs_calloc(nclasses,sizeof(int));
		for (i=0;i<nclasses;i++) c->class[i]=0;
		c->cfs=(FEATURE_LIST *)NULL;
		c->unresolvable=0;

		if (c->level < max_level)
			c->cfs_possible= -1;
		else
			c->cfs_possible=0;

		if(f->type == 'U') {
			foreach(PATHNODE,pathnode,f->pathnodes) {
				int nearest_cluster;
				nearest_cluster = find_nearest_point(pathnode->unary->value, 
													 nfeatures, 
													 cluster_centres, 
													 best_nclusters);
				if (nearest_cluster == r)
				{
					list_append(pathnode,c->pathnodes);
					list_union(pathnode->unary,c->elements);
					(c->class[pathnode->unary->class])++;
				}
			}
			endforeach
		}
		else /* f->type == 'B' */ {
			foreach(PATHNODE,pathnode,f->pathnodes) {
				if (find_nearest_point(pathnode->binary->value, nfeatures,
									   cluster_centres, best_nclusters) == r)
				{
					list_append(pathnode,c->pathnodes);
					list_union(pathnode->binary,c->elements);
					(c->class[pathnode->binary->class])++;
				}
			}
			endforeach
		}
/* 		c->npathnodes=list_length(c->pathnodes); */
					
		/* nclass */

		c->nclass=0;
        for (i=0;i<nclasses;i++) {
            if (c->class[i]>0) (c->nclass)++;
        }
		c->entropy=compute_entropy(c->class);
/* 		pratio=(float)(c->npathnodes)/(float)(c->parent->cfs->npathnodes); */
/* 		c->weighted_entropy=pratio*c->entropy; */
	}

	/* insert new clusters into cluster tree */

	parent=f->parent;
	foreach(CLUSTER,newc,clist)
		list_append(newc,parent->children);
	endforeach

	/* print new clusters */

	if (crg_debug_clusters) {

		if (parent->level==0)
			printf("Expanding root cluster\n");
		else
			printf("\nExpanding cluster %s\n",parent->name);
	}

	iprs_free(cluster_centres);
	return(clist);
}

/* --------------------------------------------------------------------
 * Procedure: cluster_feature_list
 * Comments: cluster a feature list. List of clusters is returned.
 *           exactly the same as previous function except that the 
 *           clusters formed are not added to the parent's child list.
 * -------------------------------------------------------------------- */

LIST recluster_cluster(CLUSTER *clus, int nclusters, float fuzzyq,
					   float delta_centres_threshold)
{
	float **cluster_centres;
	CLUSTER *c,*parent;
	LIST clist;
	int nrules,i,r,nfeatures;
	float pratio;
	char *s;
	FEATURE_LIST *f;

	f = create_feature_list_for_cluster(clus);
	cluster_centres = fuzzy_kmeans_cluster(f, nclusters, fuzzyq,
										   delta_centres_threshold);

	clist=NIL;
    nrules=nclusters;
    for (r=0;r<nrules;r++) {
        c=(CLUSTER *)iprs_calloc(1,sizeof(CLUSTER));
        list_append(c,clist);
		c->type=f->type;
		c->level=f->level;
		if (c->type=='U') nfeatures=nunary_features;
		else nfeatures=nbinary_features;
		s=(char *)iprs_calloc(strlen(f->parent->name)+10,sizeof(char));
		sprintf(s,"%s%c%-d",f->parent->name,f->type,
				r+list_length(f->parent->children));
		c->name=s;
		c->parent=f->parent;
		c->children=NIL;
/* 		c->rule=(RULE *)iprs_calloc(1,sizeof(RULE)); */
/*         c->rule->nfeatures=nfeatures; */
/* 		c->rule->important=(int *)iprs_calloc(nfeatures,sizeof(int)); */
/*         c->rule->lower=(float*)iprs_calloc(nfeatures, sizeof(float)); */
/*         c->rule->upper=(float*)iprs_calloc(nfeatures, sizeof(float)); */
		c->mean = (float*)iprs_calloc(nfeatures, sizeof(float));
		for (i=0; i<nfeatures; i++)
			c->mean[i] = cluster_centres[r][i];
        c->elements=NIL;
		c->pathnodes=NIL;
		c->nclass=0;
		c->class=(int *)iprs_calloc(nclasses,sizeof(int));
		for (i=0;i<nclasses;i++) c->class[i]=0;
		c->cfs=(FEATURE_LIST *)NULL;
		c->unresolvable=0;

		if (c->level < max_level)
			c->cfs_possible= -1;
		else
			c->cfs_possible=0;

		if(f->type == 'U') {
			foreach(PATHNODE,pathnode,f->pathnodes) {
				int nearest_cluster;
				nearest_cluster = find_nearest_point(pathnode->unary->value, 
													 nfeatures, 
													 cluster_centres, 
													 nclusters);
				if (nearest_cluster == r)
				{
					list_append(pathnode,c->pathnodes);
					list_union(pathnode->unary,c->elements);
					(c->class[pathnode->unary->class])++;
				}
			}
			endforeach
		}
		else /* f->type == 'B' */ {
			foreach(PATHNODE,pathnode,f->pathnodes) {
				if (find_nearest_point(pathnode->binary->value, nfeatures,
									   cluster_centres, nclusters) == r)
				{
					list_append(pathnode,c->pathnodes);
					list_union(pathnode->binary,c->elements);
					(c->class[pathnode->binary->class])++;
				}
			}
			endforeach
		}
/* 		c->npathnodes=list_length(c->pathnodes); */
					
		/* nclass */

		c->nclass=0;
        for (i=0;i<nclasses;i++) {
            if (c->class[i]>0) (c->nclass)++;
        }
		c->entropy=compute_entropy(c->class);
/* 		pratio=(float)(c->npathnodes)/(float)(c->parent->cfs->npathnodes); */
/* 		c->weighted_entropy=pratio*c->entropy; */
	}

	/* print new clusters */

	if (crg_debug_clusters) {

		if (f->parent->level==0)
			printf("Reclustering root cluster\n");
		else
			printf("\nReclustering cluster %s\n",clus->name);
	}

	iprs_free(cluster_centres);
	free_feature_list(f);

	return(clist);
}

/* --------------------------------------------------------------------
 * Procedure: check_labels
 * Comments:  check that label information in data file is correct and
 *            meets the assumptions made in crg.
 * -------------------------------------------------------------------- */

void check_labels(float **label, int nlabel, int nbinary,
				  int npart, int class, int view)
{
	int i,j,found;

	/* number of binary and label must match */

	if (nlabel != nbinary)
		error_exit("read_database: label and binary do not match");

	/* label values must match parts in view */

	for (i=0;i<nlabel;i++) {
		if (label[i][0] < 0 || label[i][0] >= npart ||
			label[i][1] < 0 || label[i][1] >= npart) {
			printf("read_database: label information incorrect\n");
			printf("Class: %d View: %d Label { %d %d }\n",
				   class,view,(int)label[i][0],(int)label[i][1]);
			exit(1);
		}
	}

	/* inverse relations must be given */

	if (!symmetric_relations) {
		for (i=0;i<nlabel;i++) {
			found=0;
			for (j=0;j<nlabel;j++) {
				if (label[i][0]==label[j][1] && label[i][1]==label[j][0]) {
					found=1;
					break;
				}
			}
			if (!found)
				error_exit("read_database: inverse relation missing");
		}
	}
}

/* --------------------------------------------------------------------
 * Procedure: compute_mean_sdev
 * Comment:   calculate mean and standard deviation for unary and binary
 *            features
 * -------------------------------------------------------------------- */

void compute_mean_sdev()
{
	int i,n;

	/* unary */

	for (i=0;i<nunary_features;i++) {
		unary_mean[i]=0.0;
		unary_sdev[i]=0.0;
	}
	n=0;
	foreach(UNARY,unary,unary_list) {
		n++;
		for (i=0;i<nunary_features;i++) {
			unary_mean[i] += unary->value[i];
			unary_sdev[i] += (unary->value[i] * unary->value[i]);
		}
	}
	endforeach
	for (i=0;i<nunary_features;i++) {
		unary_mean[i] /= (float)n;
		unary_sdev[i] /= (float)n;
		unary_sdev[i] -= (unary_mean[i] * unary_mean[i]);
/* 		if ((unary_sdev[i] == 0.0)&&(zscore)) */
/* 			error_exit("compute_zscore: zero standard deviation"); */
		unary_sdev[i] = sqrtf(unary_sdev[i]);
	}

	/* binary */

	binary_mean=(float *)iprs_calloc(nbinary_features,sizeof(float));
	binary_sdev=(float *)iprs_calloc(nbinary_features,sizeof(float));
	for (i=0;i<nbinary_features;i++) {
		binary_mean[i]=0.0;
		binary_sdev[i]=0.0;
	}
	n=0;
	foreach(BINARY,binary,binary_list) {
		n++;
		for (i=0;i<nbinary_features;i++) {
			binary_mean[i] += binary->value[i];
			binary_sdev[i] += binary->value[i] * binary->value[i];
		}
	}
	endforeach
	for (i=0;i<nbinary_features;i++) {
		if (n>0)
		{
			binary_mean[i] /= (float)n;
			binary_sdev[i] /= (float)n;
			binary_sdev[i] -= binary_mean[i] * binary_mean[i];
			binary_sdev[i] = sqrtf(binary_sdev[i]);
		}
		else binary_mean[i] = binary_sdev[i] = 0.0;
	}
	if (crg_debug_varia)
	{
		for (i=0;i<nunary_features; i++)
		{
			printf("umean[%d]=%f, usdev[%d]=%f\n", i, unary_mean[i], i,
				   unary_sdev[i]);
		}
		for (i=0; i<nbinary_features; i++)
		{
			printf("bmean[%d]=%f, bsdev[%d]=%f\n", i, binary_mean[i], i,
				   binary_sdev[i]);
		}
	}
}

/* --------------------------------------------------------------------
 * Procedure: compute_zscore
 * Comment:   convert all (unary and binary) feature values to zvalues
 * -------------------------------------------------------------------- */

void compute_zscore()
{
	int i,n;

	/* unary */

	foreach(UNARY,unary,unary_list) {
		for (i=0;i<nunary_features;i++) {
			if (unary_sdev[i]!=0.0)
			{
				unary->value[i] -= unary_mean[i];
				unary->value[i] /= unary_sdev[i];
			}
		}
	}
	endforeach

	/* binary */

	foreach(BINARY,binary,binary_list) {
		for (i=0;i<nbinary_features;i++) {
			if (binary_sdev[i]!=0.0)
			{
				binary->value[i] -= binary_mean[i];
				binary->value[i] /= binary_sdev[i];
			}
		}
	}
	endforeach;
}

/* --------------------------------------------------------------------
 * Procedure: compute_entropy
 * Comments:
 * -------------------------------------------------------------------- */

float compute_entropy(int *class)
{
    int i,n;
    float p,e;

	n=0;
	for (i=0;i<nclasses;i++) {
		n+=class[i];
	}
	e=0.0;
    for (i=0;i<nclasses;i++) {
        if (class[i]>0.0) {
			p = (float)class[i]/(float)n;
            e -= p*logf(p);
        }
    }
    return e;
}

/* --------------------------------------------------------------------
 * Procedure: compute_range
 * Comment:   compute range of all unary and binary features. Range
 *            information is used in splitting clusters.
 *            feature_difference_threshold defines minimum difference
 *            in feature value for splitting.
 * -------------------------------------------------------------------- */

void compute_range(float feature_difference_threshold)
{
	int f;
	UNARY *unary;
	BINARY *binary;

	unary_range=(float *)iprs_calloc(nunary_features,sizeof(float));
	unary_difference_threshold=
		(float *)iprs_calloc(nunary_features,sizeof(float));
	unary_lower=(float *)iprs_calloc(nunary_features,sizeof(float));
	unary_upper=(float *)iprs_calloc(nunary_features,sizeof(float));
	unary=list_first(unary_list);
	for (f=0;f<nunary_features;f++) {
		unary_lower[f]=unary->value[f];
		unary_upper[f]=unary->value[f];
	}
	foreach(UNARY,unary,list_rest(unary_list)) {
		for (f=0;f<nunary_features;f++) {
			if (unary->value[f] < unary_lower[f])
				unary_lower[f]=unary->value[f];
			if (unary->value[f] > unary_upper[f])
				unary_upper[f]=unary->value[f];
		}
	}
	endforeach
	for (f=0;f<nunary_features;f++) {
		unary_range[f]=unary_upper[f]-unary_lower[f];
		/*
		 * Changed by B. McCane because if there are outliers, the
		 * difference threshold will be larger than expected.
		 * Also, if zscore, variance=1, hence make diff_thresh equal to
		 * the global difference thresh.
		 */
		if (zscore) 
			unary_difference_threshold[f]=feature_difference_threshold;
		else
			unary_difference_threshold[f]=
				unary_sdev[f]*feature_difference_threshold;
	}

	if (crg_debug_varia) {
		printf("Range of unary feature values:\n");
		for (f=0;f<nunary_features;f++) {
			printf("%2d: %10f - %10f = %12f, diff_thresh = %f\n",
				   f,unary_lower[f],unary_upper[f],unary_range[f],
				   unary_difference_threshold[f]);
		}
	}

	binary_range=(float *)iprs_calloc(nbinary_features,sizeof(float));
	binary_difference_threshold=
		(float *)iprs_calloc(nbinary_features,sizeof(float));
	binary_lower=(float *)iprs_calloc(nbinary_features,sizeof(float));
	binary_upper=(float *)iprs_calloc(nbinary_features,sizeof(float));
	for (f=0;f<nbinary_features;f++) {
		binary_lower[f]=FLT_MAX/2.0;
		binary_upper[f]=-FLT_MAX/2.0;
	}
	foreach(BINARY,binary,binary_list) {
		for (f=0;f<nbinary_features;f++) {
			if (binary->value[f] < binary_lower[f])
				binary_lower[f]=binary->value[f];
			if (binary->value[f] > binary_upper[f])
				binary_upper[f]=binary->value[f];
		}
	}
	endforeach
	for (f=0;f<nbinary_features;f++) {
		binary_range[f]=binary_upper[f]-binary_lower[f];
		if (zscore) 
			binary_difference_threshold[f]=feature_difference_threshold;
		else
			binary_difference_threshold[f]=
				binary_sdev[f]*feature_difference_threshold;
	}

	if (crg_debug_varia) {
		printf("Range of binary feature values:\n");
		for (f=0;f<nbinary_features;f++) {
			printf("%2d: %10f - %10f = %12f; diff_thresh = %f\n",
				   f,binary_lower[f],binary_upper[f],binary_range[f],
				   binary_difference_threshold[f]);
		}
	}
}

/* --------------------------------------------------------------------
 * Procedure: feature_in_rule
 * Comments: does feature value fall within the rule hyper-rectangle?
 * -------------------------------------------------------------------- */

int feature_in_rule(float *value, RULE *rule)
{
	int i;
	for (i=0;i<rule->nfeatures;i++) {
		if (value[i] < rule->lower[i]) return(0);
		if (value[i] > rule->upper[i]) return(0);
	}
	return(1);
}
/* --------------------------------------------------------------------
 * Procedure: add_clist_to_queue
 * Comment:   just a little shorthand
 * -------------------------------------------------------------------- */

void add_clist_to_queue(LIST clist, LIST queue)
{
	if (breadth_first) {
		foreach(CLUSTER,c,clist)
			list_append(c,queue);
		endforeach
	}
	else /* depth first */ {
		foreach(CLUSTER,c,clist)
			list_prepend(c,queue);
		endforeach
	}
}

/* --------------------------------------------------------------------
 * Procedure: save_training
 * Comment:   save training data to file
 * -------------------------------------------------------------------- */

void save_training(CLUSTER *root, char *infilename, char *outfilename)
{
	int i,id;
	CLUSTER *c;
	PATH *path;
	LIST clist;
	FILE *fp;

	if ((fp=fopen(outfilename,"w")) == (FILE *)NULL)
		error_exit("save_training: cannot open file");

	/* save some global values */

	fputs(infilename,fp);
	putc('\n',fp);
	fwrite(&nunary_features,sizeof(int),1,fp);
	fwrite(&nbinary_features,sizeof(int),1,fp);
	fwrite(&nclasses,sizeof(int),1,fp);
	fwrite(&nviews,sizeof(int),1,fp);
	fwrite(&nparts,sizeof(int),1,fp);
	fwrite(&feature_ordering,sizeof(int),1,fp);
	fwrite(&symmetric_relations,sizeof(int),1,fp);
	fwrite(&max_level,sizeof(int),1,fp);
	/* mod by B. McCane: save zscore transformations params */
	/*                   so test image can be similarly transfmed */
	if (zscore)
	{
		printf("zscore\n");
		fwrite(unary_mean, sizeof(float), nunary_features, fp);
		fwrite(unary_sdev, sizeof(float), nunary_features, fp);
		fwrite(binary_mean, sizeof(float), nbinary_features, fp);
		fwrite(binary_sdev, sizeof(float), nbinary_features, fp);
	}
	/* if we're not performing a zscore transform, then we need to */
	/* save the mean as 0 and the sdev as 1 for unary and binary.  */
	/* This has the effect of not transforming data at all.        */
	else
	{
		int i;
		float *mean, *sdev;

		printf("not zscore\n");
		/* unary */
		mean = (float*)iprs_calloc(nunary_features, sizeof(float));
		sdev = (float*)iprs_calloc(nunary_features, sizeof(float));
		for (i=0; i<nunary_features; i++)
		{
			mean[i] = 0.0;
			sdev[i] = 1.0;
		}
		fwrite(mean, sizeof(float), nunary_features, fp);
		fwrite(sdev, sizeof(float), nunary_features, fp);
		iprs_free(mean); iprs_free(sdev);

		/* binary */
		mean = (float*)iprs_calloc(nbinary_features, sizeof(float));
		sdev = (float*)iprs_calloc(nbinary_features, sizeof(float));
		for (i=0; i<nbinary_features; i++)
		{
			mean[i] = 0.0;
			sdev[i] = 1.0;
		}
		fwrite(mean, sizeof(float), nbinary_features, fp);
		fwrite(sdev, sizeof(float), nbinary_features, fp);
		iprs_free(mean); iprs_free(sdev);
	}

	/* save part descriptions */

	fwrite(&nparts,sizeof(int),1,fp);
	foreach(PART,part,part_list) {
		int numbinary;
		fwrite(&(part->name),sizeof(int),1,fp);
		fwrite(&(part->view),sizeof(int),1,fp);
		fwrite(&(part->class),sizeof(int),1,fp);
		numbinary = list_length(part->binary);
		fwrite(&numbinary, sizeof(int), 1, fp);
	}
	endforeach

	/* generate numerical cluster ids and count number of nodes      */

	id=0;
	clist=list_list(root);
	while (!list_empty(clist)) {
		c=list_remove_first(clist);
		c->id=id++;
		foreach(CLUSTER,child,c->children)
			list_append(child,clist);
		endforeach
	}
	fwrite(&id,sizeof(int),1,fp);

	/* save clusters in breadth first order */

	clist=list_list(root);
	while (!list_empty(clist)) {
		c=list_remove_first(clist);

		fwrite(&(c->id),sizeof(int),1,fp);
		fwrite(&(c->level),sizeof(int),1,fp);
		fputs(c->name,fp);
		putc('\n',fp);
		fwrite(&(c->cfs_possible),sizeof(int),1,fp);

		if (c->level > 0) {

			/* parent */

			fwrite(&(c->parent->id),sizeof(int),1,fp);

			/* paths */

			i=list_length(c->pathnodes);
			fwrite(&i,sizeof(int),1,fp);
			foreach(PATHNODE,pathnode,c->pathnodes) {
				path=get_path(pathnode);
				fwrite(&(path->length),sizeof(int),1,fp);
				for (i=0;i<path->length;i++)
					fwrite(&(path->part[i]->name),sizeof(int),1,fp);
				free_path(path);
			}
			endforeach

			/* rule */

/* 			fwrite(&(c->rule->nfeatures),sizeof(int),1,fp); */
/* 			fwrite(c->rule->important,sizeof(int),c->rule->nfeatures,fp); */
/* 			fwrite(c->rule->lower,sizeof(float),c->rule->nfeatures,fp); */
/* 			fwrite(c->rule->upper,sizeof(float),c->rule->nfeatures,fp); */

			/* class vector */

			fwrite(c->class,sizeof(int),nclasses,fp);

			/* cluster mean */

			if (c->type=='U')
				fwrite(c->mean, sizeof(float), nunary_features, fp);
			else
				fwrite(c->mean, sizeof(float), nbinary_features, fp);
		}

		/* Add children to printout queue */

		if (!list_empty(c->children)) {
			foreach(CLUSTER,child,c->children)
				list_append(child,clist);
			endforeach
		}

	}

	/* close */

	fclose(fp);
}
