/* ====================================================================
 * Package:		analyze
 * Comments:    analyze cluster tree
 * ==================================================================== */

/* Include files */

#include <iprs.h>
#include "crg.h"

/* External Declarations */

extern LIST      binary_list;
extern float    *binary_difference_threshold;
extern float    *binary_mean;
extern float    *binary_range;
extern float    *binary_sdev;
extern float     expected_rule_cost;
extern int       max_path_length;
extern int       nbinary_features;
extern int       nclasses;
extern int       nnodes_analyzed;
extern int       nparts;
extern int       npaths;
extern int       npaths_analyzed;
extern int       nunary_features;
extern int       nviews;
extern LIST      part_list;
extern PATHNODE *path_tree;
extern LIST      unary_list;
extern float    *unary_difference_threshold;
extern float    *unary_mean;
extern float    *unary_range;
extern float    *unary_sdev;

/* Parameters */

extern int       breadth_first;
extern int       crg_debug_clusters;
extern int       crg_debug_parts;
extern int       crg_debug_paths;
extern int       crg_debug_varia;
extern int       entropy_wrt_all_classes;
extern float     entropy_threshold;
extern int       feature_ordering;
extern int       fixed_path_length;
extern float     leader_threshold;
extern int       max_level;
extern int       symmetric_relations;
extern int       use_leader_cluster;
extern int       zscore;

/* Global Declarations */

static int nunresolved_clusters;
static int *unresolved_part;
static int *unresolved_view;
static int *unresolved_class;
static float unary_volume;
static float binary_volume;

/* --------------------------------------------------------------------
 * Procedure: print_cluster_tree
 * Comments:  print cluster tree with class vectors and rules
 * -------------------------------------------------------------------- */

void print_cluster_tree(CLUSTER *c, LIST leaf_list, int *leaf_number)
{
    int i,npathnodes;
	float h,w;

    if (c->level>0) {
        for (i=1;i<c->level;i++) printf("  ");
        printf("%s [",c->name);
		npathnodes=list_length(c->pathnodes);
		for (i=0;i<nclasses;i++) {
			if (c->class[i] > 0)
				printf(" %4.2f",(float)c->class[i]/(float)npathnodes);
			else
				printf(" --");
		}
		printf("]");

		if (c->cfs_possible==0)
			printf("(-)");
		else if (c->cfs_possible==1)
			printf("(+)");

		/* weighted entropy */

		w=(float)npathnodes/(float)(list_length(c->parent->cfs->pathnodes));
		h=w*c->entropy;

        if (list_empty(c->children)) {
			list_append(c,leaf_list);
			(*leaf_number)++;
			printf(" H=%4.2f wH=%5.3f (%d paths) Rule%d",
				   c->entropy,h,list_length(c->pathnodes),*leaf_number);
			if (c->entropy > entropy_threshold) {
				printf(" (unresolved)\n");
			}
			else {
        		printf("\n");
			}
		}
		else {
			printf("] H=%4.2f wH=%5.3f (%d paths)\n",
				   c->entropy,h,list_length(c->pathnodes));
		}
	}

    foreach(CLUSTER,child,c->children) {
        print_cluster_tree(child,leaf_list,leaf_number);
    }
    endforeach
}

/* --------------------------------------------------------------------
 * Procedure: print_rules
 * Comments:  
 * -------------------------------------------------------------------- */

/* void print_rules(LIST leaf_list) */
/* { */
/* 	int leaf_number=0,f; */
/* 	CLUSTER *c; */
/* 	LIST rules; */
/* 	float fl,fu; */
/* 	char type; */
/* 	int i; */

/* 	foreach(CLUSTER,leaf,leaf_list) { */
/* 		leaf_number++; */
/* 		printf("Rule %s %d:", leaf->name, leaf_number); */
/* 		if (leaf->type=='U') */
/* 		{ */
/* 			for (i=0; i<nunary_features; i++) */
/* 				printf(" %f", leaf->mean[i]); */
/* 		} */
/* 		else */
/* 		{ */
/* 			for (i=0; i<nbinary_features; i++) */
/* 				printf(" %5.2f", leaf->mean[i]); */
/* 		} */
/* 		printf("\n"); */
/* 		rules=NIL; */
/* 		for (c=leaf;c->level>0;c=c->parent) */
/* 			list_prepend(c->rule,rules); */
/* 		type='B'; */
/* 		foreach(RULE,rule,rules) { */
/* 			type = type=='U' ? 'B' : 'U'; */
/* 			printf("      %c ",type); */
/* 			if (zscore) { */
/* 				if (type == 'U') { */
/* 					for (f=0;f<rule->nfeatures;f++) { */
/* 						if (f>0 && f%4==0) printf("\n        "); */
/* 						fl=rule->lower[f]*unary_sdev[f]+unary_mean[f]; */
/* 						fu=rule->upper[f]*unary_sdev[f]+unary_mean[f]; */
/* 						fl = rule->lower[f]; */
/* 						fu = rule->upper[f]; */
/* 						if (rule->important[f]) */
/* 							printf("[%5.2f:%5.2f] ",fl,fu); */
/* 						else */
/* 							printf("{%5.2f:%5.2f} ",fl,fu); */
/* 					} */
/* 				} */
/* 				else  { */
/* 					for (f=0;f<rule->nfeatures;f++) { */
/* 						if (f>0 && f%4==0) printf("\n        "); */
/* 						fl=rule->lower[f]*binary_sdev[f]+binary_mean[f]; */
/* 						fu=rule->upper[f]*binary_sdev[f]+binary_mean[f]; */
/* 						fl = rule->lower[f]; */
/* 						fu = rule->upper[f]; */
/* 						if (rule->important[f]) */
/* 							printf("[%5.2f:%5.2f] ",fl,fu); */
/* 						else */
/* 							printf("{%5.2f:%5.2f} ",fl,fu); */
/* 					} */
/* 				} */
/* 			} */
/* 			else { */
/* 				for (f=0;f<rule->nfeatures;f++) { */
/* 					if (f>0 && f%4==0) printf("\n        "); */
/* 					if (rule->important[f]) */
/* 						printf("[%5.2f:%5.2f] ",rule->lower[f],rule->upper[f]); */
/* 					else */
/* 						printf("{%5.2f:%5.2f} ",rule->lower[f],rule->upper[f]); */
/* 				} */
/* 			} */
/* 			printf("\n"); */
/* 		} */
/* 		endforeach */
/* 		list_free(rules); */
/* 	} */
/* 	endforeach */
/* } */

/* --------------------------------------------------------------------
 * Procedure: analyze_rules
 * Comments:  
 * -------------------------------------------------------------------- */

/* void analyze_rules(CLUSTER *c,float cvol,int *nrule,int *ncase,float *volume) */
/* { */
/* 	RULE *r; */
/* 	PART *part; */
/* 	float v; */
/* 	int f; */

/* 	if (c->level > 0) { */
/* 		r=c->rule; */
/* 		v=1.0; */
/* 		for (f=0;f<r->nfeatures;f++) */
/* 			v *= (r->upper[f]-r->lower[f]); */
/* 		if ((c->type == 'U')&&(unary_volume>0.0)) */
/* 			v /= unary_volume; */
/* 		else if (binary_volume>0.0) */
/* 			v /= binary_volume; */
/* 		cvol += v; */
/* 	} */

/* 	if (list_empty(c->children)) { */
/* 		if (c->entropy > entropy_threshold) { */
/* 			nunresolved_clusters++; */
/* 			foreach(PATHNODE,pathnode,c->pathnodes) { */
/* 				part=pathnode->part; */
/* 				unresolved_part [part->name ]=1; */
/* 				unresolved_view [part->view ]=1; */
/* 				unresolved_class[part->class]=1; */
/* 			} */
/* 			endforeach */
/* 		} */
/* 		else { */
/* 			nrule[c->level]++; */
/* 			ncase[c->level] += list_length(c->pathnodes); */
/* 			volume[c->level]+=cvol; */
/* 		} */
/* 	} */
/* 	else { */
/* 		foreach(CLUSTER,child,c->children) */
/* 			analyze_rules(child,cvol,nrule,ncase,volume); */
/* 		endforeach */
/* 	} */
/* } */

/* --------------------------------------------------------------------
 * Procedure: compute_unary_volume
 * Comments:  
 * -------------------------------------------------------------------- */

void compute_unary_volume()
{
	float *lower,*upper;
	int f;

	lower=(float *)iprs_calloc(nunary_features,sizeof(float));
	upper=(float *)iprs_calloc(nunary_features,sizeof(float));

	for (f=0;f<nunary_features;f++) {
		lower[f] =  IPRS_FLOAT_MAX;
		upper[f] = -IPRS_FLOAT_MAX;
	}
	foreach(UNARY,unary,unary_list) {
		for (f=0;f<nunary_features;f++) {
			if (unary->value[f] < lower[f]) lower[f]=unary->value[f];
			if (unary->value[f] > upper[f]) upper[f]=unary->value[f];
		}
	}
	endforeach
	unary_volume=1.0;
	for (f=0;f<nunary_features;f++) {
		unary_volume *= (upper[f]-lower[f]);
	}

	iprs_free(lower);
	iprs_free(upper);
}

/* --------------------------------------------------------------------
 * Procedure: compute_binary_volume
 * Comments:  
 * -------------------------------------------------------------------- */

void compute_binary_volume()
{
	float *lower,*upper;
	int f;

	lower=(float *)iprs_calloc(nbinary_features,sizeof(float));
	upper=(float *)iprs_calloc(nbinary_features,sizeof(float));

	for (f=0;f<nbinary_features;f++) {
		lower[f] =  IPRS_FLOAT_MAX;
		upper[f] = -IPRS_FLOAT_MAX;
	}
	foreach(BINARY,binary,binary_list) {
		for (f=0;f<nbinary_features;f++) {
			if (binary->value[f] < lower[f]) lower[f]=binary->value[f];
			if (binary->value[f] > upper[f]) upper[f]=binary->value[f];
		}
	}
	endforeach
	binary_volume=1.0;
	for (f=0;f<nbinary_features;f++) {
		binary_volume *= (upper[f]-lower[f]);
	}

	iprs_free(lower);
	iprs_free(upper);
}

/* --------------------------------------------------------------------
 * Procedure: mark_important_features
 * Comments:  determine features used for distinguishing different 
 *            clusters.
 * -------------------------------------------------------------------- */

/* void mark_important_features(CLUSTER *c) */
/* { */
/* 	int i,nf,*mark; */
/* 	float *lower,*upper; */
/* 	CLUSTER *child; */

/* 	if (list_empty(c->children)) return; */

/* 	if (c->type=='U') */
/* 		nf=nbinary_features; */
/* 	else */
/* 		nf=nunary_features; */

/* 	mark=   (int *)iprs_calloc(nf,sizeof(int)); */
/* 	lower=(float *)iprs_calloc(nf,sizeof(float)); */
/* 	upper=(float *)iprs_calloc(nf,sizeof(float)); */
/* 	child=list_first(c->children); */

/* 	for (i=0;i<nf;i++) { */
/* 		mark[i]=0; */
/* 		lower[i]=child->rule->lower[i]; */
/* 		upper[i]=child->rule->upper[i]; */
/* 	} */

/* 	foreach(CLUSTER,child,list_rest(c->children)) { */
/* 		for (i=0;i<nf;i++) { */
/* 			if ((lower[i] != child->rule->lower[i]) || */
/* 				(upper[i] != child->rule->upper[i])) */
/* 				mark[i]=1; */
/* 		} */
/* 	} */
/* 	endforeach */

/* 	foreach(CLUSTER,child,c->children) { */
/* 		for (i=0;i<nf;i++) { */
/* 			child->rule->important[i]=mark[i]; */
/* 		} */
/* 	} */
/* 	endforeach */

/* 	iprs_free(mark); */
/* 	iprs_free(lower); */
/* 	iprs_free(upper); */

/* 	foreach(CLUSTER,child,c->children) */
/* 		mark_important_features(child); */
/* 	endforeach */

/* } */

/* --------------------------------------------------------------------
 * Procedure: print_resolved
 * Comments:  
 * -------------------------------------------------------------------- */

void print_unresolved()
{
	int nunresolved_parts;
	int i,j;

	if (nunresolved_clusters == 0) {
		if (entropy_threshold == 0.0)
			printf("All rules are unique\n");
		else
			printf("All rules have entropy <= %4.2f\n",entropy_threshold);
		return;
	}

	nunresolved_parts=0;
	for (i=0;i<nparts;i++) {
		if (unresolved_part[i]) nunresolved_parts++;
	}

	if (nunresolved_clusters == 1){
		if (entropy_threshold == 0.0)
			printf("1 rule containing %d parts is non-unique\n",
		   		   nunresolved_parts);
		else
			printf("1 rule containing %d parts has entropy > %4.2f\n",
				   nunresolved_parts,entropy_threshold);
	}
	else {
		if (entropy_threshold == 0.0)
			printf("%d rules containing %d parts are non-unique\n",
		   		   nunresolved_clusters,nunresolved_parts);
		else
			printf("%d rules containing %d parts have entropy > %4.2f\n",
				   nunresolved_clusters,nunresolved_parts,entropy_threshold);
	}

#ifdef 0

	printf("Parts:\n");
	j=0;
	for (i=0;i<nparts;i++) {
		if (unresolved_part[i]) {
			printf("%3d ",i);
			j++;
			if (j%20 == 0) printf("\n         ");
		}
	}
	printf("\n");

#endif

	printf("Views:\n");
	j=0;
	for (i=0;i<nviews;i++) {
		if (unresolved_view[i]) {
			printf("%3d ",i);
			j++;
			if (j%20 == 0) printf("\n         ");
		}
	}
	printf("\n");

	printf("Classes: ");
	j=0;
	for (i=0;i<nclasses;i++) {
		if (unresolved_class[i]) {
			printf("%3d ",i);
			j++;
			if (j%20 == 0) printf("\n         ");
		}
	}
	printf("\n");
}

/* --------------------------------------------------------------------
 * Procedure: print_rule_summary
 * Comments:  print percent correct classification up to level n and
 *            statistics of rules at that level n
 * -------------------------------------------------------------------- */

void print_rule_summary(int *nrule, int *ncase, float *volume)
{
	int tnrule=0,tncase=0,level,i,max_print_level;
	float tvolume=0.0,r;

	for (level=1;level<=max_level;level++) {
		if (nrule[level]>0) max_print_level=level;
	}

	if (entropy_threshold == 0.0)
		printf("\nAnalysis of unique rules\n\n");
	else
		printf("\nAnalysis of rules with entropy <= %4.2f\n\n",
			   entropy_threshold);
	printf("Level      p(class) rules cases/rule  volume  vol/rule\n");
	for (level=1;level<=max_print_level;level++) {
		for (i=1;i<=level;i++)
			printf("%c", i%2 == 0 ? 'B' : 'U' );
		for (i=level+1;i<=10;i++)
			printf(" ");
		if (nrule[level]>0) {
			tnrule+=nrule[level];
			tncase+=ncase[level];
			tvolume+=volume[level];
		}
		r=tncase/(float)npaths_analyzed;
		printf("%9.3f",r);
		if (nrule[level]>0) {
			r=(float)ncase[level]/(float)nrule[level];
			printf("%6d %9.2f",nrule[level],r);
			r=volume[level]/(float)nrule[level];
			printf("%9.2f %9.2f\n",volume[level],r);
		}
		else {
			printf("\n");
		}
	}
	r=(float)tncase/(float)tnrule;
	printf("Total              %6d %9.2f",tnrule,r);
	r=tvolume/(float)tnrule;
	printf("%9.2f %9.2f\n",tvolume,r);
}

/* --------------------------------------------------------------------
 * Procedure: print_summary
 * Comments:  
 * -------------------------------------------------------------------- */

void print_summary()
{
	printf("\nSummary\n-------\n");
	printf("Number of classes:         %d\n",nclasses);
	printf("Number of views:           %d\n",nviews);
	printf("Number of parts:           %d\n",nparts);
	printf("Number of paths:           %d\n",npaths);
	printf("Number of paths analyzed:  %d\n",npaths_analyzed);
	printf("Number of nodes analyzed:  %d\n",nnodes_analyzed);
	printf("Expected rule cost:        %-5.1f\n",expected_rule_cost);
}

/* --------------------------------------------------------------------
 * Procedure: analyze_cluster_tree
 * Comments:  
 * -------------------------------------------------------------------- */

void analyze_cluster_tree(CLUSTER *root)
{
	LIST leaf_list;
	int leaf_number;

	int *nrule,*ncase,i;
	float *volume;

/* 	mark_important_features(root); */

	/* print cluster tree and rules */

	leaf_list=NIL;
	leaf_number=0;
	printf("Cluster tree\n------------\n");
	print_cluster_tree(root,leaf_list,&leaf_number);
	printf("\nRules\n-----\n");
/* 	print_rules(leaf_list); */
	printf("\n");

	/* analyze cluster tree for rule volume, unresolved elements etc. */

	unresolved_part=(int *)iprs_calloc(nparts,sizeof(int));
	for (i=0;i<nparts;i++) unresolved_part[i]=0;
	unresolved_view=(int *)iprs_calloc(nviews,sizeof(int));
	for (i=0;i<nviews;i++) unresolved_view[i]=0;
	unresolved_class=(int *)iprs_calloc(nclasses,sizeof(int));
	for (i=0;i<nclasses;i++) unresolved_class[i]=0;
	nunresolved_clusters=0;

	nrule=(int *)iprs_calloc(max_level+1,sizeof(int));
	ncase=(int *)iprs_calloc(max_level+1,sizeof(int));
	volume=(float *)iprs_calloc(max_level+1,sizeof(float));
	for (i=1;i<=max_level;i++) {
		nrule[i]=0;
		ncase[i]=0;
		volume[i]=0.0;
	}

	compute_unary_volume();
	compute_binary_volume();
/* 	analyze_rules(root,0.0,nrule,ncase,volume); */

	print_unresolved();

	print_rule_summary(nrule,ncase,volume);

	print_summary();
}
