/* ====================================================================
 * Package:     cluster
 * ==================================================================== */

/* Include files */

#include <iprs.h>
#include "crg.h"

/* Definitions */

#define DZERO ((float)1.0e-10)

/* External Declarations */

extern LIST      binary_list;
extern float    *binary_lower;
extern float    *binary_difference_threshold;
extern float    *binary_mean;
extern float    *binary_range;
extern float    *binary_sdev;
extern float    *binary_upper;
extern float     expected_rule_cost;
extern int       full_feature_range;
extern int       max_path_length;
extern int       nbinary_features;
extern int       nclasses;
extern int       nnodes_analyzed;
extern int       nparts;
extern int       npaths;
extern int       npaths_analyzed;
extern int       nunary_features;
extern int       nviews;
extern LIST      part_list;
extern PATHNODE *path_tree;
extern LIST      unary_list;
extern float    *unary_lower;
extern float    *unary_difference_threshold;
extern float    *unary_mean;
extern float    *unary_range;
extern float    *unary_sdev;
extern float    *unary_upper;

/* Parameters */

extern int       breadth_first;
extern int       crg_debug_clusters;
extern int       crg_debug_parts;
extern int       crg_debug_paths;
extern int       crg_debug_varia;
extern float     entropy_threshold;
extern int       entropy_wrt_all_classes;
extern int       feature_ordering;
extern int       fixed_path_length;
extern float     leader_threshold;
extern int       max_level;
extern int       symmetric_relations;
extern int       use_leader_cluster;
extern int       zscore;

/* -----------------------------------------------------------------*/
/* dummy procedures for stuff I'm not using */

IPRS_HRECT_RULE *iprs_genhrectrule_objectspace(IPRS_IMAGE *a)
{
        return(NULL);
}

void iprs_leadercluster_objectspace(IPRS_IMAGE *a, float thresh){};

IPRS_HRECT_RULE *iprs_create_hrectrule(int nf, int nd)
{
        return(NULL);
}

void iprs_free_labelspace(IPRS_LABEL_SPACE *a){};
/* ------------------------------------------------------------------*/

/* --------------------------------------------------------------------
 * Procedure: single_cluster
 * Comments:  returns a single cluster
 * -------------------------------------------------------------------- */

IPRS_HRECT_RULE *single_cluster(FEATURE_LIST *f)
{
	int nfeatures,i;
	float margin=0.02;
	IPRS_HRECT_RULE *hrrptr;

	if (f->type=='U')
		nfeatures = nunary_features;
	else
		nfeatures = nbinary_features;

	hrrptr=iprs_create_hrectrule(nfeatures,1);
	for (i=0;i<nfeatures;i++) {
		hrrptr->lower[0][i] =  IPRS_FLOAT_MAX;
		hrrptr->upper[0][i] = -IPRS_FLOAT_MAX;
	}

	if (f->type=='U') {

		if (full_feature_range) {
			for (i=0;i<nfeatures;i++) {
				hrrptr->lower[0][i] = unary_lower[i] - unary_range[i];
				hrrptr->upper[0][i] = unary_upper[i] + unary_range[i];
			}
		}
		else {

			/* find bounding rectangle of data points */

			foreach(UNARY,unary,f->elements) {
				for (i=0;i<nfeatures;i++) {
					if (unary->value[i] < hrrptr->lower[0][i])
						hrrptr->lower[0][i] = unary->value[i];
					if (unary->value[i] > hrrptr->upper[0][i])
						hrrptr->upper[0][i] = unary->value[i];
				}
			}
			endforeach

			/* add margin (so no data point is on boundary) */

			for (i=0;i<nfeatures;i++) {
/* 				hrrptr->lower[0][i] -= margin*unary_range[i]; */
/* 				hrrptr->upper[0][i] += margin*unary_range[i]; */
				hrrptr->lower[0][i] -= unary_difference_threshold[i];
				hrrptr->upper[0][i] += unary_difference_threshold[i];
			}
		}
	}
	else /* f->type == 'B' */ {

		if (full_feature_range) {
			for (i=0;i<nfeatures;i++) {
				hrrptr->lower[0][i] = binary_lower[i] - binary_range[i];
				hrrptr->upper[0][i] = binary_upper[i] + binary_range[i];
			}
		}
		else {

			/* find bounding rectangle of data points */

			foreach(BINARY,binary,f->elements) {
				for (i=0;i<nfeatures;i++) {
					if (binary->value[i] < hrrptr->lower[0][i])
						hrrptr->lower[0][i] = binary->value[i];
					if (binary->value[i] > hrrptr->upper[0][i])
						hrrptr->upper[0][i] = binary->value[i];
				}
			}
			endforeach

			/* add margin (so no data point is on boundary) */

			for (i=0;i<nfeatures;i++) {
/* 				hrrptr->lower[0][i] -= margin*binary_range[i]; */
/* 				hrrptr->upper[0][i] += margin*binary_range[i]; */
				hrrptr->lower[0][i] -= binary_difference_threshold[i];
				hrrptr->upper[0][i] += binary_difference_threshold[i];
			}
		}
	}

	return hrrptr;
}


/* --------------------------------------------------------------------
 * Procedure: leader_cluster
 * Comments:  cluster the elements of feature list f using leader cluster.
 *            Simply set up everything so we can call the iprs routines
 *            for clustering and rule generation. We might want to call
 *            min entropy clustering from here.
 * -------------------------------------------------------------------- */

IPRS_HRECT_RULE *leader_cluster(FEATURE_LIST *f, float threshold)
{
	IPRS_IMAGE *osptr;
	IPRS_FEATURE_SPACE *fsptr;
	IPRS_HRECT_RULE *hrrptr;
	UNARY *unary;
	BINARY *binary;
	int npoints,nfeatures,i,j;

	/* printf("Fix margin of leader-cluster as in single-cluster\n"); */

	npoints=list_length(f->elements);
	if (f->type == 'U')
		nfeatures=nunary_features;
	else
		nfeatures=nbinary_features;

	/* VERY few elements */

	if (npoints<1) error_exit("leader cluster: no elements");
	if (npoints==1) {
		hrrptr=iprs_create_hrectrule(nfeatures,1);
		if (f->type == 'U') {
			unary=list_first(f->elements);
			for (i=0;i<nfeatures;i++) {
				hrrptr->lower[0][i]=unary->value[i]-
					unary_difference_threshold[i];
				/* Modification: B. McCane */
				/* added diff_thresh so as not to create */
				/* clusters of size 0 */
				hrrptr->upper[0][i]=unary->value[i]+
					unary_difference_threshold[i];
			}
		}
		else /* f->type = 'B' */ {
			binary=list_first(f->elements);
			for (i=0;i<nfeatures;i++) {
				hrrptr->lower[0][i]=binary->value[i]-
					binary_difference_threshold[i];
				/* Modification: B. McCane */
				/* added diff_thresh so as not to create */
				/* clusters of size 0 */
				hrrptr->upper[0][i]=binary->value[i]+
					binary_difference_threshold[i];
			}
		}
		return(hrrptr);
	}

	/* Prepare everything for iprs routine */

	osptr=iprs_create_image(3);
    fsptr=crg_create_featurespace(nfeatures,npoints);
    osptr->image[1]=fsptr;
	if (f->type == 'U') {
		i=0;
		foreach(UNARY,unary,f->elements) {
			fsptr->point[i++]=unary->value;
		}
		endforeach
	}
	else /* f->type == 'B' */ {
		i=0;
		foreach(BINARY,binary,f->elements) {
			fsptr->point[i++]=binary->value;
		}
		endforeach
	}
	osptr->type[0] = IPRS_SPACETYPE_LABEL;
	osptr->type[1] = IPRS_SPACETYPE_FEATURE;
	osptr->type[2] = IPRS_SPACETYPE_LABEL;
	iprs_leadercluster_objectspace(osptr,threshold);
	hrrptr=iprs_genhrectrule_objectspace(osptr);
	/* make rules greater than a single pt if necessary */
	for (i=0; i<hrrptr->numrules; i++)
	{
		for (j=0; j<hrrptr->numfeatures; j++)
		{
			if (f->type=='U')  /* unary */
			{
				hrrptr->upper[i][j] += unary_difference_threshold[j];
				hrrptr->lower[i][j] -= unary_difference_threshold[j];
			}
			else    /* binary */
			{
				hrrptr->upper[i][j] += binary_difference_threshold[j];
				hrrptr->lower[i][j] -= binary_difference_threshold[j];
			}
		}
	}
	

	crg_free_featurespace(osptr->image[1]);
	iprs_free_labelspace(osptr->image[2]);

	return(hrrptr);
}

/* --------------------------------------------------------------------
 * Procedure: agglomerate_cluster
 * Comments:  cluster elements into at most max_ncluster clusters. The
 *            number of clusters produced can be smaller than max_nclusters
 *            if a) some elements have zero distance or b) some elements
 *            have the same distance.
 * Note:      WARNING: can't be used with large number of elements. We'd
 *            have to get rid of the distance MATRIX.
 * -------------------------------------------------------------------- */

IPRS_HRECT_RULE *agglomerate_cluster(FEATURE_LIST *f, int max_ncluster)
{
	float **feature,**d,dmin;
	int *n,*chain,*label,rmin=0,cmin=0;
	int npoints,nfeatures,ncluster,i,j,l,r,c;
	IPRS_HRECT_RULE *hrrptr;

	/* collect features in array */

	npoints=list_length(f->elements);
	if (npoints<1) error_exit("agglomerate cluster: too few elements");
	if (npoints>300) error_exit("agglomerate cluster: too many elements");
	if (f->type == 'U')
		nfeatures=nunary_features;
	else
		nfeatures=nbinary_features;

	feature=(float **)iprs_calloc(npoints,sizeof(float *));
	for (i=0;i<npoints;i++)
		feature[i]=(float *)iprs_calloc(nfeatures,sizeof(float));

	if (f->type == 'U') {
		i=0;
		foreach(UNARY,u,f->elements) {
			for (j=0;j<nfeatures;j++) {
				feature[i][j]=u->value[j];
			}
			i++;
		}
		endforeach
	}
	else /* f->type == 'B' */ {
		i=0;
		foreach(BINARY,b,f->elements) {
			for (j=0;j<nfeatures;j++) {
				feature[i][j]=b->value[j];
			}
			i++;
		}
		endforeach
	}

	/* distance matrix */

	d=(float **)iprs_calloc(npoints-1,sizeof(float *));
	for (i=0;i<npoints;i++) d[i]=(float *)iprs_calloc(npoints,sizeof(float));
	for (r=0;r<npoints-1;r++) {
		for (c=r+1;c<npoints;c++) {
			d[r][c]=0.0;
			for (i=0;i<nfeatures;i++) {
				d[r][c] += SQR(feature[r][i]-feature[c][i]);
			}
		}
	}

	/* cluster size, cluster chain, labels, rmin and rmax arrays */

	n=(int *)iprs_calloc(npoints,sizeof(int));
	for (i=0;i<npoints;i++) n[i]=1;
	chain=(int *)iprs_calloc(npoints,sizeof(int));
	for (i=0;i<npoints;i++) chain[i]= -1;
	label=(int *)iprs_calloc(npoints,sizeof(int));

	/* cluster loop */

	ncluster=npoints;
	while (1) {
		dmin=IPRS_FLOAT_MAX;
		for (r=0;r<npoints-1;r++) {
			if (n[r]>0) {
				for (c=r+1;c<npoints;c++) {
					if (d[r][c] < dmin) {
						dmin = d[r][c];
						rmin = r;
						cmin = c;
					}
				}
			}
		}

		if (dmin > DZERO && ncluster <= max_ncluster) break;

		r=rmin;
		c=cmin;
		for (j=0;j<nfeatures;j++) {
			feature[r][j]=(n[r]*feature[r][j]+n[c]*feature[c][j])/
						  (float)(n[r]+n[c]);
		}
		for (i=0;i<r;i++) {
			if (n[i]>0) {
				d[i][r]=0.0;
				for (j=0;j<nfeatures;j++) {
					d[i][r] += SQR(feature[i][j]-feature[r][j]);
				}
			}
		}
		for (i=r+1;i<npoints;i++) {
			if (i != c && n[i]>0) {
				d[r][i]=0.0;
				for (j=0;j<nfeatures;j++) {
					d[r][i] += SQR(feature[i][j]-feature[r][j]);
				}
			}
		}
		for (i=0;  i<c;      i++) d[i][c]=IPRS_FLOAT_MAX;
		for (i=c+1;i<npoints;i++) d[c][i]=IPRS_FLOAT_MAX;
		n[r]=n[r]+n[c];
		n[c]=0;

		for (i=r; chain[i] != -1; i=chain[i]);
		chain[i]=c;

		ncluster--;
	}

	/* label each feature with cluster number */

	l=0;
	for (i=0;i<npoints;i++) {
		if (n[i]>0) {
			for (j=i; j != -1; j=chain[j]) label[j]=l;
			l++;
		}
	}

#if 0

	/* Debugging: print out whole mess */

	if (f->type == 'U') {
		i=0;
		foreach(UNARY,u,f->elements) {
			printf("%2d: chain=%2d label=%2d feature=(",
				   i,chain[i],label[i]);
			for (j=0;j<nfeatures;j++) {
				printf(" %f",u->value[j]);
			}
			printf(")\n");
			i++;
		}
		endforeach
	}
	else /* f->type == 'B' */ {
		i=0;
		foreach(BINARY,b,f->elements) {
			printf("%2d: chain=%2d label=%2d feature=(",
				   i,chain[i],label[i]);
			for (j=0;j<nfeatures;j++) {
				printf(" %f",b->value[j]);
			}
			printf(")\n");
			i++;
		}
		endforeach
	}

#endif

	/* make rules */

	hrrptr=iprs_create_hrectrule(nfeatures,ncluster);
	for (r=0;r<ncluster;r++) {
		for (i=0;i<nfeatures;i++) {
			hrrptr->lower[r][i]=  IPRS_FLOAT_MAX;
			hrrptr->upper[r][i]= -IPRS_FLOAT_MAX;
		}
	}
	for (i=0;i<npoints;i++) {
		l=label[i];
		for (j=0;j<nfeatures;j++) {
			if (feature[i][j] < hrrptr->lower[l][j])
				hrrptr->lower[l][j] = feature[i][j];
			if (feature[i][j] > hrrptr->upper[l][j])
				hrrptr->upper[l][j] = feature[i][j];
		}
	}

	/* free junk */

	for (i=0;i<npoints;i++) iprs_free(feature[i]);
	iprs_free(feature);
	for (i=0;i<npoints;i++) iprs_free(d[i]);
	iprs_free(d);
	iprs_free(n);
	iprs_free(chain);
	iprs_free(label);

	return(hrrptr);
}


/* --------------------------------------------------------------------
 * Procedure: fuzzy_kmeans_cluster
 * Comments:  cluster elements into nclusters clusters.
 * -------------------------------------------------------------------- */

void sum_fuzzy_members(float **feature, int npoints,
					   float **cluster_centres, int nclusters,
					   int nfeatures, float fuzzyq, float *sum_points)
{
	int pt_index, cluster_index;

	for (pt_index=0; pt_index<npoints; pt_index++)
	{
		float sum=0.0;
		for (cluster_index=0; cluster_index<nclusters; cluster_index++)
		{
			float distance;
			distance=SQR(multid_feuclid_dist(feature[pt_index],
										  cluster_centres[cluster_index],
										  nfeatures));
			distance=pow(((float)1.0/(distance+0.00001)), 1.0/(fuzzyq-1));
			sum += distance;
		}
		sum_points[pt_index]=sum;
	}
}

#define RANDOM_MAX 2147483647

float **fuzzy_kmeans_cluster(FEATURE_LIST *f, int nclusters, float fuzzyq,
							 float delta_centres_threshold)
{
	float **feature, **cluster_centres;
	float *membership, *new_centre, max_delta_centres=0.0;
	float *sum_points;
	int nfeatures, npoints, i, j, k, index, loop=0;

	npoints = list_length(f->elements);
	if (f->type=='U')
		nfeatures = nunary_features;
	else nfeatures = nbinary_features;

	feature = (float**)iprs_calloc(npoints, sizeof(float*));
	for (i=0; i<npoints; i++)
	{
		feature[i] = (float*)iprs_calloc(nfeatures, sizeof(float));
	}

	if (f->type == 'U') {
		i=0;
		foreach(UNARY,u,f->elements) {
			for (j=0;j<nfeatures;j++) {
				feature[i][j]=u->value[j];
			}
			i++;
		}
		endforeach
	}
	else /* f->type == 'B' */ {
		i=0;
		foreach(BINARY,b,f->elements) {
			for (j=0;j<nfeatures;j++) {
				feature[i][j]=b->value[j];
			}
			i++;
		}
		endforeach
	}

	/* cluster vector: coords of each cluster centre */
	cluster_centres = (float**)iprs_calloc(nclusters, sizeof(float*));
	for (i=0; i<nclusters; i++)
	{
		cluster_centres[i] = (float*)iprs_calloc(nfeatures, sizeof(float));
		/* randomly pick a cluster centre */
		for (j=0; j<nfeatures; j++)
		{
			cluster_centres[i][j] = feature[i*(int)(npoints/nclusters)][j]
				+ random()/(float)RANDOM_MAX;
		}
	}

	/* membership vector: indicates degree of membership of pt in cluster */
	membership = (float*)iprs_calloc(npoints, sizeof(float));

	/* temp variable for storing the new centre of a cluster */
	new_centre = (float*)iprs_calloc(nfeatures, sizeof(float));

	/* sum_points: vector for sums of dists of each pt to all clusters */
	sum_points = (float*)iprs_calloc(npoints, sizeof(float));

	do	
	{
		max_delta_centres=0;
		/* to improve efficiency without using bazillions of memory */
		/* calc the sum of distances from each pt to all clusters */
		sum_fuzzy_members(feature, npoints, cluster_centres, nclusters,
						  nfeatures, fuzzyq, sum_points);
		for (i=0; i<nclusters; i++)
		{
			float sum, delta_centre;
			/* calculate fuzzy memberships of each point */
			sum = 0;
			for (j=0; j<npoints; j++)
			{
				float sqd_dist, memb_fuzzyq;
				sqd_dist = SQR(multid_feuclid_dist(feature[j],
												   cluster_centres[i],
												   nfeatures));
				membership[j] = pow((1.0/(sqd_dist+0.00001)), 
									1.0/(fuzzyq-1)) / (sum_points[j]+0.00001);
				memb_fuzzyq = pow(membership[j], fuzzyq);
				sum += memb_fuzzyq;
			}
			/* calculate new cluster centre */
			for (j=0; j<nfeatures; j++)
			{
				new_centre[j] = 0.0;
				for (k=0; k<npoints; k++)
					new_centre[j] += pow(membership[k], fuzzyq)*feature[k][j];
				new_centre[j] /= sum;
			}
			
			delta_centre = multid_feuclid_dist(cluster_centres[i], new_centre,
											   nfeatures);
			for (j=0; j<nfeatures; j++)
				cluster_centres[i][j] = new_centre[j];
			if (delta_centre>max_delta_centres)
				max_delta_centres = delta_centre;
		}
/* 		printf("max_delta_centres=%f\n", max_delta_centres); */
	}
	while ((max_delta_centres>delta_centres_threshold)&&(loop++<50));

	for (i=0; i<npoints; i++)
		iprs_free(feature[i]);
	iprs_free(feature);
	iprs_free(membership);
	iprs_free(new_centre);
	iprs_free(sum_points);

	return(cluster_centres);
}

/* --------------------------------------------------------------------
 * Procedure: fuzzy_validity
 * Comments:  Calculates a fuzzy validity criterion based on compactness
 *            and separation of clusters.
 *		      From Xie and Beni, vol PAMI-13, no8, pp.841-847, August, 91.
 * -------------------------------------------------------------------- */

float fuzzy_validity(FEATURE_LIST *f, float **cluster_centres, int nclusters, 
					 float fuzzyq)
{
	float **feature;
	float *membership, *new_centre, max_delta_centres=0.0;
	float *sum_points;
	int nfeatures, npoints, i, j, k, index, loop=0;
	float valid_sum=0.0, min_dist=FLT_MAX;
	float ret, penalty;
	int num_dist;
	int *numpoints_in_clusters;

	if (nclusters<2) return(FLT_MAX);

	numpoints_in_clusters = (int*)iprs_calloc(nclusters, sizeof(int));
	npoints = list_length(f->elements);
	if (f->type=='U')
	{
		nfeatures = nunary_features;
		/* calculate number of points in each cluster */
		foreach(PATHNODE,pathnode,f->pathnodes) {
			int nearest_cluster;
			nearest_cluster = find_nearest_point(pathnode->unary->value, 
												 nfeatures, 
												 cluster_centres, 
												 nclusters);
			numpoints_in_clusters[nearest_cluster]++;
		}
		endforeach;
	}
	else 
	{
		nfeatures = nbinary_features;
		/* calculate number of points in each cluster */
		foreach(PATHNODE,pathnode,f->pathnodes) {
			int nearest_cluster;
			nearest_cluster = find_nearest_point(pathnode->binary->value, 
												 nfeatures, 
												 cluster_centres, 
												 nclusters);
			numpoints_in_clusters[nearest_cluster]++;
		}
		endforeach;
	}
	/* calculate function which penalises validity if the clusters have
	   a small number of points in them. */
	penalty=0.0;
	for (i=0; i<nclusters; i++)
		penalty += 1.0/(1.0+((float)numpoints_in_clusters[i]));
	penalty /= (float)nclusters;
	iprs_free(numpoints_in_clusters);

	feature = (float**)iprs_calloc(npoints, sizeof(float*));
	for (i=0; i<npoints; i++)
	{
		feature[i] = (float*)iprs_calloc(nfeatures, sizeof(float));
	}

	if (f->type == 'U') {
		i=0;
		foreach(UNARY,u,f->elements) {
			for (j=0;j<nfeatures;j++) {
				feature[i][j]=u->value[j];
			}
			i++;
		}
		endforeach
	}
	else /* f->type == 'B' */ {
		i=0;
		foreach(BINARY,b,f->elements) {
			for (j=0;j<nfeatures;j++) {
				feature[i][j]=b->value[j];
			}
			i++;
		}
		endforeach
	}

	/* membership vector: indicates degree of membership of pt in cluster */
	membership = (float*)iprs_calloc(npoints, sizeof(float));

	/* sum_points: vector for sums of dists of each pt to all clusters */
	sum_points = (float*)iprs_calloc(npoints, sizeof(float));

	max_delta_centres=0;
	/* to improve efficiency without using bazillions of memory */
	/* calc the sum of distances from each pt to all clusters */
	sum_fuzzy_members(feature, npoints, cluster_centres, nclusters,
					  nfeatures, fuzzyq, sum_points);
	for (i=0; i<nclusters; i++)
	{
		float sum, delta_centre;
		/* calculate fuzzy memberships of each point */
		sum = 0;
		for (j=0; j<npoints; j++)
		{
			float sqd_dist, memb_fuzzyq;
			sqd_dist = SQR(multid_feuclid_dist(feature[j],
											   cluster_centres[i],
											   nfeatures));
			membership[j] = pow((1.0/(sqd_dist+0.00001)), 
								1.0/(fuzzyq-1)) / (sum_points[j]+0.00001);
			memb_fuzzyq = pow(membership[j], fuzzyq);
			sum += memb_fuzzyq;

			valid_sum += sqd_dist*memb_fuzzyq;
		}
	}

	/* find minimum distance between any two cluster centres */
	min_dist = 0.0;
	num_dist=0;
	for (i=0; i<nclusters-1; i++)
	{
		float this_cluster_min_dist=FLT_MAX;
		for (j=i+1; j<nclusters; j++)
		{
			float tmp_dist;
			tmp_dist = multid_feuclid_dist(cluster_centres[i], 
										  cluster_centres[j], nfeatures);
			if (tmp_dist<this_cluster_min_dist)
				this_cluster_min_dist = tmp_dist;
		}
		min_dist += this_cluster_min_dist;
	}
	min_dist = SQR((min_dist/(float)(nclusters-1)));

	for (i=0; i<npoints; i++)
		iprs_free(feature[i]);
	iprs_free(feature);
	iprs_free(membership);
	iprs_free(sum_points);

/* 	printf("nclusters = %d, valid=%f, penalty=%f\n", nclusters, valid_sum/(npoints*min_dist), penalty); */
	ret = (float)(valid_sum/(npoints*min_dist)+penalty);
	return(ret);
}
