/* ====================================================================
 * Package:     list processing
 * ==================================================================== */

#include <iprs.h>
#include "crglist.h"

/* ====================================================================
 * Basic routines (shouldn't be used)
 * ==================================================================== */

#define CELL_BLOCK_SIZE 8192

static LIST FreeCells=NULL;

/* --------------------------------------------------------------------
 * Procedure: allocate_cell
 * Purpose:   return new cell
 * -------------------------------------------------------------------- */

LIST allocate_cell()
{
	LIST c;
	int i;

	if (FreeCells==(LIST)NULL) {
		FreeCells=(LIST)iprs_calloc(CELL_BLOCK_SIZE,sizeof(LIST_ITEM));
		if (FreeCells==(LIST)NULL) {
			fprintf(stderr,"list routines: Out of memory\n");
			exit(1);
		}
		for (i=1,c=FreeCells;i<CELL_BLOCK_SIZE;i++,c++) c->tail=c+1;
		c->tail=NULL;
	}
	c=FreeCells;
	FreeCells=FreeCells->tail;
	return c;
}

/* --------------------------------------------------------------------
 * Procedure: free_cell
 * Purpose:   
 * -------------------------------------------------------------------- */

void free_cell(LIST c)
{
	c->head=(void *)NULL;
	c->tail=FreeCells;
	FreeCells=c;
}

/* ====================================================================
 * List processing routines
 * ==================================================================== */

/* --------------------------------------------------------------------
 * Procedure: list_initialize
 * Purpose:   return empty list
 * Comment:   use foo=NIL for foo=list_initialize() (see crglist.h)
 * -------------------------------------------------------------------- */

LIST list_initialize()
{
	LIST n;
	n=allocate_cell();
	n->tail = (LIST)NULL;
	return(n);
}

/* --------------------------------------------------------------------
 * Procedure: list_append
 * Purpose:   append item to list
 * -------------------------------------------------------------------- */

void list_append(void *item, LIST list)
{
	LIST t;

	for (t=list; t->tail != (LIST)NULL; t=t->tail);
	t->head=item;
	t->tail=allocate_cell();
	t->tail->tail=(LIST)NULL;
}

/* --------------------------------------------------------------------
 * Procedure: list_prepend
 * Purpose:   prepend item to list
 * -------------------------------------------------------------------- */

void list_prepend(void *item, LIST list)
{
	LIST t;

	t=allocate_cell();
	t->head=list->head;
	t->tail=list->tail;
	list->head=item;
	list->tail=t;
}

/* --------------------------------------------------------------------
 * Procedure: list_copy
 * Purpose:   return copy of list
 * -------------------------------------------------------------------- */

LIST list_copy(LIST old)
{
	LIST new,n,o;

	new=allocate_cell();
	for (o=old,n=new; o->tail != (LIST)NULL; o=o->tail,n=n->tail) {
		n->head=o->head;
		n->tail=allocate_cell();
	}
	n->tail=(LIST)NULL;

	return new;
}

/* --------------------------------------------------------------------
 * Procedure: list_reverse
 * Purpose:   return reverse of list
 * -------------------------------------------------------------------- */

LIST list_reverse(LIST list)
{
	LIST rev,new,t;

	rev=allocate_cell();
	rev->tail=(LIST)NULL;
	for (t=list; t->tail != (LIST)NULL; t=t->tail ) {
		new=allocate_cell();
		new->tail=rev->tail;
		new->head=rev->head;
		rev->tail=new;
		rev->head=t->head;
	}
	return rev;
}

/* --------------------------------------------------------------------
 * Procedure: list_union
 * Purpose:   append item to list unless it is already in list
 * -------------------------------------------------------------------- */

void list_union(void *item, LIST list)
{
	LIST t;
	for (t=list; t->tail != (LIST)NULL; t=t->tail ) {
		if (t->head == item) return;
	}
	t->head=item;
	t->tail=allocate_cell();
	t->tail->tail=(LIST)NULL;
}

/* --------------------------------------------------------------------
 * Procedure: list_list
 * Purpose:   return list containing item
 * -------------------------------------------------------------------- */

LIST list_list(void *item)
{
	LIST n;
	n=allocate_cell();
	n->head=item;
	n->tail=allocate_cell();
	n->tail->tail=(LIST)NULL;
	return(n);
}

/* --------------------------------------------------------------------
 * Procedure: list_empty
 * Purpose:   return true if list is empty
 * -------------------------------------------------------------------- */

int list_empty(LIST list)
{
	return (list->tail == (LIST)NULL);
}

/* --------------------------------------------------------------------
 * Procedure: list_member
 * Purpose:   return true if item is in list
 * -------------------------------------------------------------------- */

int list_member(void *item, LIST list)
{
	LIST t;
	for (t=list; t->tail != (LIST)NULL; t=t->tail ) {
		if (t->head == item) return(1);
	}
	return(0);
}

/* --------------------------------------------------------------------
 * Procedure: list_first
 * Purpose:   return first element of list
 * -------------------------------------------------------------------- */

void *list_first(LIST list)
{
	return(list->head);
}

/* --------------------------------------------------------------------
 * Procedure: list_rest
 * Purpose:   return remainder of list
 * -------------------------------------------------------------------- */

LIST list_rest(LIST list)
{
	return(list->tail);
}

/* --------------------------------------------------------------------
 * Procedure: list_remove
 * Purpose:   remove item from list (pointer to item returned)
 * -------------------------------------------------------------------- */

void *list_remove(void *item, LIST list)
{
	void *h;
	LIST t,n;

	for (t=list;t->tail != (LIST)NULL; t=t->tail) {
		if (t->head == item) break;
	}
	if (t->tail != (LIST)NULL) {
		h=t->head;
		n=t->tail;
		t->head=n->head;
		t->tail=n->tail;
		free_cell(n);
	}
	else {
		h = (void *)NULL;
	}
	return(h);
}

/* --------------------------------------------------------------------
 * Procedure: list_remove_first
 * Purpose:   remove first item from list (pointer to item returned)
 * -------------------------------------------------------------------- */

void *list_remove_first(LIST list)
{
	void *h;
	LIST n;

	if (list->tail == (LIST)NULL)
		return((void *)NULL);
	h = list->head;
	n = list->tail;
	list->head = n->head;
	list->tail = n->tail;
	free_cell(n);
	return(h);
}

/* --------------------------------------------------------------------
 * Procedure: list_length
 * Purpose:   return length of list
 * -------------------------------------------------------------------- */

int list_length (LIST list)
{
	int n = 0;
	while (list->tail != (LIST)NULL) {
		n++;
		list=list->tail;
	}
	return(n);
}

/* --------------------------------------------------------------------
 * Procedure: list_free
 * Purpose:   delete and free a list
 * -------------------------------------------------------------------- */

void list_free (LIST list)
{
	LIST t;

	while(list != (LIST)NULL) {
		t=list;
		list=list->tail;
		free_cell(t);
	}
}

/* --------------------------------------------------------------------
 * Procedure: list_intersection
 * Purpose:   return intersection of two lists
 * -------------------------------------------------------------------- */

LIST list_intersection(LIST a, LIST b)
{
	LIST n,ta,tb;

	n=list_initialize();
	for (ta=a; ta->tail != (LIST)NULL; ta=ta->tail) {
		for (tb=b; tb->tail != (LIST)NULL; tb=tb->tail) {
			if (ta->head == tb->head) {
				list_prepend(ta->head,n);
				break;
			}
		}
	}
	return n;
}
