/* ====================================================================
 * Package:     order
 * Purpose:		order the cluster tree such that, at runtime, the number
 *              of tests for classification of an arbitrary path tupel 
 *              through sequential evaluation of rules is minimized. Let
 *              ni (i=1,...,m) be the clusters conditional on a cluster n,
 *              let pi be prob(tupel in ni|tupel in n), and let ci be the
 *              expected cost of testing the rules associated with cluster
 *              tree ni. The cost for evaluating a leaf cluster is pro-
 *              portional to the number of feature dimensions (i.e.
 *              nunary_features or nbinary_features), and the expected
 *              cost for evaluating n is 
 *              c=c1+(1-p1)*c2+(1-p1-p2)*c3+...
 *              c is minimal if the clusters ni are sorted such that
 *              ci/pi <= cj/pj j=i+1,i=1,...,m-1
 * Notes:		This is only correct for the case of disjunct rule
 *              hyperrectangles, but this is the case in most of my
 *              rules. For non-disjunct rules the expected cost is
 *              somewhat lower and more complex to evaluate.
 * ==================================================================== */

/* Include files */

#include <iprs.h>
#include "crg.h"

/* Definitions */

typedef struct _cost_table {
	CLUSTER *cluster;
	float cost;
	float prob;
	float cp;
} COST_TABLE;

/* External Declarations */

extern LIST      binary_list;
extern float    *binary_difference_threshold;
extern float    *binary_mean;
extern float    *binary_range;
extern float    *binary_sdev;
extern float     expected_rule_cost;
extern int       max_path_length;
extern int       nbinary_features;
extern int       nclasses;
extern int       nnodes_analyzed;
extern int       nparts;
extern int       npaths;
extern int       npaths_analyzed;
extern int       nunary_features;
extern int       nviews;
extern LIST      part_list;
extern PATHNODE *path_tree;
extern LIST      unary_list;
extern float    *unary_difference_threshold;
extern float    *unary_mean;
extern float    *unary_range;
extern float    *unary_sdev;

/* Parameters */

extern int       breadth_first;
extern int       crg_debug_clusters;
extern int       crg_debug_parts;
extern int       crg_debug_paths;
extern int       crg_debug_varia;
extern float     entropy_threshold;
extern int       entropy_wrt_all_classes;
extern int       feature_ordering;
extern int       fixed_path_length;
extern float     leader_threshold;
extern int       max_level;
extern int       symmetric_relations;
extern int       use_leader_cluster;
extern int       zscore;

/* --------------------------------------------------------------------
 * Procedure: sort_cost
 * Comments:  sort cost table
 * -------------------------------------------------------------------- */

void sort_cost(COST_TABLE **ct, int n)
{
	COST_TABLE *temp;
	int i,j,gap;

	for (gap=1; gap<=n; gap=3*gap+1);
	for (; gap>0; gap/=3) {
		for (i=gap; i<n; i++) {
			temp=ct[i];
			for (j=i-gap;
				 (j>=0) && (ct[j]->cp > temp->cp);
				 j-=gap)
				ct[j+gap]=ct[j];
			ct[j+gap]=temp;
		}
	}
}

/* --------------------------------------------------------------------
 * Procedure: sort_cluster
 * Comments:  
 * -------------------------------------------------------------------- */

float sort_cluster(CLUSTER *c)
{
	COST_TABLE **ct;
	LIST sorted_children;
	float cost,sum_prob,prob;
	int i,n;

	/* leaf cluster */

	if (list_empty(c->children)) {
		if (c->type == 'U')
			return((float)nunary_features);
		else
			return((float)nbinary_features);
	}

	/* allocate cost tables */

	n=list_length(c->children);
	ct=(COST_TABLE **)iprs_calloc(n,sizeof(COST_TABLE *));
	for (i=0;i<n;i++)
		ct[i]=(COST_TABLE *)iprs_calloc(1,sizeof(COST_TABLE));

	/* clusters into table */

	i=0;
	foreach(CLUSTER,child,c->children) {
		ct[i++]->cluster=child;
	}
	endforeach

	/* prob(path in child | path in parent) */

	sum_prob=0.0;
	for (i=0;i<n;i++) {
		ct[i]->prob=(float)list_length(ct[i]->cluster->pathnodes);
		sum_prob+=ct[i]->prob;
	}
	for (i=0;i<n;i++) ct[i]->prob /= sum_prob;

	/* cost */

	for (i=0;i<n;i++) ct[i]->cost=sort_cluster(ct[i]->cluster);

	/* sort ct according to cost/prob ratio */

	for (i=0;i<n;i++) ct[i]->cp = ct[i]->cost / ct[i]->prob;
	sort_cost(ct,n);

	/* compute decision cost for c */

	cost=0.0;
	prob=1.0;
	for (i=0;i<n;i++) {
		cost += prob * ct[i]->cost;
		prob -= ct[i]->prob;
	}
	if (c->level > 0) {
		if (c->type == 'U')
			cost += (float)nunary_features;
		else
			cost += (float)nbinary_features;
	}

	/* save sorted list of children */

	sorted_children=NIL;
	for (i=0;i<n;i++)
		list_append(ct[i]->cluster,sorted_children);
	list_free(c->children);
	c->children=sorted_children;

	return(cost);
}

/* --------------------------------------------------------------------
 * Procedure: rename_clusters
 * Comments:  
 * -------------------------------------------------------------------- */

void rename_clusters(CLUSTER *c)
{
	char type;
	int i;

	if (c->type == 'U')
		type = 'B';
	else /* c->type == 'B' || c->type == '0' */
		type = 'U';
	i=1;
	foreach(CLUSTER,child,c->children) {
		sprintf(child->name,"%s%c%-d",c->name,type,i++);
		rename_clusters(child);
	}
	endforeach
}

/* --------------------------------------------------------------------
 * Procedure: sort_cluster_tree
 * Comments:  
 * -------------------------------------------------------------------- */

void sort_cluster_tree(CLUSTER *root)
{
	expected_rule_cost=sort_cluster(root);
	rename_clusters(root);
}
