/* ====================================================================
 * Package:   path (revised version)
 * Comments:  Paths are used to track how features are included in a 
 *            feature list to avoid cycles. Given a graph
 *
 *                             A - B - D
 *                              \ /
 *                               C
 *
 *            the path A-B-C-A must be excluded.
 *            For the above graph we have the following path tree:
 *
 *            level0                          root
 *                                             |
 *                           -------------------------------------
 *                           |            |         |            |
 *            level1         A            B         C            D
 *                          / \         / | \      / \           |
 *                         /   \       /  |  \    /   \          |
 *            level2     AB     AC   BA  BC  BD  CA    CB        DB
 *                      /  \     |    |   |      |    /  \      /  \
 *            level3  ABC  ABD  ACB  BAC BCA    CAB CBA  CBD  DBA  DBC
 *                               |               |             |    |
 *            level4            ACBD            CABD          DBAC DBCA
 *
 *            The first path version operated on paths of fixed length
 *            only. If max_path_length was set 3 the partial
 *            path BD was never analyzed because it cannot be extended
 *            further. Now we consider also the partial paths. To analyze
 *            paths of fixed length set "fixed_path_length=1".
 *
 * Future:    feature ordering has been dropped (for the moment) and will
 *            have to be introduced again.
 * ==================================================================== */

/* Include files */

#include <iprs.h>
#include "crg.h"

/* External Declarations */

extern LIST      binary_list;
extern float    *binary_difference_threshold;
extern float    *binary_mean;
extern float    *binary_range;
extern float    *binary_sdev;
extern float     expected_rule_cost;
extern int       max_path_length;
extern int       nbinary_features;
extern int       nclasses;
extern int       nnodes_analyzed;
extern int       nparts;
extern int       npaths;
extern int       npaths_analyzed;
extern int       nunary_features;
extern int       nviews;
extern LIST      part_list;
extern PATHNODE *path_tree;
extern LIST      unary_list;
extern float    *unary_difference_threshold;
extern float    *unary_mean;
extern float    *unary_range;
extern float    *unary_sdev;

/* Parameters */

extern int       breadth_first;
extern int       crg_debug_clusters;
extern int       crg_debug_parts;
extern int       crg_debug_paths;
extern int       crg_debug_varia;
extern float     entropy_threshold;
extern int       entropy_wrt_all_classes;
extern int       feature_ordering;
extern int       fixed_path_length;
extern float     leader_threshold;
extern int       max_level;
extern int       symmetric_relations;
extern int       use_leader_cluster;
extern int       zscore;

/* Static */

static int      *path_length_histo;

/* --------------------------------------------------------------------
 * Procedure: create_path_tree
 * Comments:  create complete path tree
 * -------------------------------------------------------------------- */

PATHNODE *create_path_tree()
{
	int missing_parts,i;
	PATHNODE *root,*child;

	max_path_length=max_level/2+1;
	npaths=0;
	missing_parts=0;
	path_length_histo=(int *)iprs_calloc(max_path_length+1,sizeof(int));
	for (i=1;i<=max_path_length;i++) path_length_histo[i]=0;

	/* create root node */

	root=(PATHNODE *)iprs_malloc(sizeof(PATHNODE));
	root->part=(PART *)NULL;
	root->level=0;
	root->parent=(PATHNODE *)NULL;
	root->children=NIL;
	root->unary=(UNARY *)NULL;
	root->binary=(BINARY *)NULL;

	/* create all subtrees */

	foreach(PART,part,part_list) {
		child=create_path_subtree(part,root,part->unary,(BINARY *)NULL);
		if (child != (PATHNODE *)NULL) {
			list_append(child,root->children);
		}
		else {
			printf("create_path_tree: Part %d not in a path\n",part->name);
			missing_parts=1;
		}
	}
	endforeach

	if (missing_parts) exit(1);

	/* for feature ordering, npaths_analyzed <= npaths */

	npaths_analyzed=npaths;

	/* print path length districution */

	if (crg_debug_varia) {
		printf("\nPath length distribution\nlength number\n");
		for (i=1;i<=max_path_length;i++) {
			if (path_length_histo[i]>0)
				printf("%4d %6d\n",i,path_length_histo[i]);
		}
		printf("\n");
	}

	return root;
}

/* --------------------------------------------------------------------
 * Procedure: create_path_subtree
 * Comments:  create a partial path tree
 * -------------------------------------------------------------------- */

PATHNODE *create_path_subtree(PART *part, PATHNODE *parent,
							  UNARY *unary, BINARY *binary)
{
	int level;
	PART *part2;
	PATHNODE *node, *child;

	level=parent->level+1;

	/* create node */

	node=(PATHNODE *)iprs_malloc(sizeof(PATHNODE));
	node->part=part;
	node->level=level;
	node->parent=parent;
	node->children=NIL;
	node->unary=unary;
	node->binary=binary;

	if (level >= max_path_length) {
		npaths++;
		path_length_histo[level]++;
		return node;
	}
	else {
		foreach(BINARY,binary,part->binary) {
			part2=binary->part2;
			if (!part_in_path(part2,parent)) {
				child=create_path_subtree(part2,node,part2->unary,binary);
				if (child != (PATHNODE *)NULL)
					list_append(child,node->children);
			}
		}
		endforeach
		if (list_empty(node->children)) {
			if (fixed_path_length) {
				iprs_free(node);
				return((PATHNODE *)NULL);
			}
			else {
				npaths++;
				path_length_histo[level]++;
				return node;
			}
		}
		else {
			return node;
		}
	}
}

/* --------------------------------------------------------------------
 * Procedure: part_in_path
 * Comments:  test whether part occurs in parent chain
 * -------------------------------------------------------------------- */

int part_in_path(PART *part, PATHNODE *pathnode)
{
	PATHNODE *node;

	for (node=pathnode;node != (PATHNODE *)NULL;node=node->parent) {
		if (node->part == part) return 1;
	}
	return 0;
}

/* --------------------------------------------------------------------
 * Procedure: get_path
 * Comments:  given a node in the path tree return the full path
 * -------------------------------------------------------------------- */

PATH *get_path(PATHNODE *node)
{
	int level,i;
	PATH *path;
	PATHNODE *n;

	level=node->level;

	path=(PATH *)iprs_malloc(sizeof(PATH));
	path->length=level;
	path->part=(PART **)iprs_calloc(level,sizeof(PART *));

	for (i=0,n=node;i<level;i++,n=n->parent)
		path->part[i]=n->part;

	return path;
}

/* --------------------------------------------------------------------
 * Procedure: free_path
 * Comments:  free path
 * -------------------------------------------------------------------- */

void free_path(PATH *path)
{
	iprs_free(path->part);
	iprs_free(path);
}
