/* ====================================================================
 * Package:     split
 * Author:      Walter F. Bischof
 * ==================================================================== */

/* Include files */

#include <iprs.h>
#include <string.h>
#include "crg.h"

/* Definitions */

#define NULL_CLUSTER ((CLUSTER *)NULL)

typedef struct _cv {                /* class-value pairs for sorting */
	int class;
	float *value;
} CV;

void split_sort(CV **cv, int feature, int n);
int collect_elements(CLUSTER *c, CLUSTER *leaf, CV **cv, int nelements);

/* External Declarations */

extern LIST      binary_list;
extern float    *binary_difference_threshold;
extern float    *binary_mean;
extern float    *binary_range;
extern float    *binary_sdev;
extern float     expected_rule_cost;
extern int       max_path_length;
extern float     min_cluster_size;
extern int       nbinary_features;
extern int       nclasses;
extern int       nnodes_analyzed;
extern int       nparts;
extern int       npaths;
extern int       npaths_analyzed;
extern int       nunary_features;
extern int       nviews;
extern LIST      part_list;
extern PATHNODE *path_tree;
extern LIST      unary_list;
extern float    *unary_difference_threshold;
extern float    *unary_mean;
extern float    *unary_range;
extern float    *unary_sdev;

/* Parameters */

extern int       breadth_first;
extern int       crg_debug_clusters;
extern int       crg_debug_parts;
extern int       crg_debug_paths;
extern int       crg_debug_varia;
extern float     entropy_threshold;
extern int       entropy_wrt_all_classes;
extern int       feature_ordering;
extern int       fixed_path_length;
extern float     leader_threshold;
extern int       max_level;
extern int       symmetric_relations;
extern int       use_leader_cluster;
extern int       zscore;

float sum_clist_entropy(LIST clist)
{
	float entropy=0;
	int total=0;

	foreach (CLUSTER, c, clist)
	{
		entropy += ((c->entropy)*(list_length(c->pathnodes)));
		total += list_length(c->pathnodes);
	}
	endforeach;
	entropy /= (float)total;
	return(entropy);
}

void free_cluster_list(LIST clist)
{
	foreach (CLUSTER,c,clist)
		free_cluster(c);
	endforeach;
	list_free(clist);
}
	  
/* --------------------------------------------------------------------
 * Procedure: fuzzy_split
 * Comments:  Determine the best place to split a fuzzy cluster. Fuzzy 
 *            splitting is performed here by reclustering the entire 
 *            space where the best split is found. Hence, all sibling
 *            clusters and their corresponding subtrees are lost.
 * Notes:     This is a very inefficient method of building the cluster
 *            tree, but is probably the most correct in terms of the
 *            goodness of the clusters formed.
 * -------------------------------------------------------------------- */

LIST fuzzy_split(CLUSTER *leaf, LIST queue, float fuzzyq, 
		   float delta_centres_threshold)
{
	CLUSTER *c, *worst_cluster=(CLUSTER *)NULL;
	LIST best_clist, temp_clist;
	float best_split_entropy=FLT_MAX, sum_entropy;
	FEATURE_LIST *feature_list;
	int num_new_clusters;

	/* search the chain leaf - ... - UB - U for the best split  */
	/* with respect to the leaf class vector. With ties, splits */
	/* higher up in the chain are preferred.                    */

	best_clist=NIL;
	for (c=leaf->parent; c != NULL; c=c->parent) {
/* 		sum_entropy = c->entropy_1more - */
/* 			          sum_clist_entropy(c->children); */
		if (sum_entropy<=best_split_entropy)
		{
			worst_cluster = c;
			best_split_entropy = sum_entropy;
		}
	}
	
	/* split best_cluster if possible */

	if (crg_debug_clusters) {
		printf("\nsplit: %s unresolved, reclustering %s\n",
			   leaf->name,worst_cluster->name);
	}
	
	num_new_clusters = list_length(worst_cluster->children)+1;
	foreach (CLUSTER, child, worst_cluster->children)
	{
		remove_cluster_tree_from_queue(child,queue);
		free_cluster_tree(child);
	}
	endforeach;
	list_free(worst_cluster->children);
	worst_cluster->children=NIL;
	best_clist = cluster_feature_list(worst_cluster->cfs,
                                          num_new_clusters,
                                          fuzzyq, delta_centres_threshold);
/* 	free_feature_list(feature_list); */

	return(best_clist);
}

/* --------------------------------------------------------------------
 * Procedure: fuzzy_split
 * Comments:  Determine the best place to split a fuzzy cluster. Fuzzy 
 *            splitting is performed here by clustering the original
 *            cluster into two fuzzy clusters.
 * Notes:    
 * -------------------------------------------------------------------- */

LIST fuzzy_split2(CLUSTER *leaf, LIST queue, float fuzzyq, 
		   float delta_centres_threshold)
{
	CLUSTER *c, *worst_cluster=(CLUSTER *)NULL, *parent;
	LIST best_clist, temp_clist;
	float best_split_entropy=FLT_MAX, split_entropy;
	int num_new_clusters;

	/* search the chain leaf - ... - UB - U for the best split  */
	/* with respect to the leaf class vector. With ties, splits */
	/* higher up in the chain are preferred.                    */

	best_clist=NIL;
	for (c=leaf; c->parent != NULL; c=c->parent) {
		CLUSTER *tmp;
		LIST clist;
		
		clist = recluster_cluster(c, 2, fuzzyq, delta_centres_threshold);
		split_entropy = fuzzy_split_entropy(clist);
/* 		printf("split entropy for %s = %f\n", c->name, split_entropy); */
		if (split_entropy<=best_split_entropy)
		{
			worst_cluster = c;
			best_split_entropy = split_entropy;
			while ((tmp=(CLUSTER*)list_remove_first(best_clist)))
				free_cluster(tmp);
			list_free(best_clist);
			best_clist = clist;
		}
		else
		{
			while ((tmp=(CLUSTER*)list_remove_first(clist)))
				free_cluster(tmp);
			list_free(clist);
		}
	}
	
	/* split best_cluster if possible */

	if (list_length(worst_cluster->elements)<4)
	{
		CLUSTER *tmp;
		while ((tmp=(CLUSTER*)list_remove_first(best_clist)))
			free_cluster(tmp);
		list_free(best_clist);
		best_clist = NIL;
		if (crg_debug_clusters) 
			printf("\nsplit: not enough points in %s\n", worst_cluster->name);
	}
	else
	{
		if (crg_debug_clusters) {
			printf("\nsplit: %s unresolved, reclustering %s\n",
				   leaf->name,worst_cluster->name);
		}

		parent = worst_cluster->parent;
		list_remove(worst_cluster, parent->children);
		foreach (CLUSTER, c, best_clist)
			list_append(c, parent->children);
		endforeach;
		remove_cluster_tree_from_queue(worst_cluster,queue);
		free_cluster_tree(worst_cluster);
	}

	return(best_clist);
}

/* --------------------------------------------------------------------
 * Procedure: best_split
 * Comments:  For cluster c and the class vector of a leaf cluster,
 *            determine the feature and the optimal cut for the feature
 *            that leads to the smallest entropy wrt to the class
 *            vector of the leaf cluster.
 * Notes:     range calculations: if two feature dimensions have the
 *            same split entropy, the dimension with the larger
 *            (normalized) feature range is selected. This eliminates
 *            the creation of very narrow hyperrectangles.
 * -------------------------------------------------------------------- */

void best_split(CLUSTER *c, CLUSTER *leaf,
				int   *best_found,
				float *best_entropy,
				int   *best_feature,
				float *best_threshold)
{
	int cmin_found,fmin_found,cmin_feature=-1;
	int fmin_points, cmin_points;  /* diagnostics telling me how many */
	int fmin_npoints, cmin_npoints;/* points either side of the split point. */
	float cmin_entropy,fmin_entropy,cmin_threshold,fmin_threshold;
	float cmin_range,range;
	int npoints,feature,nfeatures,i,nelements;
	int *c0,*c1,*c2;
	float *difference_threshold,h,rsize;
	CV **cv;

	/* I hate warnings: uninitialized */

	cmin_threshold=IPRS_FLOAT_MAX;
	fmin_threshold=IPRS_FLOAT_MAX;
	cmin_range=IPRS_FLOAT_MAX;

	/* collect elements */

	nelements=list_length(c->elements);
	cv=(CV **)iprs_calloc(nelements,sizeof(CV *));
	for (i=0;i<nelements;i++) cv[i]=(CV *)iprs_malloc(sizeof(CV));
	npoints=collect_elements(c,leaf,cv,nelements);

	/* Feature information */

	if (c->type == 'U') {
		nfeatures=nunary_features;
		difference_threshold=unary_difference_threshold;
	}
	else /* c->type == 'B' */ {
		nfeatures=nbinary_features;
		difference_threshold=binary_difference_threshold;
	}

	/* Initialize class vectors */

	c0=(int *)iprs_calloc(nclasses,sizeof(int)); /* class vector for all   */
	for (i=0;i<nclasses;i++) c0[i]=0;
	for (i=0;i<npoints;i++) c0[cv[i]->class]++;
	c1=(int *)iprs_calloc(nclasses,sizeof(int)); /* class vector for left  */
	c2=(int *)iprs_calloc(nclasses,sizeof(int)); /* class vector for right */

	/* Find best split */

	cmin_found=0;
	cmin_entropy=IPRS_FLOAT_MAX;

	for (feature=0;feature<nfeatures;feature++) {

		fmin_found=0;
		fmin_entropy=IPRS_FLOAT_MAX;

		split_sort(cv,feature,npoints);

		for (i=0;i<nclasses;i++) c1[i]=0;
		for (i=0;i<nclasses;i++) c2[i]=c0[i];

		/* This is where I can change the code to ensure clusters end up */
		/* with > 1 pts. Instead of going from 0 to npoints-1, I can go  */
		/* from 1 to npoints-2, or whatever.                             */
		for (i=0;i<npoints-1;i++) {
			float threshold;

			c1[cv[i]->class]++;
			c2[cv[i]->class]--;

			threshold = (cv[i+1]->value[feature]+cv[i]->value[feature])/2.0;
 			if (((threshold-cv[0]->value[feature]) >
				 difference_threshold[feature])&&
				((cv[npoints-1]->value[feature]-threshold) >
				 difference_threshold[feature]))
			{

				h=split_entropy(c1,c2,i,npoints);

				if (h<fmin_entropy) {
					fmin_found=1;
					fmin_entropy=h;
					fmin_threshold=threshold;
					fmin_points=i;
					fmin_npoints=npoints;
				}
			}
/* 			else if (crg_debug_varia) */
/* 				printf("feature diff %d not large enough: %f %f\n", feature, */
/* 					   (threshold-cv[0]->value[feature]),  */
/* 					   (cv[npoints-1]->value[feature]-threshold)); */
		}

		if (fmin_found) {

			/* feature range */

			range=cv[npoints-1]->value[feature]-cv[0]->value[feature];
			if (c->type == 'U')
				range /= unary_range[feature];
			else
				range /= binary_range[feature];
			if (cmin_found) {
				if (fmin_entropy<cmin_entropy) {
					cmin_entropy=fmin_entropy;
					cmin_threshold=fmin_threshold;
					cmin_feature=feature;
					cmin_range=range;
					cmin_points = fmin_points;
					cmin_npoints = fmin_npoints;
				}
				else if (fmin_entropy == cmin_entropy && range > cmin_range) {
					cmin_threshold=fmin_threshold;
					cmin_feature=feature;
					cmin_range=range;
					cmin_points = fmin_points;
					cmin_npoints = fmin_npoints;
				}
			}
			else {
				cmin_found=1;
				cmin_entropy=fmin_entropy;
				cmin_threshold=fmin_threshold;
				cmin_feature=feature;
				cmin_range=range;
				cmin_points = fmin_points;
				cmin_npoints = fmin_npoints;
			}
		}
	}

	if (cmin_found) {
		if (crg_debug_varia) {
			printf("best_split: S-entropy = %f of feature %2d of cluster %s\n",
				   cmin_entropy,cmin_feature,c->name);
			printf("c1->np=%d, c2->np=%d, crange=%f\n", cmin_points, 
				   cmin_npoints-cmin_points, cmin_range);
			/* getchar(); */
		}
		rsize=(float)(list_length(c->pathnodes))/
			  (float)(list_length(c->parent->cfs->pathnodes));
		if (rsize < min_cluster_size)
			cmin_entropy += 100.0;

		*best_found=1;
		*best_entropy=cmin_entropy;
		*best_feature=cmin_feature;
		*best_threshold=cmin_threshold;
	}
	else {

		if (crg_debug_varia) {
			printf("best_split: Cannot split any feature of cluster %s\n",
				   c->name);
		}

		*best_found=0;
	}

	/* free junk */

	for (i=0;i<nelements;i++) iprs_free(cv[i]);
	iprs_free(cv);
	iprs_free(c0);
	iprs_free(c1);
	iprs_free(c2);
}

/* --------------------------------------------------------------------
 * Procedure: collect_elements
 * Comment:   
 * -------------------------------------------------------------------- */

int collect_elements(CLUSTER *c, CLUSTER *leaf, CV **cv, int nelements)
{
	int npoints;

	/* if (entropy_wrt_all_classes)                             */
	/*      collect all elements                                */
	/* else                                                     */
	/*      collect only elements of classes unresolved in leaf */

	if (c->type == 'U') {
		npoints=0;
		if (entropy_wrt_all_classes) {
			foreach(UNARY,unary,c->elements) {
				cv[npoints]->class=unary->class;
				cv[npoints]->value=unary->value;
				npoints++;
			}
			endforeach
		}
		else {
			foreach(UNARY,unary,c->elements) {
				if (leaf->class[unary->class]>0) {
					cv[npoints]->class=unary->class;
					cv[npoints]->value=unary->value;
					npoints++;
				}
			}
			endforeach
		}
	}
	else /* c->type == 'B' */ {
		npoints=0;
		if (entropy_wrt_all_classes) {
			foreach(BINARY,binary,c->elements) {
				cv[npoints]->class=binary->class;
				cv[npoints]->value=binary->value;
				npoints++;
			}
			endforeach
		}
		else {
			foreach(BINARY,binary,c->elements) {
				if (leaf->class[binary->class]>0) {
					cv[npoints]->class=binary->class;
					cv[npoints]->value=binary->value;
					npoints++;
				}
			}
			endforeach
		}
	}
	return npoints;
}

/* --------------------------------------------------------------------
 * Procedure: split_sort
 * Comments:  sort feature values using shell sort
 * -------------------------------------------------------------------- */

void split_sort(CV **cv, int feature, int n)
{
	CV *temp;
	int i,j,gap;

	for (gap=1; gap<=n; gap=3*gap+1);
	for (; gap>0; gap/=3) {
		for (i=gap; i<n; i++) {
			temp=cv[i];
			for (j=i-gap;
				 (j>=0) && (cv[j]->value[feature] > temp->value[feature]);
				 j-=gap)
				cv[j+gap]=cv[j];
			cv[j+gap]=temp;
		}
	}
}

/* --------------------------------------------------------------------
 * Procedure: split_entropy
 * Comments:  
 * -------------------------------------------------------------------- */

float split_entropy(int *c1, int *c2, int n, int npoints)
{
	int i;
	float h1,h2;

	h1=0.0;
	h2=0.0;
	for (i=0;i<nclasses;i++) {
		if (c1[i]>0) h1-=c1[i]*logf((float)c1[i]/(float)(n+1));
		if (c2[i]>0) h2-=c2[i]*logf((float)c2[i]/(float)(npoints-n-1));
	}
	/* modified by B.McCane 24/9/94 */
	/* weight h1 and h2 so that a split of a cluster with more points */
	/* is as likely as a split of a cluster with less points */
/* 	h1 /= (float)(n+1); */
/* 	h2 /= (float)(npoints-n-1); */
	return((h1+h2)/(float)npoints);
}

float fuzzy_split_entropy(LIST cluster_list)
{
	int i, npoints=0;
	float h=0.0;

	foreach(CLUSTER, cluster, cluster_list)
	{
		float ch=0.0;
		int np=list_length(cluster->elements);
		for (i=0; i<nclasses; i++)
		{
			if (cluster->class[i]>0)
				ch -= cluster->class[i]*
					   logf((float)cluster->class[i]/(float)np);
		}
		h += (float)ch;
		npoints += np;
	}
	endforeach;
	return(h/(float)npoints);
}
		

/* --------------------------------------------------------------------
 * Procedure: split_cluster
 * Comments:  split cluster c into two by cutting the rule hyperrectangle
 *            along dimension best_feature at value best_threshold.
 *            Return list of (two) new clusters.
 * -------------------------------------------------------------------- */

/* LIST split_cluster(CLUSTER *c, int best_feature, float best_threshold) */
/* { */
/* 	RULE **nrule; */
/* 	CLUSTER *n[2], *parent; */
/* 	LIST ncluster; */
/* 	int i,j; */
/* 	float pratio; */
/* 	char *s; */
/* 	UNARY *unary; */
/* 	BINARY *binary; */

	/* Get new rules */

/* 	nrule=(RULE **)iprs_calloc(2,sizeof(RULE *)); */
/* 	for (i=0;i<2;i++) nrule[i]=(RULE *)iprs_calloc(1,sizeof(RULE)); */
/* 	split_rule(c,best_feature,best_threshold,nrule); */

	/* Create new clusters */

/* 	for (i=0;i<2;i++) { */
/*         n[i]=(CLUSTER *)iprs_calloc(1,sizeof(CLUSTER)); */
/*         n[i]->type=c->type; */
/*         n[i]->level=c->level; */
/* 		s=(char *)iprs_calloc(strlen(c->name)+10,sizeof(char)); */
/* 		n[i]->name=s; */
/*         n[i]->parent=c->parent; */
/*         n[i]->children=NIL; */
/* 		n[i]->rule=nrule[i]; */
/*         n[i]->elements=NIL; */
/*         n[i]->pathnodes=NIL; */
/* 		n[i]->nclass=0; */
/* 		n[i]->class=(int *)iprs_calloc(nclasses,sizeof(int)); */
/* 		n[i]->cfs=(FEATURE_LIST *)NULL; */
/* 		n[i]->cfs_possible=c->cfs_possible; */
/* 		n[i]->unresolvable=0; */
/* 	} */

	/* Names: n[0] same as c, n[1] new number */

/* 	strcpy(n[0]->name,c->name); */
/* 	i=list_length(c->parent->children); */
/* 	sprintf(n[1]->name,"%s%c%-d",c->parent->name,c->type,i); */

	/* initialize class vector for n[0] and n[1] */

/* 	for (i=0;i<2;i++) { */
/* 		for(j=0;j<nclasses;j++) n[i]->class[j]=0; */
/* 	} */

	/* distribute elements of c between n[0] and n[1] */

/* 	if (c->type=='U') { */
/* 		foreach(PATHNODE,pathnode,c->pathnodes) { */
/* 			unary=pathnode->unary; */
/* 			if (feature_in_rule(unary->value,n[0]->rule)) { */
/* 				list_append(pathnode,n[0]->pathnodes); */
/* 				list_union(unary,n[0]->elements); */
/* 				(n[0]->class[unary->class])++; */
/* 			} */
/* 			else { */
/* 				list_append(pathnode,n[1]->pathnodes); */
/* 				list_union(unary,n[1]->elements); */
/* 				(n[1]->class[unary->class])++; */
/* 			} */
/* 		} */
/* 		endforeach */
/* 	} */
/* 	else */
/* 		foreach(PATHNODE,pathnode,c->pathnodes) { */
/* 			binary=pathnode->binary; */
/* 			if (feature_in_rule(binary->value,n[0]->rule)) { */
/* 				list_append(pathnode,n[0]->pathnodes); */
/* 				list_union(binary,n[0]->elements); */
/* 				(n[0]->class[binary->class])++; */
/* 			} */
/* 			else { */
/* 				list_append(pathnode,n[1]->pathnodes); */
/* 				list_union(binary,n[1]->elements); */
/* 				(n[1]->class[binary->class])++; */
/* 			} */
/* 		} */
/* 		endforeach */
/* 	} */
/* 	for (i=0;i<2;i++) { */
/* 		n[i]->npathnodes=list_length(n[i]->pathnodes); */
/* 	} */

	/* nclass */

/* 	for (i=0;i<2;i++) { */
/* 		n[i]->nclass=0; */
/* 		for (j=0;j<nclasses;j++) { */
/* 			if (n[i]->class[j]>0) (n[i]->nclass)++; */
/* 		} */
/* 		n[i]->entropy=compute_entropy(n[i]->class); */
/* 		pratio=(float)(n[i]->npathnodes)/(float)(n[i]->parent->cfs->npathnodes); */
/* 		n[i]->weighted_entropy=pratio*n[i]->entropy; */
/* 	} */

	/* insert new clusters into cluster tree */

/* 	parent=c->parent; */
/* 	list_append(n[0],parent->children); */
/* 	list_append(n[1],parent->children); */
/* 	list_remove(c,parent->children); */

	/* print new clusters */

/* 	if(crg_debug_clusters) { */
/* 		printf("split_cluster: printing cluster n0\n"); */
/* 		print_cluster(n[0]); */
/* 		printf("split_cluster: printing cluster n1\n"); */
/* 		print_cluster(n[1]); */
/* 	} */

/* 	ncluster=NIL; */
/* 	for (i=0;i<2;i++) list_append(n[i],ncluster); */
/* 	return(ncluster); */

/* } */

/* --------------------------------------------------------------------
 * Procedure: split_rule
 * Comments:  Split the hyperrectangle rule of cluster c along dimension
 *            best_feature at value best_threshold
 * -------------------------------------------------------------------- */

/* void split_rule(CLUSTER *c, int best_feature, */
/* 				float best_threshold, RULE **nrule) */
/* { */
/* 	int feature,nfeatures,i; */

/* 	if (c->type == 'U') */
/* 		nfeatures=nunary_features; */
/* 	else */
/* 		nfeatures=nbinary_features; */

/* 	for (i=0;i<2;i++) { */
/* 		nrule[i]->nfeatures=nfeatures; */
/* 		nrule[i]->important=(int *)iprs_calloc(nfeatures,sizeof(int)); */
/* 		nrule[i]->lower=(float *)iprs_calloc(nfeatures,sizeof(float)); */
/* 		nrule[i]->upper=(float *)iprs_calloc(nfeatures,sizeof(float)); */
/* 		for (feature=0;feature<nfeatures;feature++) { */
/* 			nrule[i]->lower[feature]=c->rule->lower[feature]; */
/* 			nrule[i]->upper[feature]=c->rule->upper[feature]; */
/* 		} */
/* 	} */
/* 	nrule[0]->upper[best_feature]=best_threshold; */
/* 	nrule[1]->lower[best_feature]=best_threshold; */
/* } */

/* --------------------------------------------------------------------
 * Procedure: remove_cluster_tree_from_queue
 * Comments:  for a cluster that has been split, the cluster and its
 *            children must be removed from the queue in
 *            create_cluster_tree
 * -------------------------------------------------------------------- */

void remove_cluster_tree_from_queue(CLUSTER *c, LIST queue)
{
	list_remove(c,queue);
	foreach(CLUSTER,child,c->children) {
		remove_cluster_tree_from_queue(child,queue);
	}
	endforeach
}

/* --------------------------------------------------------------------
 * Procedure: split_paths
 * Comment:   split a cluster into two parts, one containing paths that
 *            can be expanded, the other with the rest of the paths.
 * -------------------------------------------------------------------- */

/* LIST split_paths(CLUSTER *c, LIST queue) */
/* { */
/* 	RULE **nrule; */
/* 	CLUSTER *n[2]; */
/* 	LIST ncluster,plist[2]; */
/* 	int feature,nfeatures,i,j; */
/* 	float pratio; */
/* 	char *s; */
/* 	UNARY *unary; */

/* 	if (c->type != 'U') */
/* 		error_exit("split_paths: cannot split paths of unary cluster"); */

/* 	if (crg_debug_clusters) */
/* 		printf("\nsplit_paths: %s cannot be expanded, splitting paths\n",c->name); */

	/* Split paths */

/* 	plist[0]=NIL; */
/* 	plist[1]=NIL; */
/* 	foreach(PATHNODE,pathnode,c->pathnodes) { */
/* 		if (list_empty(pathnode->children)) */
/* 			list_append(pathnode,plist[0]); */
/* 		else */
/* 			list_append(pathnode,plist[1]); */
/* 	} */
/* 	endforeach */

/* 	if (list_empty(plist[0])) */
/* 		error_exit("split_paths: no justification for splitting"); */

/* 	if (list_empty(plist[1])) { */

		/* We can do with a simple relabeling */

/* 		c->cfs_possible=0; */
/* 		ncluster=list_list(c); */

/* 	} */
/* 	else { */

		/* copy rules from c */

/* 		nfeatures=nunary_features; */
/* 		nrule=(RULE **)iprs_calloc(2,sizeof(RULE *)); */
/* 		for (i=0;i<2;i++) { */
/* 			nrule[i]=(RULE *)iprs_calloc(1,sizeof(RULE)); */
/* 			nrule[i]->nfeatures=nfeatures; */
/* 			nrule[i]->important=(int *)iprs_calloc(nfeatures,sizeof(int)); */
/* 			nrule[i]->lower=(float *)iprs_calloc(nfeatures,sizeof(float)); */
/* 			nrule[i]->upper=(float *)iprs_calloc(nfeatures,sizeof(float)); */
/* 			for (feature=0;feature<nfeatures;feature++) { */
/* 				nrule[i]->lower[feature]=c->rule->lower[feature]; */
/* 				nrule[i]->upper[feature]=c->rule->upper[feature]; */
/* 			} */
/* 		} */

		/* Create new clusters */

/* 		for (i=0;i<2;i++) { */
/* 			n[i]=(CLUSTER *)iprs_calloc(1,sizeof(CLUSTER)); */
/* 			n[i]->type=c->type; */
/* 			n[i]->level=c->level; */
/* 			s=(char *)iprs_calloc(strlen(c->name)+10,sizeof(char)); */
/* 			n[i]->name=s; */
/* 			n[i]->parent=c->parent; */
/* 			n[i]->children=NIL; */
/* 			n[i]->rule=nrule[i]; */
/* 			n[i]->elements=NIL; */
/* 			n[i]->pathnodes=plist[i]; */
/* 			n[i]->nclass=0; */
/* 			n[i]->class=(int *)iprs_calloc(nclasses,sizeof(int)); */
/* 			for(j=0;j<nclasses;j++) n[i]->class[j]=0; */
/* 			n[i]->cfs=(FEATURE_LIST *)NULL; */
/* 			n[i]->unresolvable=0; */
/* 		} */

		/* Names: n[0] same as c, n[1] new number */

/* 		strcpy(n[0]->name,c->name); */
/* 		i=list_length(c->parent->children); */
/* 		sprintf(n[1]->name,"%s%c%-d",c->parent->name,c->type,i); */

/*         n[0]->cfs_possible=0; */
/*         if (n[1]->level < max_level) */
/*             n[1]->cfs_possible=1; */
/*         else */
/*             n[1]->cfs_possible=0; */

/* 		ncluster=NIL; */
/* 		for (i=0;i<2;i++) { */

/* 			n[i]->npathnodes=list_length(n[i]->pathnodes); */

			/* elements and class vector */

/* 			foreach(PATHNODE,pathnode,n[i]->pathnodes) { */
/* 				unary=pathnode->unary; */
/* 				list_union(unary,n[i]->elements); */
/* 				(n[i]->class[unary->class])++; */
/* 			} */
/* 			endforeach */

			/* nclass */

/* 			n[i]->nclass=0; */
/* 			for (j=0;j<nclasses;j++) { */
/* 				if (n[i]->class[j]>0) (n[i]->nclass)++; */
/* 			} */

/* 			n[i]->entropy=compute_entropy(n[i]->class); */
/* 			pratio=(float)(n[i]->npathnodes)/ */
/* 				   (float)(n[i]->parent->cfs->npathnodes); */
/* 			n[i]->weighted_entropy=pratio*n[i]->entropy; */

			/* other things */

/* 			list_append(n[i],c->parent->children); */
/* 			if(crg_debug_clusters) */
/* 			{ */
/* 				printf("\nsplit_paths: printing cluster n[%d]", i); */
/* 				print_cluster(n[i]); */
/* 			} */
/* 			list_append(n[i],ncluster); */
/* 		} */

/* 		list_remove(c,c->parent->children); */
/* 		free_cluster_tree(c); */

/* 	} */
/* 	return(ncluster); */
/* } */
