/* ====================================================================
 * Packages:		print,free,varia
 * ==================================================================== */

/* Include files */

#include <iprs.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include "crg.h"

/* Global Declarations */

extern LIST      binary_list;
extern float    *binary_difference_threshold;
extern float    *binary_mean;
extern float    *binary_range;
extern float    *binary_sdev;
extern float     expected_rule_cost;
extern int       max_path_length;
extern int       nbinary_features;
extern int       nclasses;
extern int       nnodes_analyzed;
extern int       nparts;
extern int       npaths;
extern int       npaths_analyzed;
extern int       nunary_features;
extern int       nviews;
extern LIST      part_list;
extern PATHNODE *path_tree;
extern LIST      unary_list;
extern float    *unary_difference_threshold;
extern float    *unary_mean;
extern float    *unary_range;
extern float    *unary_sdev;

/* Parameters */

extern int       breadth_first;
extern int       crg_debug_clusters;
extern int       crg_debug_parts;
extern int       crg_debug_paths;
extern int       crg_debug_varia;
extern float     entropy_threshold;
extern int       entropy_wrt_all_classes;
extern int       feature_ordering;
extern int       fixed_path_length;
extern float     leader_threshold;
extern int       max_level;
extern int       symmetric_relations;
extern int       use_leader_cluster;
extern int       zscore;

/* ====================================================================
 * Package:        print
 * ==================================================================== */

void print_part(PART *part)
{
	int f;

	printf("Part %2d View %2d Class %2d (",
		   part->name,part->view,part->class);
	for (f=0;f<nunary_features;f++) {
		printf(" %f",part->unary->value[f]);
	}
	printf(")\n");
	foreach(BINARY,b,part->binary) {
		printf("   with Part %2d (", b->part2->name);
		for (f=0;f<nbinary_features;f++) {
			printf(" %f",b->value[f]);
		}
		printf(")\n");
	}
	endforeach
}

/* --------------------------------------------------------------------
 * Procedure: output_xgobi
 * Comments:  output unary and binary features to separate files. Mainly
 *            for use with the visualisation tool xgobi.
 * -------------------------------------------------------------------- */

void output_xgobi(char *basename)
{
	int f;
	char filename[100];
	FILE *unaryfile, *binaryfile, *uclassfile, *bclassfile;

	sprintf(filename, "%s_unary.dat", basename);
	unaryfile = fopen(filename, "w");
	sprintf(filename, "%s_unary.glyphs", basename);
	uclassfile = fopen(filename, "w");
	sprintf(filename, "%s_binary.dat", basename);
	binaryfile = fopen(filename, "w");
	sprintf(filename, "%s_binary.glyphs", basename);
	bclassfile = fopen(filename, "w");

	foreach(PART,part,part_list) {
		for (f=0;f<nunary_features;f++) {
			fprintf(unaryfile, "%f ",part->unary->value[f]);
		}
		fprintf(unaryfile, "\n");
		fprintf(uclassfile, "%d\n", part->class);
		foreach(BINARY,b,part->binary) {
			for (f=0;f<nbinary_features;f++) {
				fprintf(binaryfile, "%f ",b->value[f]);
			}
			fprintf(binaryfile, "\n");
			fprintf(bclassfile, "%d\n", part->class);
		}
		endforeach
	}
	endforeach

	fclose(unaryfile);
	fclose(uclassfile);
	fclose(binaryfile);
	fclose(bclassfile);
}

/* --------------------------------------------------------------------
 * Procedure: print_parts
 * Comments:  print part descriptions
 * -------------------------------------------------------------------- */

void print_parts()
{
	int f;

	foreach(PART,part,part_list) {
		printf("Part %2d View %2d Class %2d (",
			   part->name,part->view,part->class);
		for (f=0;f<nunary_features;f++) {
			printf(" %f",part->unary->value[f]);
		}
		printf(")\n");
		foreach(BINARY,b,part->binary) {
			printf("   with Part %2d (", b->part2->name);
			for (f=0;f<nbinary_features;f++) {
				printf(" %f",b->value[f]);
			}
			printf(")\n");
		}
		endforeach
	}
	endforeach
}

/* --------------------------------------------------------------------
 * Procedure: print_cluster
 * Comments:  print cluster description
 * -------------------------------------------------------------------- */

void print_cluster(CLUSTER *c)
{
    int i;
	float r;

    printf("\nCluster %s",c->name);
	if (c->cfs_possible==0) printf(" <not expansible>");
	else if (c->cfs_possible==1) printf(" <expansible>");
	printf("\n");

	if (c->level==0) return;

	printf("Level: %d\n", c->level);
	printf("Entropy: %4.2f\n",c->entropy);

	r=(float)(list_length(c->pathnodes))/
	  (float)(list_length(c->parent->cfs->pathnodes));
	printf("%d paths ( = %5.1f%%)\n",list_length(c->pathnodes),r*100.0);

/* 	printf("Rule:  "); */
/* 	for (i=0;i<c->rule->nfeatures;i++) { */
/* 		printf("[%6f,%6f] ",c->rule->lower[i],c->rule->upper[i]); */
/* 	} */
/* 	printf("\n"); */

	if (crg_debug_parts) {
		printf("Parts: ");
		if (c->type == 'U') {
			int cnt=0;
			foreach(UNARY,u,c->elements) {
				printf("%d ",u->part->name);
				if (cnt++>40)
				{
					printf("\n");
					cnt = 0;
				}
			}
			endforeach
		}
		else {
			int cnt=0;
			foreach(BINARY,b,c->elements) {
				printf("%d-%d ",b->part1->name,b->part2->name);
				if (cnt++>40) 
				{
					printf("\n");
					cnt=0;
				}
			}
			endforeach
		}
		printf("\n");
	}

	printf("Classes: ");
	for (i=0;i<nclasses;i++) {
		if (c->class[i] > 0) printf(" %d", i);
	}
	printf("\n");
}

/* ====================================================================
 * Package:     free
 * Comments:	free all sorts of structures
 * ==================================================================== */

/* --------------------------------------------------------------------
 * Procedure: free_cluster
 * Purpose:   free a cluster data structure
 * -------------------------------------------------------------------- */

void free_cluster(CLUSTER *c)
{
	iprs_free(c->name);
	list_free(c->children);
	list_free(c->elements);
	list_free(c->pathnodes);
/* 	if (c->rule) */
/* 	{ */
/* 		iprs_free(c->rule->lower); */
/* 		iprs_free(c->rule->upper); */
/* 		iprs_free(c->rule); */
/* 	} */
	iprs_free(c->mean);
	iprs_free(c->class);
	if (c->cfs != (FEATURE_LIST *)NULL) {
		list_free(c->cfs->elements);
		iprs_free(c->cfs);
	}
	iprs_free(c);
}

/* ====================================================================
 * Package:     modified IPRS routines
 * Comments:    IPRS is perfect but some of us prefer a less than
 *              perfect world.
 * ==================================================================== */

/* --------------------------------------------------------------------
 * Procedure: free_cluster_tree
 * Purpose:   free a cluster tree
 * -------------------------------------------------------------------- */

void free_cluster_tree(CLUSTER *c)
{
	if (!list_empty(c->children)) {
		foreach(CLUSTER,child,c->children)
			free_cluster_tree(child);
		endforeach
	}
	free_cluster(c);
}

/* --------------------------------------------------------------------
 * Procedure: crg_create_feature_space
 * Purpose:   This procedure is the same as iprs_create_feature_space
 *            except that no space is allocated for the feature value
 *            arrays (they are simply pointers to unary->value and
 *            binary->value, respectively).
 * -------------------------------------------------------------------- */

IPRS_FEATURE_SPACE *crg_create_featurespace(int numf,int nump)
{
	IPRS_FEATURE_SPACE *sptr;

	iprs_debug_enter("crg_create_featurespace");
	sptr=(IPRS_FEATURE_SPACE *)iprs_calloc(1,sizeof(IPRS_FEATURE_SPACE));
	sptr->numfeatures=numf;
	sptr->numpoints=nump;
	sptr->parameter=(char *)NULL;
	if(nump!=0) {
		sptr->point=(float **)iprs_calloc(nump,sizeof(float *));
	} else {
		sptr->point=NULL;
	}
	iprs_debug_exit();
	return(sptr);
}

/* --------------------------------------------------------------------
 * Procedure: crg_free_featurespace
 * Purpose:   same as iprs_free_featurespace but feature value arrays
 *            are not free'd (see above).
 * -------------------------------------------------------------------- */

void crg_free_featurespace(IPRS_FEATURE_SPACE *sptr)
{
	iprs_debug_enter("crg_free_featurespace");
	if(!sptr) {
        fprintf(stderr, "IPRS_NULL_IMAGE in crg_free_featurespace\n");
		return;
    }
	if(sptr->numpoints) {
		iprs_free(sptr->point);
	}
	iprs_free(sptr);
	iprs_debug_exit();
}

/* ====================================================================
 * varia
 * ==================================================================== */

/* --------------------------------------------------------------------
 * Procedure:	memory_used
 * Purpose:     return memory used in bytes
 * Comment:     SGI specific
 * -------------------------------------------------------------------- */

int memory_used()
{
	char path[100];
	pid_t pid;
	struct stat *s;

	s=(struct stat *)iprs_malloc(sizeof(struct stat));

	pid=getpid();
	sprintf(path,"/debug/%-d",pid);
	stat(path,s);
	return((int)(s->st_size));
}

/* --------------------------------------------------------------------
 * Procedure: timer
 * Purpose:   return interval (in secs) since the last call to timer.
 * Comments:  Results for intervals longer than 1 year are incorrect.
 *            (365 days = 31536000 seconds)
 * -------------------------------------------------------------------- */

float timer()
{
        struct itimerval old, new;
        float v;

        new.it_value.tv_sec     = 31536000L;
        new.it_value.tv_usec    =         0;
        new.it_interval.tv_sec  =         0;
        new.it_interval.tv_usec =         0;
        setitimer(ITIMER_VIRTUAL,&new,&old);
        v = (float)(31536000L-old.it_value.tv_sec)-
            (float)old.it_value.tv_usec*1.e-6;
        return v;
}

/* --------------------------------------------------------------------
 * Procedure: process_debug_options
 * Comment:   scan debug options and set global variables
 * -------------------------------------------------------------------- */

void process_debug_options(int n, char **option)
{
	int i;

	crg_debug_clusters=0;
	crg_debug_parts=0;
	crg_debug_paths=0;
	crg_debug_varia=0;

	for (i=0;i<n;i++) {
		if (!strncmp(option[i],"cluster",7))
			crg_debug_clusters=1;
		else if (!strncmp(option[i],"part",4))
			crg_debug_parts=1;
		else if (!strncmp(option[i],"path",4))
			crg_debug_paths=1;
		else if (!strncmp(option[i],"varia",5))
			crg_debug_varia=1;
		else
			error_exit("Unknown debug option");
	}
}

/* --------------------------------------------------------------------
 * Procedure: error_exit
 * Comments:  print error message ON STDOUT and exit
 * -------------------------------------------------------------------- */

void error_exit(char *message)
{
	fprintf(stdout,"%s\n",message);
	exit(1);
}


/* ---------------------------------------------------------------------
 * Procedure: convert_binary_todecimal
 * Comments:
 * --------------------------------------------------------------------- */

int convert_base_todecimal(int *bin, int numdigs, int base)
{
	int ret=0, i;

	for (i=numdigs-1; i>=0; i--)
		ret = bin[i] + ret*base;
	return(ret);
}

/* ---------------------------------------------------------------------
 * Procedure: print_hyper_rectrule
 * Comments: Prints out all vertices of the hyper-rect rule and also
 *           which vertices should be joined. This function is
 *           principally for use fr rule visualisation with xgobi.
 * --------------------------------------------------------------------- */

void print_hyperrect_rule(RULE *rule, int *curr_ind,
						  FILE *ptfile, FILE *lnfile, FILE *labelfile,
						  FILE *glyphfile)
{
	int *dim, *ndim, index, i, j, num;
	float *pts[2]; /* the actual data */
	int totalnums;
	int cnt;

	if (!rule) return;

	pts[0] = rule->lower;
	pts[1] = rule->upper;
	dim = (int*)iprs_calloc(rule->nfeatures, sizeof(int));
	ndim = (int*)iprs_calloc(rule->nfeatures, sizeof(int));
	
	totalnums = 1;
	for (i=0; i<rule->nfeatures; i++)
		totalnums = 2*totalnums;

	cnt=0;
	for (i=0; i<totalnums; i++)
	{
		/* output the current vertex */
		index = i;
		for (j=0; j<rule->nfeatures; j++)
		{
			dim[j] = index%2;
			index = index / 2;
			fprintf(ptfile, "%f ", pts[dim[j]][j]);
		}
		fprintf(ptfile, "\n");
		fprintf(glyphfile, "31\n");
		fprintf(labelfile, "rule_vertex\n");

		/* output which vertices should be joined by lines. */
		/* this is a hypercube type application - pts whose */
		/* index differs by only one bit in their index should */
		/* be joined. */

		for (j=0; j<rule->nfeatures; j++)
		{
			int k;
			memcpy(ndim, dim, rule->nfeatures*sizeof(int));
			if (ndim[j]==0)    /* flip this bit */
			{
				cnt++;
				ndim[j] = 1;
				num = convert_base_todecimal(ndim, rule->nfeatures, 2);
				fprintf(lnfile, "%d %d\n", i+*curr_ind, num+*curr_ind);
			}
		}
	}
	*curr_ind += totalnums;
}


/* ---------------------------------------------------------------------
 * Procedure: output_clustertree_preorder
 * Comments: Performs a preorder traversal of the cluster tree and
 *           prints out all the rules at a given point.
 * --------------------------------------------------------------------- */

void output_clustertree_preorder(CLUSTER *root)
{
	char ptname[40], lnname[40];
	int curr_ind=1, i;
	FILE *ptfile, *lnfile, *labelfile, *glyphfile;

	if (!root->children) return;
	
	sprintf(ptname, "%s.dat", root->name);
	sprintf(lnname, "%s.lines", root->name);
	ptfile = fopen(ptname, "w");
	lnfile = fopen(lnname, "w");
	sprintf(lnname, "%s.rowlab", root->name);
	labelfile = fopen(lnname, "w");
	sprintf(lnname, "%s.glyphs", root->name);
	glyphfile = fopen(lnname, "w");

	/* visit root */
	printf("printing rules for %s\n", root->name);
/* 	print_hyperrect_rule(root->rule, &curr_ind, ptfile, lnfile, labelfile, glyphfile); */
	if (root->type=='U')   /* unary cluster */
	{
		foreach (UNARY, u, root->elements)
		{
			fprintf(glyphfile, "%d\n", u->class*5+1);
			fprintf(labelfile, "%d\n", u->part->name);
			for (i=0; i<nunary_features; i++)
				fprintf(ptfile, "%f ", u->value[i]);
			fprintf(ptfile, "\n");
		}
		endforeach;
	}
	else if (root->type=='B')
	{
		foreach (BINARY, b, root->elements)
		{
			fprintf(glyphfile, "%d\n", b->class*5+1);
			fprintf(labelfile, "%d_%d\n", b->part1->name, b->part2->name);
			for (i=0; i<nbinary_features; i++)
				fprintf(ptfile, "%f ", b->value[i]);
			fprintf(ptfile, "\n");
		}
		endforeach;
	}
			
/* 	foreach (CLUSTER, c, root->children) */
/* 		print_hyperrect_rule(c->rule, &curr_ind, ptfile, lnfile); */
/* 	endforeach; */
	printf("finished printing rules for %s\n", root->name);

	fclose(ptfile);
	fclose(lnfile);
	fclose(labelfile);
	fclose(glyphfile);

	/* recursive call for each child */
	foreach (CLUSTER, c, root->children)
		output_clustertree_preorder(c);
	endforeach;
}

void output_fuzzyclustertree_preorder(CLUSTER *root)
{
	char ptname[40], lnname[40];
	int curr_ind=1, i;
	FILE *ptfile, *lnfile, *labelfile, *glyphfile;
	char name[100];

	if (!root->children) return;
	
	if (root->type=='0') sprintf(name, "root");
	else sprintf(name, "%s", root->name);

	sprintf(ptname, "%s.dat", name);
	sprintf(lnname, "%s.lines", name);
	ptfile = fopen(ptname, "w");
	lnfile = fopen(lnname, "w");
	sprintf(lnname, "%s.rowlab", name);
	labelfile = fopen(lnname, "w");
	sprintf(lnname, "%s.glyphs", name);
	glyphfile = fopen(lnname, "w");

	/* visit root */
	printf("printing rules for %s\n", name);
		
	if ((root->type=='B')||(root->type=='0'))   /* unary cluster */
	{
		int cluster_number=0;
		foreach (CLUSTER, c, root->children)
		{
			fprintf(glyphfile, "%d\n", 5);
			fprintf(labelfile, "centre %d\n", cluster_number++);
			for (i=0; i<nunary_features; i++)
				fprintf(ptfile, "%f ", c->mean[i]);
			fprintf(ptfile,"\n");
			foreach (UNARY, u, c->elements)
			{
				fprintf(glyphfile, "%d\n", u->class*5+1);
				fprintf(labelfile, "%d\n", u->part->name);
				for (i=0; i<nunary_features; i++)
					fprintf(ptfile, "%f ", u->value[i]);
				fprintf(ptfile, "\n");
			}
			endforeach;
		}
		endforeach;
	}
	else if (root->type=='U')
	{
		int cluster_number=0;
		foreach (CLUSTER, c, root->children)
		{
			fprintf(glyphfile, "%d\n", 5);
			fprintf(labelfile, "centre %d\n", cluster_number++);
			for (i=0; i<nbinary_features; i++)
				fprintf(ptfile, "%f ", c->mean[i]);
			fprintf(ptfile,"\n");
			foreach (BINARY, b, c->elements)
			{
				fprintf(glyphfile, "%d\n", b->class*5+1);
				fprintf(labelfile, "%d_%d\n", b->part1->name, b->part2->name);
				for (i=0; i<nbinary_features; i++)
					fprintf(ptfile, "%f ", b->value[i]);
				fprintf(ptfile, "\n");
			}
			endforeach;
		}
		endforeach;
	}
			
/* 	foreach (CLUSTER, c, root->children) */
/* 		print_hyperrect_rule(c->rule, &curr_ind, ptfile, lnfile); */
/* 	endforeach; */
	printf("finished printing rules for %s\n", name);

	fclose(ptfile);
	fclose(lnfile);
	fclose(labelfile);
	fclose(glyphfile);

	/* recursive call for each child */
	foreach (CLUSTER, c, root->children)
		output_fuzzyclustertree_preorder(c);
	endforeach;
}
