/* ====================================================================
 * Program:     sure: scene understanding using rule evaluation
 * Author:      Walter F. Bischof
 * ==================================================================== */

/* Include files */

#include "sure.h"

/* Global Declarations (described in sureextern.h) */

CLUSTER *cluster_tree;
int      max_level;
int      nbinary_features;
int      nclasses;
int      nunary_features;
SCENE   *scene;
int      symmetric_relations;
int      classification_known;
LIST     part_list_training;
PART   **training_parts;
int      nparts;

/* Parameters (described in sureextern.h) */

char	*classification_file;
int      runtime_max_level;
char    *scene_filename;
char    *training_filename;
int      verbosity;
int      use_intersnake_compatibility;
int      use_permutation_constraint;
int      use_relaxation_labeling;
int      use_single_classification_constraint;

/*
 * The leaveout flags are used for removing a particular feature from
 * the input feature spaces. This is useful for testing the performance
 * of the classifier without having to re-extract the features from scratch.
 */

int       nunary_leaveout=0, unary_leaveout[40];
int       nbinary_leaveout=0, binary_leaveout[40];

/* Mean and standard dev for features. Must transform scene values */
/* according to the zscore trans.                                  */

float *unary_mean, *unary_sdev;
float *binary_mean, *binary_sdev;

/* fuzzy clustering parameters */

float fuzzyq=2.0;
int fuzzy_nclusters=2;
float cutoff_snake_percentage=0.1;
float minimum_fuzzy_membership=0.15;
int class_number=-1;

/* hypothesis verification parameters */
int perform_hypothesis_verification=0;
char *view_file="intens_files";
char *reg_file="regs_file";
char *part_file="parts_file";

/* --------------------------------------------------------------------
 * Procedure: Main
 * Usage:     sure [switches] filename
 * -------------------------------------------------------------------- */

int main(int argc, char **argv)
{
	timer();

	scene=(SCENE *)iprs_calloc(1, sizeof(SCENE));
	process_options(argc,argv);
	read_training();
	read_scene();
	if (classification_known) read_classification();
	if (verbosity>1) print_cluster_tree(cluster_tree);
	classify_scene();
	if (class_number>=0)
		generate_classified_regions(scene->parts, scene->regions, class_number);

	printf("Time                       %-3.1f secs\n",timer());
	printf("Memory                     %-3.1f MB\n",memory_used()/1048576.);

	return 0;
}

/* --------------------------------------------------------------------
 * Procedure: process_options
 * Comment:   set default values of all parameters and process options
 * -------------------------------------------------------------------- */

void process_options(int argc, char **argv)
{
    int opt;
    extern char *optarg;
    extern int optind;
	int leaveout_feat;

    /* Default values for parameters */

    runtime_max_level = -1;
    scene_filename = "scene.data";
    training_filename = "training.result";
    classification_known=0;
    use_intersnake_compatibility=1;
    use_permutation_constraint=1;
    use_relaxation_labeling=1;
    use_single_classification_constraint=1;
	verbosity=0;

	use_intersnake_compatibility=0;
	use_permutation_constraint=0;
	use_single_classification_constraint=0;
    while ( (opt=getopt(argc,argv,"a:b:c:l:m:rt:v:%:o:e:x:y:hi:s:p:"))!= EOF) {
        switch (opt) {
		case 'c':
			classification_known=1;
			classification_file=optarg;
			break;
		case 'l':
			runtime_max_level=(int)atoi(optarg);
			break;
		case 'r':
			use_relaxation_labeling=0;
			break;
		case 't':
			training_filename=optarg;
			break;
        case 'v':
            verbosity = atoi(optarg);
            break;
		case '%':
			cutoff_snake_percentage = atof(optarg);
			break;
		case 'o':
			class_number = atoi(optarg);
			break;
		case 'e':
			scene->regions = iprs_load_regionimage(optarg);
			break;
		case 'x':
			unary_leaveout[nunary_leaveout++] = atoi(optarg);
			break;
		case 'y':
			binary_leaveout[nbinary_leaveout++] = atoi(optarg);
			break;
		case 'h':
			perform_hypothesis_verification = 1;
			break;
		case 'i':
			view_file = optarg;
			break;
		case 's':
			reg_file = optarg;
			break;
		case 'p':
			part_file = optarg;
			break;
		case 'a':
		{
			IPRS_IMAGE *input, *int_input;
			input = iprs_load_image(optarg);
			int_input = iprs_imagetoint(input);
			iprs_free_image(input);
			scene->intensity = (IPRS_INT_IMAGE*)int_input->image[0];
			break;
		}
		case 'm':
		{
			minimum_fuzzy_membership = (float)atof(optarg);
			break;
		}
		case '?':
            usage();
            break;
        }
    }

    /* assign normal arguments */

    if (optind < argc) {
        scene_filename = argv[optind];
    }
	/* remove background from intensity image */
	if ((scene->regions)&&(scene->intensity))
	{
		IPRS_INT_IMAGE *tmp;
		int width, height, i, j;
		width = iprs_getwidth_intimage(scene->intensity);
		height = iprs_getheight_intimage(scene->intensity);
		tmp = iprs_create_2dintimage(width, height);
		bren_highlight_regionimage(scene->regions, tmp);
		for (i=0; i<width; i++)
		{
			for (j=0; j<height; j++)
			{
				if (tmp->image[i+j*width]) 
					tmp->image[i+j*width] = scene->intensity->image[i+j*width];
			}
		}
		iprs_free_intimage(scene->intensity);
		scene->intensity = tmp;
	}
}

/* --------------------------------------------------------------------
 * Procedure: usage
 * Comment:   describe usage
 * -------------------------------------------------------------------- */

void usage()
{
    char *help =
    "sure [switches] filename\n"
    "-c file   file containing true region classifications\n"
    "-i        do not use intersnake compatibility\n"
    "-l level  expand CRG rules to maximal depth level\n"
    "-p        do not use snake permutation constraint\n"
    "-r        do not use intrasnake relaxation constraint\n"
    "-s        do not use single classification constraint\n"
    "-t file   training file (default=training.result)\n"
    "-v n      verbosity level (default=0)\n"
    "filename  scene description (default=scene.data)\n";

    error_exit(help);
}

/* --------------------------------------------------------------------
 * Procedure: read_training
 * Comment:   read training results from file
 * -------------------------------------------------------------------- */

void bren_reverse_path(PATH *path)
{
	PART **part;
	UNARY **unary;
	BINARY **binary;
	int i;

	part = (PART**)iprs_calloc(path->length, sizeof(PART*));
	unary = (UNARY**)iprs_calloc(path->length, sizeof(UNARY*));
	binary = (BINARY**)iprs_calloc(path->length, sizeof(BINARY*));
	for (i=path->length-1; i>=0; i--)
	{
		part[i] = path->part[path->length-1-i];
		if (path->unary) unary[i] = path->unary[i];
		if (path->binary) binary[i] = path->binary[i];
	}
	iprs_free(path->part);
	iprs_free(path->unary);
	iprs_free(path->binary);
	path->part = part;
	path->unary = unary;
	path->binary = binary;
}

void read_training()
{
	int i,j,n,ic,nc,ip,nviews,ntotal,plength;
	char datafile[100],*cptr,header[10];
	PART **part;
	PATH *path;
	RULE *rule;
	CLUSTER **cluster;
	FILE *fp;
	int feature_ordering;
	int *num_unary_parts, *num_binary_relations;

	if ((fp=fopen(training_filename,"r")) == (FILE *)NULL)
	{
		printf("filename is %s\n", training_filename);
		error_exit("read_training: cannot open file");
	}

	/* file header */

/* 	fgets(header,10,fp); */
/* 	if (strcmp(header,"CRG2.0\n") != 0) */
/* 		error_exit("read_training: incorrect header training file"); */

	/* load global values */

	fgets(datafile,100,fp);
	fread(&nunary_features,sizeof(int),1,fp);
	fread(&nbinary_features,sizeof(int),1,fp);
	fread(&nclasses,sizeof(int),1,fp);
	fread(&nviews,sizeof(int),1,fp);
	fread(&nparts,sizeof(int),1,fp);
	fread(&feature_ordering,sizeof(int),1,fp);
	fread(&symmetric_relations,sizeof(int),1,fp);
	fread(&max_level,sizeof(int),1,fp);

	if (runtime_max_level < 0) runtime_max_level=max_level;

	/* Mod by B. McCane to read zscore transform params */
	unary_mean = (float*)iprs_calloc(nunary_features, sizeof(float));
	unary_sdev = (float*)iprs_calloc(nunary_features, sizeof(float));
	binary_mean = (float*)iprs_calloc(nbinary_features, sizeof(float));
	binary_sdev = (float*)iprs_calloc(nbinary_features, sizeof(float));
	fread(unary_mean, sizeof(float), nunary_features, fp);
	fread(unary_sdev, sizeof(float), nunary_features, fp);
	fread(binary_mean, sizeof(float), nbinary_features, fp);
	fread(binary_sdev, sizeof(float), nbinary_features, fp);

	if (verbosity>0)
	{
		int i;

		printf("unary_mean = [ ");
		for (i=0; i<nunary_features; i++)
			printf("%f ", unary_mean[i]);
		printf("]\nunary sdev = [ ");
		for (i=0; i<nunary_features; i++)
			printf("%f ", unary_sdev[i]);

		printf("]\nbinary_mean = [ ");
		for (i=0; i<nbinary_features; i++)
			printf("%f ", binary_mean[i]);
		printf("\nbinary sdev = [ ");
		for (i=0; i<nbinary_features; i++)
			printf("%f ", binary_sdev[i]);
		printf("]\n");
	}

	printf("Training data file: %s",datafile);

	/* load part descriptions */

	num_unary_parts = (int*)iprs_calloc(nclasses, sizeof(int));
	num_binary_relations = (int*)iprs_calloc(nclasses, sizeof(int));
	part_list_training=NIL;
	part=(PART **)iprs_calloc(nparts,sizeof(PART *));
	fread(&nparts,sizeof(int),1,fp);
	for (i=0;i<nparts;i++) {
		int numbinary;
		part[i]=(PART *)iprs_calloc(1,sizeof(PART));
		fread(&(part[i]->name),sizeof(int),1,fp);
		fread(&(part[i]->view),sizeof(int),1,fp);
		fread(&(part[i]->class),sizeof(int),1,fp);
		fread(&numbinary, sizeof(int), 1, fp);
		list_append(part[i],part_list_training);

		num_unary_parts[part[i]->class]++;
		num_binary_relations[part[i]->class] += numbinary;
	}
	training_parts = part;

	fread(&nc,sizeof(int),1,fp);
	cluster=(CLUSTER **)iprs_calloc(nc,sizeof(CLUSTER *));
	for (ic=0;ic<nc;ic++) {
		cluster[ic]=(CLUSTER *)iprs_calloc(1,sizeof(CLUSTER));
		cluster[ic]->name=(char *)iprs_malloc(100);
		fread(&(cluster[ic]->id),sizeof(int),1,fp);
		fread(&(cluster[ic]->level),sizeof(int),1,fp);
		fgets(cluster[ic]->name,100,fp);
		cptr=cluster[ic]->name+strlen(cluster[ic]->name)-1;  /* remove \n */
		*cptr='\0';

		fread(&(cluster[ic]->cfs_possible),sizeof(int),1,fp);
		cluster[ic]->children=NIL;
		if (cluster[ic]->level > 0) {
			if ((cluster[ic]->level)%2 == 1)
				cluster[ic]->type='U';
			else
				cluster[ic]->type='B';
			fread(&ip,sizeof(int),1,fp);
			cluster[ic]->parent=cluster[ip];
			list_append(cluster[ic],cluster[ip]->children);

			/* paths */

			/*
			 * for some reason paths are in reverse order, so reverse them.
			 */
			cluster[ic]->paths=NIL;
			fread(&n,sizeof(int),1,fp);
			for (i=0;i<n;i++) {
				path=(PATH *)iprs_calloc(1,sizeof(PATH));
				fread(&plength,sizeof(int),1,fp);
				path->length=plength;
				path->part=(PART **)iprs_calloc(plength,sizeof(PART *));
				for (j=0;j<plength;j++) {
					fread(&ip,sizeof(int),1,fp);
					path->part[j]=part[ip];
				}
				bren_reverse_path(path);
				list_append(path,cluster[ic]->paths);
			}

			/* rule */

/* 			rule=(RULE *)iprs_calloc(1,sizeof(RULE)); */
/* please note that the following lines need to be deleted */
/* 			fread(&n,sizeof(int),1,fp); */
/* 			rule->nfeatures=n; */
/* 			rule->important=(int *)iprs_calloc(n,sizeof(int)); */
/* 			rule->lower=(float *)iprs_calloc(n,sizeof(float)); */
/* 			rule->upper=(float *)iprs_calloc(n,sizeof(float)); */
/* 			cluster[ic]->rule=rule; */
/* 			fread(rule->important,sizeof(int),n,fp); */
/* 			fread(rule->lower,sizeof(float),n,fp); */
/* 			fread(rule->upper,sizeof(float),n,fp); */
/* up till here */

			/* class */

			cluster[ic]->class=(int *)iprs_calloc(nclasses,sizeof(int));
			cluster[ic]->cvector=(float *)iprs_calloc(nclasses,sizeof(float));
			fread(cluster[ic]->class,sizeof(int),nclasses,fp);
			cluster[ic]->nclass=0;
			ntotal=0;
			for (i=0;i<nclasses;i++) {
				if (cluster[ic]->class[i] > 0) {
					(cluster[ic]->nclass)++;
					ntotal+=cluster[ic]->class[i];
				}
			}
			if (ntotal>0)
			{
				float total=0.0;
				for (i=0;i<nclasses;i++) {
					cluster[ic]->cvector[i]=
						(float)(cluster[ic]->class[i])/(float)ntotal;
					/* normalise weights by number of parts in class */
					if (cluster[ic]->type=='U')
						cluster[ic]->cvector[i] /= (float)num_unary_parts[i];
					else
						cluster[ic]->cvector[i]/=(float)num_binary_relations[i];
					total += cluster[ic]->cvector[i];
				}
				/* normalise to 1 */
				for (i=0;i<nclasses;i++) {
					cluster[ic]->cvector[i] /= total;
				}
			}

			/* cluster mean */
			if (cluster[ic]->type=='U') n = nunary_features;
			else n = nbinary_features;
			cluster[ic]->mean = (float*)iprs_calloc(n, sizeof(float));
			fread(cluster[ic]->mean, sizeof(float), n, fp);
		}
	}
	cluster_tree=cluster[0];

    /* file trailer */

/*     fgets(header,10,fp); */
/*     if (strcmp(header,"CRG2.0\n") != 0) */
/*         error_exit("read_training: incorrect trailer training file"); */

	/* close */

	fclose(fp);
}

/* --------------------------------------------------------------------
 * Procedure: read_scene
 * Comments:  read a single scene description
 * -------------------------------------------------------------------- */

void read_scene()
{
	IPRS_IMAGE *db,*cptr,*vptr;
	IPRS_FEATURE_SPACE *fs0ptr,*fs1ptr,*fs2ptr;
	int c,v,i,j,n,p1,p2,nparts;
	PART **part;
	BINARY *binary;
	UNARY *unary;
	LIST plist=NULL;
	int scene_nunary_features, scene_nbinary_features, tmp;
	int *binary_usefeatures, *unary_usefeatures;

	db=iprs_load_image(scene_filename);

	printf("Scene data file: %s\n",scene_filename);

	/* nunary_features and nbinary_features */

	nparts=0;
	if (db->numimages > 1)
		error_exit("Cannot analyze multiple scenes");
	cptr=(IPRS_IMAGE *)(db->image[0]);
	vptr=(IPRS_IMAGE *)(cptr->image[0]);
	fs0ptr=(IPRS_FEATURE_SPACE *)(vptr->image[0]);
	fs1ptr=(IPRS_FEATURE_SPACE *)(vptr->image[1]);

    scene_nunary_features = fs0ptr->numfeatures;
    scene_nbinary_features = fs1ptr->numfeatures;
	
	unary_usefeatures = (int*)iprs_calloc(scene_nunary_features, sizeof(int));
	for (i=0; i<nunary_leaveout; i++)
	{
		if (unary_leaveout[i]<scene_nunary_features) 
			unary_usefeatures[unary_leaveout[i]]=1;
		else fprintf(stderr, "Not that many unary features to leave out\n");
	}
	tmp = scene_nunary_features;
	for (i=0; i<tmp; i++)
		if (unary_usefeatures[i]) scene_nunary_features--;
	binary_usefeatures = (int*)iprs_calloc(scene_nbinary_features, sizeof(int));
	for (i=0; i<nbinary_leaveout; i++)
	{
		if (binary_leaveout[i]<scene_nbinary_features) 
			binary_usefeatures[binary_leaveout[i]]=1;
		else fprintf(stderr, "Not that many binary features to leave out\n");
	}
	tmp = scene_nbinary_features;
	for (i=0; i<tmp; i++)
		if (binary_usefeatures[i]) scene_nbinary_features--;

	if (nunary_features != scene_nunary_features)
		error_exit("number of unary features different from training");
	if (nbinary_features != scene_nbinary_features)
		error_exit("number of binary features different from training");

	/* create part, unary and binary descriptors */

	for(c=0;c<db->numimages;c++) {
		cptr=(IPRS_IMAGE *)(db->image[c]);
		for(v=0;v<cptr->numimages;v++) {
			vptr=(IPRS_IMAGE *)(cptr->image[v]);
			fs0ptr=bren_removefeatures_featurespace(
					   (IPRS_FEATURE_SPACE *)(vptr->image[0]),
					   unary_usefeatures, nunary_features);
			fs1ptr=bren_removefeatures_featurespace(
                       (IPRS_FEATURE_SPACE *)(vptr->image[1]),
					   binary_usefeatures, nbinary_features);
			fs2ptr=(IPRS_FEATURE_SPACE *)(vptr->image[2]);

			plist=NIL;

			/* part */

			n=fs0ptr->numpoints;
			part=(PART **)iprs_calloc(n,sizeof(PART *));
			for (i=0;i<n;i++) {
				part[i]=(PART *)iprs_calloc(1,sizeof(PART));
				part[i]->name=i+nparts;
				part[i]->view=v;
				part[i]->class=c;
				part[i]->binary=NIL;
				part[i]->snakes=NIL;
				part[i]->cvector=(float *)iprs_calloc(nclasses,sizeof(float));
				part[i]->cvector1=(float *)iprs_calloc(nclasses,sizeof(float));
				part[i]->modelparts=NIL;
				part[i]->modelparts_defined=0;
				part[i]->true_classification=-1;
				list_append(part[i],plist);
			}

			/* unary */

			for(i=0;i<n;i++) {
				unary=(UNARY *)iprs_calloc(1,sizeof(UNARY));
				unary->part=part[i];
				unary->view=v;
				unary->class=c;
				unary->value=fs0ptr->point[i];
				for (j=0;j<nunary_features;j++)
					if (unary_sdev[j]!=0.0)
					{
						unary->value[j] = (unary->value[j]-unary_mean[j])
							/unary_sdev[j];
					}

				part[i]->unary=unary;
			}

			/* binary */

			check_labels(fs2ptr->point,fs2ptr->numpoints,fs1ptr->numpoints,
						 n,c,v);

			for (i=0;i<fs1ptr->numpoints;i++) {
				p1=(int)(fs2ptr->point[i][0]);
				p2=(int)(fs2ptr->point[i][1]);
				binary=(BINARY *)iprs_calloc(1,sizeof(BINARY));
				binary->part1=part[p1];
				binary->part2=part[p2];
				binary->view=v;
				binary->class=c;
				binary->value=fs1ptr->point[i];
				for (j=0;j<nbinary_features;j++)
					if (binary_sdev[j]!=0.0)
					{
						binary->value[j] = (binary->value[j]-binary_mean[j])
							/binary_sdev[j];
					}

				list_append(binary,part[p1]->binary);

				if (symmetric_relations) {
					binary=(BINARY *)iprs_calloc(1,sizeof(BINARY));
					binary->part1=part[p2];
					binary->part2=part[p1];
					binary->view=v;
					binary->class=c;
					binary->value=fs1ptr->point[i];
					list_append(binary,part[p2]->binary);
				}
			}

/* 			scene=(SCENE *)iprs_malloc(sizeof(SCENE)); */
			scene->parts=plist;

			iprs_free(part);
		}
	}

	/* Check for isolated parts */

	n=0;
	foreach(PART,p,plist) {
		if (list_empty(p->binary)) {
			n++;
			if (verbosity > 0) {
				printf("Part %d isolated\n",p->name);
			}
		}
	}
	endforeach
	if (n>0) {
		fprintf(stdout,
				"Warning: %d out of %d parts are isolated\n"
				"Conditional rule instantiation severely limited\n",
				n,nparts);
	}
}

/* --------------------------------------------------------------------
 * Procedure: read_classification
 * Comments:  
 * -------------------------------------------------------------------- */

void read_classification()
{
	int i,j;
	char line[256];

	FILE *file;

	if ((file=fopen(classification_file,"r"))==(FILE *)NULL) {
		error_exit("cannot open classification file");
	}

	/* read region-class pairs, skipping comments */

	while (1) {
		if (fgets(line,256,file)==NULL) break;
		if (line[0]=='#') continue;
		if (sscanf(line,"%d %d",&i,&j)<2) continue;
		foreach(PART,part,scene->parts) {
			if (part->name == i) {
				part->true_classification=j;
				break;
			}
		}
		endforeach
	}

	/* check that every part is classified */

	foreach(PART,part,scene->parts) {
		if (part->true_classification == -1) {
			error_exit("read_classification: classification incomplete");
		}
	}
	endforeach
}

/* --------------------------------------------------------------------
 * Procedure: check_labels
 * Comments:  check that label information in data file is correct and
 *            meets the assumptions made in crg.
 * -------------------------------------------------------------------- */

void check_labels(float **label, int nlabel, int nbinary,
                  int npart, int class, int view)
{
    int i,j,found;

    /* number of binary and label must match */

    if (nlabel != nbinary)
        error_exit("read_database: label and binary do not match");

    /* label values must match parts in view */

    for (i=0;i<nlabel;i++) {
        if (label[i][0] < 0 || label[i][0] >= npart ||
            label[i][1] < 0 || label[i][1] >= npart) {
            printf("read_database: label information incorrect\n");
            printf("Class: %d View: %d Label { %d %d }\n",
                   class,view,(int)label[i][0],(int)label[i][1]);
            exit(1);
        }
    }

    /* inverse relations must be given */

    if (!symmetric_relations) {
        for (i=0;i<nlabel;i++) {
            found=0;
            for (j=0;j<nlabel;j++) {
                if (label[i][0]==label[j][1] && label[i][1]==label[j][0]) {
                    found=1;
                    break;
                }
            }
            if (!found) {
                printf("read_database: inverse relation missing\n");
				printf("Class: %d View: %d Label { %d %d }\n",
					   class,view,(int)label[i][0],(int)label[i][1]);
				exit(1);
			}

        }
    }
}


void generate_classified_regions(LIST parts, IPRS_REGION_IMAGE *regs,
								 int class_number)
{
	IPRS_INT_IMAGE *image;
	int index=0, i, num_correct=0, num_orig=0, max=0;
	int *ranks, part_num=0;
	IPRS_EDGE_IMAGE *edges;
	FILE *res, *orig;

	if (!regs)
	{
		fprintf(stderr, "A region image must be specified to generate classification output\n");
		return;
	}
	image = iprs_create_2dintimage(256, 256);
	ranks = (int*)iprs_calloc(nclasses, sizeof(int));
	foreach (PART, part, parts)
	{
		if (part->class==class_number)  /* part was classified correctly */
		{
			bren_highlight_region(regs->region[index], image, 2);
			num_correct++;
		}
		else
		{
			bren_highlight_region(regs->region[index], image, 1);
		}

		/* test original classif before hypothesis verification */
		/* find out where the correct class ranked */
		ranks[index]=0;
		for (i=0; i<nclasses; i++)
		{
			if (part->cvector[i]>part->cvector[max])
				max = i;
			if (part->cvector[i]>=part->cvector[class_number])
				ranks[index]++;
		}
		if (max==class_number)
			num_orig++;

		index++;
	}
	endforeach;

	res = fopen("recog_results.txt", "w");
	fprintf(res, "%d %d\n", num_correct, index);
	fclose(res);
	orig = fopen("sure_recog.txt", "w");
	fprintf(orig, "%d %d\n", num_orig, index);
	index=0;
	foreach (PART, part, parts)
	{
		fprintf(orig, "rank[%d] = %d\n", index, ranks[index++]);
	}
	endforeach;
	fclose(orig);
	edges = bren_regionimage_bdry(regs);
	for (i=0; i<edges->numedges; i++)
		bren_highlight_edge(edges->edge[i], image, 3);

/* 	make_intgif(image, "classif.gif"); */
	iprs_free(ranks);
	iprs_free_intimage(image);
	iprs_free_edgeimage(edges);
}
