/* ====================================================================
 * Package:		inter-snake compatibility
 * ==================================================================== */

/* Include files */

#include <iprs.h>
#include "sure.h"
#include "sureextern.h"

/* --------------------------------------------------------------------
 * Procedure: intersnake_compatibility
 * Comments:  compute average (neighbourhood) compatibility of each snake
 * -------------------------------------------------------------------- */

void intersnake_compatibility()
{
	int i,n;
	float w,total;
	PATH *path;
	LIST slist;

	foreach(PART,part,scene->parts) {
		foreach(SNAKE,snake1,part->snakes) {

			if (verbosity > 0) {
				printf("\nAnalysis of part %d, snake ",part->name);
				print_path(snake1->path);
				printf("\n");
			}

			/* get all snakes intersecting this snake */

			slist=NIL;
			path=snake1->path;
			for (i=0;i<path->length;i++) {
				foreach(SNAKE,snake2,path->part[i]->touched_snakes) {
					if (snake1 != snake2) {
						snake_union(snake2,slist);
					}
				}
				endforeach
			}

			/* compute compatibility of snake with all  */
			/* intersecting ones                        */

			total=0.0;
			n=list_length(slist);
			foreach(SNAKE,snake2,slist) {
				w=snake_compatibility(snake1,snake2);
				total += w;
			}
			endforeach
			snake1->compatibility=total/n;
		}
		endforeach
	}
	endforeach

	if (verbosity > 0) {
		foreach(PART,part,scene->parts) {
			foreach(SNAKE,snake,part->snakes) {
				print_path(snake->path);
				printf(" = %f\n",snake->compatibility);
			}
			endforeach
		}
		endforeach
	}
}

/* --------------------------------------------------------------------
 * Procedure: compute_touched_snakes
 * Comments:  compute list of all snakes that start/pass through a part
 * -------------------------------------------------------------------- */

void compute_touched_snakes(PART *part)
{
	int i;

	foreach(SNAKE,snake,part->snakes) {
		for (i=0;i<snake->path->length;i++) {
			snake_union(snake,snake->path->part[i]->touched_snakes);
		}
	}
	endforeach
}

/* --------------------------------------------------------------------
 * Procedure: snake_compatibility
 * Comments:  compute compatibility between two snakes:
 *            For all parts in intersection of the snakes
 *            #(intersection(modelparts(part|snake1),
 *                           modelparts(part,snake2)))
 *            /
 *            #(union(modelparts(part|snake1),
 *                    modelparts(part,snake2)))
 * -------------------------------------------------------------------- */

float snake_compatibility(SNAKE *snake1, SNAKE *snake2)
{
	int n;
	float total;
	LIST plist;

	plist=snake_intersection(snake1,snake2);
	if ((n=list_length(plist))==0) {
		error_exit("snake_compatibility: zero intersection");
	}
	total=0.0;
	foreach(PART,part,plist) {
		total += part_snake_compatibility(part,snake1,snake2);
	}
	endforeach
	total /= (float)n;

	if (total > 0.0 && verbosity>0) {
		printf("Comp [ ");
		print_path(snake1->path);
		printf(" , ");
		print_path(snake2->path);
		printf(" ] = %f\n",total);
	}

	return total;
}

/* --------------------------------------------------------------------
 * Procedure: snake_intersection
 * Comments:  set of parts common to two snakes
 * -------------------------------------------------------------------- */

LIST snake_intersection(SNAKE *snake1,SNAKE *snake2)
{
	int i,j;
	PATH *path1,*path2;
	LIST plist;

	plist=NIL;
	path1=snake1->path;
	path2=snake2->path;

	for (i=0;i<path1->length;i++) {
		for (j=0;j<path2->length;j++) {
			if (path1->part[i] == path2->part[j]) {
				list_append(path1->part[i],plist);
			}
		}
	}

	return plist;
}

/* --------------------------------------------------------------------
 * Procedure: part_snake_compatibility
 * Comments:  see snake_compatibility
 * -------------------------------------------------------------------- */

float part_snake_compatibility(PART *part,SNAKE *snake1,SNAKE *snake2)
{
	int i1,i2,ni,nu;
	float c;
	LIST intersecting;

	for (i1=0;i1<snake1->path->length;i1++) {
		if (part == snake1->path->part[i1]) break;
	}
	if (i1 >= snake1->path->length) {
		error_exit("snake_part_compatibility: empty intersection");
	}

	for (i2=0;i2<snake2->path->length;i2++) {
		if (part == snake2->path->part[i2]) break;
	}
	if (i2 >= snake2->path->length) {
		error_exit("snake_part_compatibility: empty intersection");
	}
	
	intersecting = list_intersection(snake1->modelparts[i1],
									 snake2->modelparts[i2]);
	ni=list_length(intersecting);
	list_free(intersecting);
	nu=list_length(snake1->modelparts[i1]);
	foreach(PART,modelpart,snake2->modelparts[i2]) {
		if (!list_member(modelpart,snake1->modelparts[i1])) nu++;
	}
	endforeach

	c=(float)ni/(float)nu;
	return c;
}
