/* ====================================================================
 * Package:		snake permutation analysis
 * ==================================================================== */

/* Include files */

#include <iprs.h>
#include "sure.h"
#include "sureextern.h"

/* --------------------------------------------------------------------
 * Procedure: snake_permutation_constraint
 * Comments:  
 * -------------------------------------------------------------------- */

void snake_permutation_constraint()
{
	int change;

	if (verbosity>1) printf("\nPermutation constraint\n");

	while (1) {
		change=0;
		foreach(PART,part,scene->parts) {
			foreach(SNAKE,snake1,part->snakes) {
				foreach(SNAKE,snake2,snake1->permutations) {
					if (intersect_model_parts(snake1,snake2))
						change=1;
				}
				endforeach
			}
			endforeach
		}
		endforeach
		if (!change) break;
	}

	if (verbosity>0) {
		printf("\nModel parts for all snakes after constraint analysis\n");
		print_model_parts();
	}

	delete_inconsistent_snakes();

	foreach(PART,part,scene->parts)
		foreach(SNAKE,snake,part->snakes)
			snake_classification_from_modelparts(snake);
		endforeach
	endforeach
}

/* --------------------------------------------------------------------
 * Procedure: compute_model_parts
 * Comments:  
 * -------------------------------------------------------------------- */

void compute_model_parts()
{
	int i;

	foreach(PART,part,scene->parts)
		foreach(SNAKE,snake,part->snakes)
			foreach(CLASSIFICATION,classification,snake->classifications)
				foreach(PATH,modelpath,classification->cluster->paths) {
					if (snake->path->length != modelpath->length) {
						printf("Oops! bug in compute_model_parts\n");
						printf("snake pl=%d, model pl=%d\n", snake->path->length, modelpath->length);
						exit(1);
					}
					for (i=0;i<modelpath->length;i++) {
						list_union(modelpath->part[i],snake->modelparts[i]);
					}
				}
				endforeach
			endforeach
		endforeach
	endforeach
}

/* --------------------------------------------------------------------
 * Procedure: intersect_model_parts
 * Comments:  
 * -------------------------------------------------------------------- */

int intersect_model_parts(SNAKE *snake1, SNAKE *snake2)
{
	int change,i,j,m,m1,m2;
	LIST new;

	change=0;
	for (i=0; i<snake1->path->length; i++) {
		for (j=0; j<snake2->path->length; j++) {
			if ((snake1->path->part[i] == snake2->path->part[j]) &&
				(snake1->modelparts[i] != snake2->modelparts[j])) {
				new=list_intersection(snake1->modelparts[i],
									  snake2->modelparts[j]);
				m=list_length(new);
				m1=list_length(snake1->modelparts[i]);
				m2=list_length(snake2->modelparts[j]);
				if (m<m1 || m<m2) {
					snake1->modelparts[i]=new;
					snake2->modelparts[j]=new;
					change=1;
				}
			}
		}
	}

	if (verbosity>1) {
		if (change) {
			printf("Intersecting ");
			print_path(snake1->path);
			printf(" with ");
			print_path(snake2->path);
			printf("\n");
		}
	}

	return change;
}

/* --------------------------------------------------------------------
 * Procedure: compute_permutation_snakes
 * Comments:  
 * -------------------------------------------------------------------- */

void compute_permutation_snakes()
{
	int i;
	PART *part2;

	foreach(PART,part1,scene->parts) {
		foreach(SNAKE,snake1,part1->snakes) {
			snake1->permutations=NIL;
			for (i=1;i<snake1->path->length;i++) {
				part2=snake1->path->part[i];
				foreach(SNAKE,snake2,part2->snakes) {
					if (path_permutation(snake1->path,snake2->path)) {
						list_append(snake2,snake1->permutations);
					}
				}
				endforeach
			}
		}
		endforeach
	}
	endforeach
}

/* --------------------------------------------------------------------
 * Procedure: delete_inconsistent_snakes
 * Comments:  delete snakes with empty set of model parts after
 *            apply snake permutation constraint
 * -------------------------------------------------------------------- */

void delete_inconsistent_snakes()
{
	int i;
	LIST delete_list;

	foreach(PART,part,scene->parts) {
		delete_list=NIL;
		foreach(SNAKE,snake,part->snakes) {
			for (i=0;i<snake->path->length;i++) {
				if (list_empty(snake->modelparts[i])) {
					list_append(snake,delete_list);
					break;
				}
			}
		}
		endforeach

		foreach(SNAKE,snake,delete_list) {
			list_remove(snake,part->snakes);
		}
		endforeach
	}
	endforeach
}

/* --------------------------------------------------------------------
 * Procedure: snake_classification_from_modelparts
 * Comments:  
 * -------------------------------------------------------------------- */

void snake_classification_from_modelparts(SNAKE *snake)
{
	int *class,i,last;
	float *cvector,total;

	last=snake->path->length-1;
	total=(float)list_length(snake->modelparts[last]);
	cvector=(float *)iprs_calloc(nclasses,sizeof(float));

	if (total > 0.0) {

		class=(int *)iprs_calloc(nclasses,sizeof(int));
		for (i=0;i<nclasses;i++) {
			class[i]=0;
		}
	
		foreach(PART,model,snake->modelparts[last]) {
			(class[model->class])++;
		}
		endforeach

		for (i=0;i<nclasses;i++) {
			cvector[i]=(float)class[i]/total;
		}
	}
	else {
		for (i=0;i<nclasses;i++) {
			cvector[i]=0.0;
		}
	}

	for (i=0;i<nclasses;i++) {
		snake->cvector[i]=cvector[i];
	}
	iprs_free(cvector);
}

/* ====================================================================
 * Package:		single snake - single classification - constraint
 * ==================================================================== */

/* --------------------------------------------------------------------
 * Procedure: single_classification_constraint
 * Comments:  
 * -------------------------------------------------------------------- */

int single_classification_constraint()
{
	int i,change;
	PART *part;
	LIST slist;

	slist=NIL;

	/* search and print all single classification snakes */

	if (verbosity > 0) printf("\nSingle classification snakes\n");
	foreach(PART,part,scene->parts) {
		if (list_length(part->snakes) == 1) {
			SNAKE *snake;
			snake=list_first(part->snakes);
			if (list_length(snake->classifications) == 1) {
				list_append(snake,slist);
				if (verbosity > 0) {
					printf("Part %d snake ",part->name);
					print_path(snake->path);
					printf("\n");
				}
			}
		}
	}
	endforeach

	if (list_empty(slist)) return 0;

	change=0;
	foreach(SNAKE,snake1,slist) {
		for (i=0;i<snake1->path->length;i++) {
			part=snake1->path->part[i];
			foreach(SNAKE,snake2,part->touched_snakes) {
				if (constrain_model_parts(snake1,snake2)) change=1;
			}
			endforeach
		}
	}
	endforeach

	return change;
}

/* --------------------------------------------------------------------
 * Procedure: constrain_model_parts
 * Comments:  same as intersect_model_parts except that modelparts of
 *            snake2 but NOT snake1 are constrained.
 * -------------------------------------------------------------------- */

int constrain_model_parts(SNAKE *snake1, SNAKE *snake2)
{
	int change,i,j,m,m2;
	LIST new;

	change=0;
	for (i=0; i<snake1->path->length; i++) {
		for (j=0; j<snake2->path->length; j++) {
			if ((snake1->path->part[i] == snake2->path->part[j]) &&
				(snake1->modelparts[i] != snake2->modelparts[j])) {
				new=list_intersection(snake1->modelparts[i],
									  snake2->modelparts[j]);
				m=list_length(new);
				m2=list_length(snake2->modelparts[j]);
				if (m<m2) {
					snake2->modelparts[j]=new;
					change=1;
				}
			}
		}
	}

	if (verbosity>1) {
		if (change) {
			printf("Intersecting ");
			print_path(snake1->path);
			printf(" with ");
			print_path(snake2->path);
			printf("\n");
		}
	}

	return change;
}

/* --------------------------------------------------------------------
 * Procedure: intersnake_compatibility
 * Comments:  compute average (neighbourhood) compatibility of each snake
 * -------------------------------------------------------------------- */

void analyze_intersnake_compatibility()
{
	int i,n;
	float w,total;
	PATH *path;
	LIST slist;

	foreach(PART,part,scene->parts) {
		foreach(SNAKE,snake1,part->snakes) {

			if (verbosity > 0) {
				printf("\nAnalysis of part %d, snake ",part->name);
				print_path(snake1->path);
				printf("\n");
			}

			/* get all snakes intersecting this snake */

			slist=NIL;
			path=snake1->path;
			for (i=0;i<path->length;i++) {
				foreach(SNAKE,snake2,path->part[i]->touched_snakes) {
					if (snake1 != snake2) {
						snake_list_union(snake2,slist);
					}
				}
				endforeach
			}

			/* compute compatibility of snake with all  */
			/* intersecting ones                        */

			total=0.0;
			n=list_length(slist);
			foreach(SNAKE,snake2,slist) {
				w=snake_compatibility(snake1,snake2);
				total += w;
			}
			endforeach
			if (n>0) snake1->compatibility=total/n;
			else snake1->compatibility=0;
			list_free(slist);
		}
		endforeach
	}
	endforeach

	if (verbosity > 0) {
		foreach(PART,part,scene->parts) {
			foreach(SNAKE,snake,part->snakes) {
				printf("Compat of ");
				print_path(snake->path);
				printf(" = %f\n",snake->compatibility);
			}
			endforeach
		}
		endforeach
	}
}

/* --------------------------------------------------------------------
 * Procedure: snake_compatibility
 * Comments:  compute compatibility between two snakes:
 *            For all parts in intersection of the snakes
 *
 *            Cardinality(intersection(mp(part|snake1),mp(part|snake2)))
 *            ----------------------------------------------------------
 *            Cardinality(union(mp(part|snake1),mp(part|snake2)))
 *
 * -------------------------------------------------------------------- */

float snake_compatibility(SNAKE *snake1, SNAKE *snake2)
{
	int n;
	float total;
	LIST plist;

	plist=snake_intersection(snake1,snake2);
	if ((n=list_length(plist))==0) {
		error_exit("snake_compatibility: zero intersection");
	}
	total=0.0;
	foreach(PART,part,plist) {
		total += part_snake_compatibility(part,snake1,snake2);
	}
	endforeach
	total /= (float)n;

	if (total > 0.0 && verbosity>0) {
		printf("Comp [ ");
		print_path(snake1->path);
		printf(" , ");
		print_path(snake2->path);
		printf(" ] = %f\n",total);
	}

	list_free(plist);
	return total;
}

/* --------------------------------------------------------------------
 * Procedure: snake_intersection
 * Comments:  set of parts common to two snakes
 * -------------------------------------------------------------------- */

LIST snake_intersection(SNAKE *snake1,SNAKE *snake2)
{
	int i,j;
	PATH *path1,*path2;
	LIST plist;

	plist=NIL;
	path1=snake1->path;
	path2=snake2->path;

	for (i=0;i<path1->length;i++) {
		for (j=0;j<path2->length;j++) {
			if (path1->part[i] == path2->part[j]) {
				list_append(path1->part[i],plist);
			}
		}
	}

	return plist;
}

/* --------------------------------------------------------------------
 * Procedure: part_snake_compatibility
 * Comments:  see snake_compatibility
 * -------------------------------------------------------------------- */

float part_snake_compatibility(PART *part,SNAKE *snake1,SNAKE *snake2)
{

	int i1,i2,ni,nu;
	float c;
	LIST intersecting;

	for (i1=0;i1<snake1->path->length;i1++) {
		if (part == snake1->path->part[i1]) break;
	}
	if (i1 >= snake1->path->length) {
		error_exit("snake_part_compatibility: empty intersection");
	}

	for (i2=0;i2<snake2->path->length;i2++) {
		if (part == snake2->path->part[i2]) break;
	}
	if (i2 >= snake2->path->length) {
		error_exit("snake_part_compatibility: empty intersection");
	}
	
	intersecting = list_intersection(snake1->modelparts[i1],
									 snake2->modelparts[i2]);
	ni=list_length(intersecting);
	list_free(intersecting);
	nu=list_length(snake1->modelparts[i1]);
	foreach(PART,modelpart,snake2->modelparts[i2]) {
		if (!list_member(modelpart,snake1->modelparts[i1])) nu++;
	}
	endforeach;

	if (nu>0) c=(float)ni/(float)nu;
	else c=0.0;
	return c;
}

/* ====================================================================
 * Package:		relaxation labeling analysis
 * ==================================================================== */

/* --------------------------------------------------------------------
 * Procedure: relaxation labeling
 * Comments:
 * -------------------------------------------------------------------- */

void relaxation_labeling()
{
	int i,iteration,niterations=10;
	float c,*cvector,total,winter,wintra;
	float *compat_cvector;

	cvector=(float *)iprs_calloc(nclasses,sizeof(float));
	compat_cvector=(float *)iprs_calloc(nclasses,sizeof(float));

	foreach (PART,part,scene->parts) {
		part->terminate_relaxation=0;
	}
	endforeach

	for (iteration=1;iteration<=niterations;iteration++) {

		foreach (PART,part,scene->parts) {

			if (part->terminate_relaxation) continue;

			if (verbosity>1) printf("relaxing part %d\n", part->name);
			for (i=0;i<nclasses;i++) {
				cvector[i]=0.0;
			}

			foreach(SNAKE,snake,part->snakes) {
				winter=snake->compatibility;
/* 				wintra=intrasnake_compatibility(snake); */
				brendans_snake_compatibility(snake, compat_cvector);
				for (i=0;i<nclasses;i++) {
/* 					cvector[i] += winter * wintra * snake->cvector[i]; */
					cvector[i] += winter * compat_cvector[i] * 
						snake->cvector[i];
				}
				if (verbosity>1)
				{
					printf("compat: ");
					print_classification_vector(compat_cvector);
					printf("\nsnake: "); print_path(snake->path);
					print_classification_vector(snake->cvector);
					printf("\n");
/* 					printf(" wer=%f, wra=%e\n", winter, wintra); */
				}
			}
			endforeach;

			if (verbosity>1)
			{
				printf("original class vector: ");
				print_classification_vector(part->cvector);
				printf("class vector after inter/intra: ");
				print_classification_vector(cvector);
				printf("\n");
			}
			/* normalization */

			total=0.0;
			for (i=0;i<nclasses;i++) total+=cvector[i];
			if (total <= 0.0) {
				if (verbosity > 1) {
					printf("relaxation1: %e for %d\n",total,part->name);
				}
				part->terminate_relaxation=1;
				continue;
			}
			for (i=0;i<nclasses;i++) {
				cvector[i] /= total;
			}

			if (verbosity>1)
			{
				printf("class vector before transform: ");
				print_classification_vector(cvector);
				printf("\n");
			}
			/* logistic transform */

/* 			for (i=0;i<nclasses;i++) { */
/* 				c=cvector[i]; */
/*  				c=(1-exp((double)(-(c-0.4)/0.05)))/ */
/*  				  (1+exp((double)(-(c-0.4)/0.05))); */
/*  				c=0.5+c/2; */
/*  				cvector[i] = c; */
/* 			} */

			if (verbosity>1)
			{
				printf("class vector after transform: ");
				print_classification_vector(cvector);
				printf("\n");
			}
			/* normalization */

			total=0.0;
			for (i=0;i<nclasses;i++) total+=cvector[i];
			if (total <= 0.0) {
				if (verbosity > 0) {
					printf("relaxation2: %e for %d\n",total,part->name);
				}
				part->terminate_relaxation=1;
				continue;
			}

			for (i=0;i<nclasses;i++) {
				cvector[i] /= total;
			}

			if (verbosity>1)
			{
				printf("class vector after normal: ");
				print_classification_vector(cvector);
				printf("\n");
			}
			/* save new evidence vector */

			for (i=0;i<nclasses;i++) {
				part->cvector1[i]=cvector[i];
			}
		}
		endforeach

		/* Update evidence vectors for next iteration */

		foreach(PART,part,scene->parts) {
			if (!part->terminate_relaxation) {
				for (i=0;i<nclasses;i++) {
					part->cvector[i]=part->cvector1[i];
				}
			}
		}
		endforeach

		if (verbosity > 0) {
			printf("\nIteration %d\n",iteration);
			print_part_classification();
		}
	}
}

/* --------------------------------------------------------------------
 * Procedure: intrasnake_compatibility
 * Comments:  my measure
 * -------------------------------------------------------------------- */


float intrasnake_compatibility(SNAKE *snake)
{
	int i,p;
	float result;
	static float *cvector;
	static int firstcall=1;

	if (firstcall) {
		cvector=(float *)iprs_calloc(nclasses,sizeof(float));
		firstcall=0;
	}

	for (i=0;i<nclasses;i++) {
		cvector[i]=1.0;
	}

	for (p=0;p<snake->path->length;p++) {
		if (verbosity>1)
		{
			printf("snake[%d] : ", p);
			print_classification_vector(snake->path->part[p]->cvector);
			printf("\n");
		}
		for (i=0;i<nclasses;i++) {
			cvector[i] *= snake->path->part[p]->cvector[i];
		}
	}

	result=0.0;
	for (i=0;i<nclasses;i++) {
		result += cvector[i];
	}

	return result;
}

/* --------------------------------------------------------------------
 * Procedure: intrasnake_compatibility
 * Comments:  Brendan McCane's measure.
 * -------------------------------------------------------------------- */

void brendans_snake_compatibility(SNAKE *snake, float *cvector)
{
	int i,p;
	float result;

	for (i=0;i<nclasses;i++) {
		cvector[i]=0.0;
	}

	for (p=0;p<snake->path->length;p++) {
		if (verbosity>1)
		{
			printf("snake[%d] : ", p);
			print_classification_vector(snake->path->part[p]->cvector);
			printf("\n");
		}
		for (i=0;i<nclasses;i++) {
			cvector[i] += snake->path->part[p]->cvector[i];
		}
	}

	result=0.0;
	for (i=0;i<nclasses;i++) {
		cvector[i] /= (float)snake->path->length;
	}
}

/* --------------------------------------------------------------------
 * Procedure: intrasnake_compatibility
 * Comments:  Dennis Moore's measure
 * -------------------------------------------------------------------- */

/*
** float intrasnake_compatibility(SNAKE *snake)
** {
** 	int i,j,k,len;
** 	float c1,c2,s11,s12,s22,angle,result;
** 
** 	len=snake->path->length;
** 
** 	for (i=0;i<len-1;i++) {
** 		for (j=i+1;j<len;j++) {
** 			s11=0.0;
** 			s12=0.0;
** 			s22=0.0;
** 			for (k=0;k<nclasses;k++) {
** 				c1=snake->path->part[i]->cvector[k];
** 				c2=snake->path->part[j]->cvector[k];
** 				s11 += c1*c1;
** 				s12 += c1*c2;
** 				s22 += c2*c2;
** 			}
**			angle=s12/sqrtf(s11*s22);
**			result += angle;
** 		}
** 	}
** 
** 	result /= (float)(len*(len-1));
** 
** 	return result;
** }
*/
