/*****************************************************************************\
*******************************************************************************
*******                                                                 *******
*****   FILE:             bren_reorder_edge                               *****
*****   AUTHOR:           brendan (mccane@coral.cs.jcu.edu.au)            *****
*****   DATE:             12/1/93                                         *****
*****   COMMENTS:         Reorders an edge so that edge points are in     *****
*****					  their natural order from one endpoint to        *****
*****					  or from an arbitrary point if the edge is       *****
*****                     a closed contour.                               *****
*****   IMPORTANT:		  The edge must be initially ordered on y-value   *****
*****                     or this algorithm won't work!!!                 *****
*****   MODIFICATIONS:                                                    *****
*******                                                                 *******
*******************************************************************************
\*****************************************************************************/


#include "iprs.h"
#include "list.h"

#define INVALID -500

int x_compare(POINT *pt1, POINT *pt2)
{
	if (pt1->x<pt2->x) return(-1);	/* order on x first */
	else if (pt1->x==pt2->x) 
	{
		if (pt1->y<pt2->y) return(-1);		/* then order on y */
		else if (pt1->y==pt2->y) return(0);
		else return(1);
	}
	else return(1);
}

/************* orders according to y-value *****************/

void bren_reorder_xedge(IPRS_EDGE *edge)
{
	POINT *new_edge;
	int i;

	iprs_debug_enter("bren_reorder_yedge");

	new_edge = (POINT *)iprs_calloc(edge->numpixels, sizeof(POINT));
	for (i=0; i<edge->numpixels; i++)
	{
		new_edge[i].x = edge->x[i];
		new_edge[i].y = edge->y[i];
	}

	qsort(new_edge, edge->numpixels, sizeof(POINT), 
			(int (*)())x_compare);

	for (i=0; i<edge->numpixels; i++)
	{
		edge->x[i] = new_edge[i].x;
		edge->y[i] = new_edge[i].y;
	}

	bren_free(new_edge, iprs_free);

	iprs_debug_exit();
}

int y_compare(POINT *pt1, POINT *pt2)
{
	if (pt1->y<pt2->y) return(-1);	/* order on y first */
	else if (pt1->y==pt2->y) 
	{
		if (pt1->x<pt2->x) return(-1);		/* then order on x */
		else if (pt1->x==pt2->x) return(0);
		else return(1);
	}
	else return(1);
}

/************* orders according to y-value *****************/

void bren_reorder_yedge(IPRS_EDGE *edge)
{
	POINT *new_edge;
	int i;

	iprs_debug_enter("bren_reorder_yedge");

	new_edge = (POINT *)iprs_calloc(edge->numpixels, sizeof(POINT));
	for (i=0; i<edge->numpixels; i++)
	{
		new_edge[i].x = edge->x[i];
		new_edge[i].y = edge->y[i];
	}

	qsort(new_edge, edge->numpixels, sizeof(POINT), 
			(int (*)())y_compare);

	for (i=0; i<edge->numpixels; i++)
	{
		edge->x[i] = new_edge[i].x;
		edge->y[i] = new_edge[i].y;
	}

	bren_free(new_edge, iprs_free);

	iprs_debug_exit();
}


void bren_reorder_yedgeimage(IPRS_EDGE_IMAGE* edge_image)
{
	int i;

	iprs_debug_enter("bren_reorder_yedgeimage");

	for (i=0; i<edge_image->numedges; i++)
		bren_reorder_yedge(edge_image->edge[i]);

	iprs_debug_exit();
}


/********** performs a natural ordering **************/

int bren_reorder_edge(IPRS_EDGE *edge)
{
	int i, k, neighbour, endpoint2=-1, endpoint1=-1, count;
	int *newx, *newy, nb[3];

	iprs_debug_enter("bren_reorder_edge");

	if (edge->numpixels<=1)
	{	
		iprs_debug_exit();
		return(1);
	}
	bren_reorder_yedge(edge);

	for (i=0; (i<edge->numpixels)&&(endpoint2==-1); i++)
	{
		neighbour = 0;

		k = i-1; /** search backward for neighbouring points **/
		while ((k>=0)&&(k<edge->numpixels)&&(edge->y[k]>=edge->y[i]-1)&&
			   (edge->y[k]<=edge->y[i]+1)&&(neighbour<2))
		{
			if ((edge->x[k]>=edge->x[i]-1)&&(edge->x[k]<=edge->x[i]+1))
				nb[neighbour++]=k;
			k--;
		}

		k = i+1; /** search forward for neighbouring points **/
		while ((k>=0)&&(k<edge->numpixels)&&(edge->y[k]>=edge->y[i]-1)&&
			   (edge->y[k]<=edge->y[i]+1)&&(neighbour<3))
		{
			if ((edge->x[k]>=edge->x[i]-1)&&(edge->x[k]<=edge->x[i]+1))
				nb[neighbour++]=k;
			k++;
		}

		/*
		 * Due to configurations like this :
		 *
		 *           ####
		 *              #
		 *
		 * where the hashes represent a line. The bottom # which is a
		 * valid endpoint will not be considered as such because it has
		 * two valid neighbours. This problem can be corrected by
		 * checking that the distance between the two neighbours of a
		 * point is less than sqrt(2) (the dist between diagonal nbrs).
		 */

		if (neighbour == 2)
		{
			if (euclid_dist(edge->x[nb[0]], edge->y[nb[0]], edge->x[nb[1]],
							edge->y[nb[1]]) < 1.4)
				neighbour = 1;
		}

/* only one neighbour, must be an endpoint */
		if ((neighbour == 1)&&(endpoint1==-1))
			endpoint1 = i;
		else if ((neighbour==1)&&(endpoint2==-1))
			endpoint2 = i;
	}

/** want to start at the lowest y-value **/
	if ((endpoint1==-1)&&(endpoint2==-1))	/* closed curve */
	{
		neighbour=0;
	}
	else if ((endpoint1==-1)||(endpoint2==-1))	/* only one endpoint */
	{											/* must be single pt edge */
		fprintf(stdout, "Only one endpoint!!!\n");
		iprs_debug_exit();
		/* This was return(1) but fucks me why */
		return(0);
	}
	else
	{
		float dist1, dist2;

		dist1 = euclid_dist(edge->x[endpoint1], edge->y[endpoint1], 0, 0);
		dist2 = euclid_dist(edge->x[endpoint2], edge->y[endpoint2], 0, 0);

		neighbour = ((dist1<=dist2) ? endpoint1 : endpoint2);
	}

	newx = (int *)iprs_calloc(edge->numpixels, sizeof(int));
	newy = (int *)iprs_calloc(edge->numpixels, sizeof(int));

	count = 0;
	while ((neighbour!=-1)&&(count<edge->numpixels))
	{
		int tmp;

		i = neighbour;
		newx[count] = edge->x[i];
		newy[count++] = edge->y[i];
		
		neighbour = tmp = -1;
		for (k=0; k<edge->numpixels; k++)
		{
			if (k==i) continue;
			if ((edge->x[k]>=edge->x[i]-1)&&(edge->x[k]<=edge->x[i]+1)&&
				(edge->y[k]>=edge->y[i]-1)&&(edge->y[k]<=edge->y[i]+1)&&
				(edge->y[k]!=INVALID))
			{
				if (neighbour==-1) neighbour = k;
				else tmp = k;	
			}
		}

		if ((tmp!=INVALID)&&(tmp!=neighbour)&&	/* two valid neighbours */
			(tmp!=-1))
		{
			float dist1, dist2;
			dist1=euclid_dist(edge->x[i],edge->y[i],edge->x[tmp],edge->y[tmp]);
			dist2=euclid_dist(edge->x[i],edge->y[i],edge->x[neighbour],
													edge->y[neighbour]);
			if (dist1<dist2) neighbour=tmp;
			else if ((fabs(dist1-dist2)<0.0001)&&
					 (edge->x[tmp]>edge->x[neighbour]))
			 	neighbour=tmp;
		}
		edge->x[i] = INVALID;
		edge->y[i] = INVALID;
	}

	bren_free(edge->x, iprs_free);
	bren_free(edge->y, iprs_free);

	edge->x = newx;
	edge->y = newy;

	iprs_debug_exit();
	if (count<edge->numpixels) return(0);
	else return(1);
}

IPRS_EDGE *recursive_follow(IPRS_EDGE *edge, int current_index, 
							IPRS_EDGE *current_edge, int *visited)
{
	int tmp, i, neighbour, k, l;
	IPRS_EDGE *new_edge, *ret_edge;
	int *new_visited;

	if (visited[current_index]) return(iprs_create_edge(0));
	new_visited = iprs_calloc(edge->numpixels, sizeof(int));
	for (i=0; i<edge->numpixels; i++)
		new_visited[i] = visited[i];
	new_visited[current_index] = TRUE;

	/* add the next pixel to the current edge */
	new_edge = bren_addpixel_edge(current_edge, edge->x[current_index], 
								  edge->y[current_index]);

	i = current_index;
		
	/* find valid neighbours for this point - hopefully at most 2 */
	neighbour = tmp = -1;
	for (k=0; k<edge->numpixels; k++)
	{
		if (k==i) continue;
		if ((edge->x[k]>=edge->x[i]-1)&&(edge->x[k]<=edge->x[i]+1)&&
			(edge->y[k]>=edge->y[i]-1)&&(edge->y[k]<=edge->y[i]+1))
		{
			char found=TRUE;
			for (l=0; l<new_edge->numpixels; l++)
				if ((edge->x[k]==new_edge->x[l])&&(edge->y[k]==new_edge->y[l]))
					found=FALSE;
			if (found)
			{
				if (neighbour==-1) neighbour = k;
				else tmp = k;	
			}
		}
	}

	/* if there are two valid neighbours, then follow both possibilities and
	   return the one with the most pixels as the correct ordering */
	if ((tmp!=INVALID)&&(tmp!=neighbour)&&	/* two valid neighbours */
		(tmp!=-1))
	{
		IPRS_EDGE *ed1, *ed2;
		ed1 = recursive_follow(edge, neighbour, new_edge, new_visited);
		ed2 = recursive_follow(edge, tmp, new_edge, new_visited);
		
		if (ed1->numpixels>=ed2->numpixels)
		{
			iprs_free_edge(ed2);
			ret_edge = ed1;
		}
		else
		{
			iprs_free_edge(ed1);
			ret_edge = ed2;
		}
	}
	/* only one neighbour, follow that one */
	else if (neighbour != -1)
	{
		ret_edge = recursive_follow(edge, neighbour, new_edge, new_visited);
	}
	/* no neighbours, at the end of the edge */
	else ret_edge = iprs_ccopy_edge(new_edge);

	iprs_free_edge(new_edge);
	iprs_free(new_visited);
	return(ret_edge);
}

void bren_newreorder_edge(IPRS_EDGE *edge)
{
	int i, k, neighbour, count;
	int *newx, *newy, nb[3];
	IPRS_EDGE *new_edge=NULL;
	int *visited;
	List *endpoints;

	iprs_debug_enter("bren_reorder_edge");

	endpoints = List_init();
	if (edge->numpixels<=1)
	{	
		iprs_debug_exit();
		return;
	}
	bren_reorder_yedge(edge);

	for (i=0; (i<edge->numpixels); i++)
	{
		neighbour = 0;

		k = i-1; /** search backward for neighbouring points **/
		while ((k>=0)&&(k<edge->numpixels)&&(edge->y[k]>=edge->y[i]-1)&&
			   (edge->y[k]<=edge->y[i]+1)&&(neighbour<2))
		{
			if ((edge->x[k]>=edge->x[i]-1)&&(edge->x[k]<=edge->x[i]+1))
				nb[neighbour++]=k;
			k--;
		}

		k = i+1; /** search forward for neighbouring points **/
		while ((k>=0)&&(k<edge->numpixels)&&(edge->y[k]>=edge->y[i]-1)&&
			   (edge->y[k]<=edge->y[i]+1)&&(neighbour<3))
		{
			if ((edge->x[k]>=edge->x[i]-1)&&(edge->x[k]<=edge->x[i]+1))
				nb[neighbour++]=k;
			k++;
		}

		/*
		 * Due to configurations like this :
		 *
		 *           ####
		 *              #
		 *
		 * where the hashes represent a line. The bottom # which is a
		 * valid endpoint will not be considered as such because it has
		 * two valid neighbours. This problem can be corrected by
		 * checking that the distance between the two neighbours of a
		 * point is less than sqrt(2) (the dist between diagonal nbrs).
		 */

		if (neighbour == 2)
		{
			if (euclid_dist(edge->x[nb[0]], edge->y[nb[0]], edge->x[nb[1]],
							edge->y[nb[1]]) < 1.4)
				neighbour = 1;
		}

/* only one neighbour, must be an endpoint */
		if (neighbour == 1)
			List_insert_head(endpoints, (void*)(i+1));
	}

/** want to start at the lowest y-value **/
	if (List_number_elements(endpoints)==0)	/* closed curve */
	{
		neighbour=0;
		visited = (int*)iprs_calloc(edge->numpixels, sizeof(int));
		new_edge = recursive_follow(edge, neighbour, NULL, visited);
		iprs_free(visited);
	}
	/*
	 * start at all possible endpoints and choose the one that results in
	 * the ordering with the most pixels.
	 */
	else
	{
		IPRS_EDGE *tmp_edge;
		void *info;

		if (List_number_elements(endpoints)==1)	/* only one endpoint */
		{											/* must be single pt edge */
			fprintf(stderr,"Curve has only one endpoint - impossible\n");
			fprintf(stdout, "Only one endpoint - reordering anyway!!!\n");
		}
	    while ((info=List_delete_head(endpoints)))
		{
			neighbour = (int)(info)-1;
			visited = (int*)iprs_calloc(edge->numpixels, sizeof(int));
			tmp_edge = recursive_follow(edge, neighbour, NULL, visited);
			iprs_free(visited);
			if (!new_edge)
				new_edge = tmp_edge;
			else if (tmp_edge->numpixels>new_edge->numpixels)
			{
				iprs_free_edge(new_edge);
				new_edge = tmp_edge;
			}
		}
	}

	List_delete_list(endpoints);
	bren_free(edge->x, iprs_free);
	bren_free(edge->y, iprs_free);

	edge->x = new_edge->x;
	edge->y = new_edge->y;
	edge->numpixels = new_edge->numpixels;

	iprs_debug_exit();
}


void bren_reorder_edgeimage(IPRS_EDGE_IMAGE *edge_image)
{
	int i;

	iprs_debug_enter("bren_reorder_edgeimage");

	for (i=0; i<edge_image->numedges; i++)
		bren_newreorder_edge(edge_image->edge[i]);

	iprs_debug_exit();
}
