/*****************************************************************************\
*******************************************************************************
*******                                                                 *******
*****   FILE:             list.c                                          *****
*****   AUTHOR:           brendan (mccane@coral.cs.jcu.edu.au)            *****
*****   DATE:             23/7/93                                         *****
*****   COMMENTS:         Generic linked list structure.                  *****
*****   MODIFICATIONS:                                                    *****
*******                                                                 *******
*******************************************************************************
\*****************************************************************************/

#include "list.h"
#include "iprs.h"

Node *Node_alloc(void *info)
{
	Node *tmp;

	iprs_debug_enter("Node_alloc");
	tmp = (Node *)iprs_calloc(1, sizeof(Node));
	tmp->next = tmp->prev = NULL;
	tmp->info = info;
	iprs_debug_exit();
	return(tmp);
}
	
List *List_init()
{
	List *list;

	iprs_debug_enter("List_init");
	list = (List*)iprs_calloc(1, sizeof(List));
	list->head = list->tail = list->current = NULL;
	list->num_elts = 0;

	iprs_debug_exit();
	return(list);
}

List *List_ccopy(List *list)
{
	List *copy;
	int i;

	iprs_debug_enter("List_ccopy");
	copy = List_init();
	List_reset_current(list);
	for (i=0; i<list->num_elts; i++)
		List_insert_tail(copy, List_find_next(list));

	iprs_debug_exit();
	return(copy);
}	

void List_insert_tail(List *list, void *info)
{
	Node *tmp;

	iprs_debug_enter("List_insert_tail");
	if (list->tail == NULL)		/* NULL list */
	{
		list->head = list->tail = Node_alloc(info);
		list->num_elts = 1;
		iprs_debug_exit();
		return;
	}	

	tmp = Node_alloc(info);
	list->tail->next = tmp;
	tmp->prev = list->tail;
	list->tail = tmp;
	list->num_elts++;
	iprs_debug_exit();
}

void List_insert_head(List *list, void *info)
{
	Node *tmp;

	iprs_debug_enter("List_insert_head");
	if (list->head == NULL)		/* empty list */
	{
		list->head = list->tail = Node_alloc(info);
		list->num_elts = 1;
		iprs_debug_exit();
		return;
	}

	tmp = Node_alloc(info);
	list->head->prev = tmp;
	tmp->next = list->head;
	list->head = tmp;
	list->num_elts++;
	iprs_debug_exit();
}

void List_insert_before(List *list, void *info)
{
	Node *tmp;

	iprs_debug_enter("List_insert_before");
	if (list->head == NULL)		/* empty list */
	{
		list->head = list->tail = Node_alloc(info);
		list->num_elts = 1;
		iprs_debug_exit();
		return;
	}

	if (list->current==list->head)
	{
		List_insert_head(list, info);
		iprs_debug_exit();
		return;
	}

	tmp = Node_alloc(info);
	tmp->prev = list->current->prev;
	tmp->next = list->current;
	tmp->prev->next = tmp;
	list->current->prev = tmp; 
	list->num_elts++;
	iprs_debug_exit();
}

void List_insert_ordered(List *list, void *info, int (*comp)(void *, void *))
{

	iprs_debug_enter("List_insert_ordered");
	if (list->head == NULL)		/* empty list */
	{
		list->head = list->tail = Node_alloc(info);
		list->num_elts = 1;
		list->current = list->head;
		iprs_debug_exit();
		return;
	}

	list->current = list->head;
	while (list->current)
	{
		if (comp(list->current->info, info)>0)
			list->current = list->current->next;
		else break;
	}
	if (list->current)
	{
		List_insert_before(list, info);
		list->current = list->current->prev;
	}
	else 
	{
		List_insert_tail(list, info);
		list->current = list->tail;
	}
	iprs_debug_exit();
}

void *List_find_head(List *list)
{
	return(list->head->info);
}

void *List_find_tail(List *list)
{
	return(list->tail->info);
}

void List_reset_current(List *list)
{
	list->current = list->head;
}

void *List_find_next(List *list)
{
	if (list->head == NULL) return(NULL);
	else if (list->current == NULL) return(NULL);
	else
	{
		void *tmp;

		tmp = list->current->info;
		list->current = list->current->next;
		return(tmp);
	}
}

void *List_delete_head(List *list)
{
	void *tmp;
	Node *head;

	iprs_debug_enter("List_delete_head");
	if (list->head == NULL)
	{
		iprs_debug_exit();
		return(NULL);
	}

	tmp = list->head->info;
	head = list->head;
	list->head = list->head->next;
	if (list->head!=NULL) list->head->prev = NULL;
	else list->tail=NULL;
	bren_free(head, iprs_free);
	list->num_elts--;
	iprs_debug_exit();

	return(tmp);
}

void *List_delete_tail(List *list)
{
	void *tmp;
	Node *node;

	iprs_debug_enter("List_delete_tail");
	if (list->tail == NULL)
	{
		iprs_debug_exit();
		return(NULL);
	}

	tmp = list->tail->info;
	node = list->tail;
	list->tail = list->tail->prev;
	if (list->tail!=NULL) list->tail->next = NULL;
	else (list->head=NULL);
	bren_free(node, iprs_free);
	list->num_elts--;

	iprs_debug_exit();
	return(tmp);
}	

void List_delete_list(List *list)
{
	Node *tmp;

	iprs_debug_enter("List_delete_list");
	tmp = list->head;
	while (tmp!=NULL)
	{
		list->current = tmp->next;
		bren_free(tmp, iprs_free);
		tmp = list->current;
	}
	bren_free(list, iprs_free);
	iprs_debug_exit();
}

void *List_delete_current(List *list)
{
	void *info;
	Node *tmp;

	iprs_debug_enter("List_delete_current");
	if (list->current==NULL)
	{
		iprs_debug_exit();
		return(NULL);
	}

	tmp = list->current;
	info = tmp->info;

	if (tmp->prev != NULL)
		tmp->prev->next = tmp->next;
	else
	{
		iprs_debug_exit();
		return(List_delete_head(list));
	}
	if (tmp->next != NULL)
		tmp->next->prev = tmp->prev;	
	else
	{
		iprs_debug_exit();
		return(List_delete_tail(list));
	}

	bren_free(tmp, iprs_free);
	iprs_debug_exit();
	return(info);
}

void *List_return_info(Node *node)
{
	if (node) return(node->info);
	else return(NULL);
}

char List_find_info(List *list, void *info, int (*comp)(void *, void *))
{
	char found=FALSE;
 
	List_reset_current(list);
	while ((list->current)&&((found=comp(list->current->info, info))>0))
		List_find_next(list);

	return(!found);
}

char List_find_unordered(List *list, void *info, int (*comp)(void *, void *))
{
	char found=TRUE;
 
	List_reset_current(list);
	while ((list->current)&&((found=comp(info, list->current->info))!=0))
		List_find_next(list);

	return(!found);
}

void List_print(List *list, void (*print)(void*))
{
	int i;

	iprs_debug_enter("List_ccopy");
	List_reset_current(list);
	for (i=0; i<list->num_elts; i++)
		print((void*)List_find_next(list));
	List_reset_current(list);
	iprs_debug_exit();
}	

int List_number_elements(List *list)
{
	return(list->num_elts);
}
